package demo.barcharts.xy ;

import java.awt.Font ;
import java.awt.GridLayout ;
import java.text.NumberFormat ;
import java.util.Calendar ;

import javax.swing.JPanel ;

import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.client.ChartFactory ;
import com.ijchart.xychart.client.ChartPanel ;
import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.data.IntervalXNumberYBaseChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.plot.XYPlot ;
import com.ijchart.xychart.render.xy.IntervalXNumberYBaseBar2DRender ;
import com.ijchart.xychart.title.TextTitle ;
import com.ijchart.xychart.tooltip.DefaultTooltipGenerator ;
import com.ijchart.xychart.util.DateUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYBarChartDemo1 extends JPanel {
  private ChartOrientation orientation = ChartOrientation.VERTICAL ;

  private Calendar cal = Calendar.getInstance() ;

  private NumberFormat labelFormat = NumberFormat.getIntegerInstance() ;

  public XYBarChartDemo1() {
    setLayout(new GridLayout(1, 1)) ;
    addPanel() ;
  }

  private void addPanel() {
    final IMatrixDataset dataset = createDataset() ;
    final IJChart chart = ChartFactory.createIntervalXBaseYBar2DChart(
      "State Executions - USA", "Year", "Number of People", dataset, orientation, true,
      true, true, false) ;

    chart.addSubtitle(new TextTitle("Source: http://www.amnestyusa.org/abolish/listbyyear.do",
                                    new Font("Dialog", 2, 10))) ;
    XYPlot xyplot = (XYPlot) chart.getPlot() ;
    IntervalXNumberYBaseBar2DRender render = (IntervalXNumberYBaseBar2DRender) xyplot.getRender() ;
    render.setToolTipGenerator(new DefaultTooltipGenerator() {
      public String generateToolTip(IMatrixDataset dataset, int row, int column) {
        IntervalXNumberYBaseChartData data = (IntervalXNumberYBaseChartData)
                                             dataset.getData(row, column) ;
        cal.setTimeInMillis((long) data.getEndXValue()) ;
        return cal.get(Calendar.YEAR) + " = " + labelFormat.format(data.getYValue()) ;
      }
    }) ;

    AbstractValueAxis valueAxis = xyplot.getDomainAxis() ;
    valueAxis.setLowerMargin(0.01D) ;
    valueAxis.setUpperMargin(0.01D) ;

    ChartPanel chartpanel = new ChartPanel(chart) ;
    add(chartpanel) ;
  }

  private static IMatrixDataset createDataset() {
    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addRowKey("Executions") ;

    int data[][] = { {1976, 0}, {1977, 1}, {1978, 0}, {1979, 2}, {1980, 0}, {1981, 1}, {1982, 2},
                   {1983, 5}, {1984, 21}, {1985, 18}, {1986, 18}, {1987, 25}, {1988, 11}, {1989,
                   16}, {1990, 23}, {1991, 14}, {1992, 31}, {1993, 38}, {1994, 31}, {1995, 56},
                   {1996, 45}, {1997, 74}, {1998, 68}, {1999, 98}, {2000, 85}, {2001, 66}, {2002,
                   71}, {2003, 65}, {2004, 59}, {2005, 60}
    } ;

    dataset.addRowData() ;

    for (int i = 0 ; i < data.length ; i++) {
      long time1 = DateUtil.dateToTimeInMillis(data[i][0] - 1, 8) ;
      long time2 = DateUtil.dateToTimeInMillis(data[i][0], 5) ;
      dataset.addData(0, new IntervalXNumberYBaseChartData(time1, time2, data[i][1])) ;
    }

    return dataset ;
  }

}
