package demo.miscellaneous.dynamiccharts ;

import java.awt.Color ;
import java.awt.GradientPaint ;
import java.awt.GridLayout ;
import java.util.List ;

import javax.swing.JPanel ;

import com.ijchart.xychart.axis.value.NumberValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.client.ChartFactory ;
import com.ijchart.xychart.client.ChartPanel ;
import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.GradientPaintTransformType ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.event.ChartMouseEvent ;
import com.ijchart.xychart.listener.AbstractChartMouseAdapter ;
import com.ijchart.xychart.plot.CategoryPlot ;
import com.ijchart.xychart.render.category.Bar2DRender ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.tickunit.LongTickUnit ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class MouseOverDemo1 extends JPanel {
  private ChartOrientation orientation = ChartOrientation.VERTICAL ;

  public MouseOverDemo1() {
    setLayout(new GridLayout(1, 1)) ;
    addPanel() ;
  }

  private void addPanel() {
    final IMatrixDataset dataset = createDataset() ;
    final IJChart chart = ChartFactory.createCategoryBar2DChart(
      "Mouseover Demo 1", "Category", "Value", dataset, orientation, true, true) ;

    CategoryPlot plot = (CategoryPlot) chart.getPlot() ;
    plot.setDomainGridlinePaint(Color.white) ;
    plot.setValueGridlinePaint(Color.white) ;
    plot.getRender().setBackgroundPaint(Color.lightGray) ;

    NumberValueAxis numberaxis = (NumberValueAxis) plot.getValueAxis() ;
    numberaxis.setTickUnit(new LongTickUnit()) ;

    Bar2DRender render = (Bar2DRender) plot.getRender() ;
    render.setBaseItemDrawBorder(false) ;
    GradientPaint gradientpaint = new GradientPaint(0.0F, 0.0F, Color.blue, 0.0F, 0.0F,
                                                    new Color(0, 0, 64)) ;
    GradientPaint gradientpaint1 = new GradientPaint(0.0F, 0.0F, Color.green, 0.0F, 0.0F,
                                                     new Color(0, 64, 0)) ;
    GradientPaint gradientpaint2 = new GradientPaint(0.0F, 0.0F, Color.red, 0.0F, 0.0F,
                                                     new Color(64, 0, 0)) ;
    render.setAutoSelectFillPaint(false) ;
    render.setBaseItemFillPaintTransformer(GradientPaintTransformType.HORIZONTAL) ;
    render.setColumnItemFillPaint(0, gradientpaint) ;
    render.setColumnItemFillPaint(1, gradientpaint1) ;
    render.setColumnItemFillPaint(2, gradientpaint2) ;

    ChartPanel chartpanel = new ChartPanel(chart) ;
    chartpanel.addChartMouseListener(new AbstractChartMouseAdapter() {
      public void chartMouseMoved(ChartMouseEvent event) {
        RenderShapeCollection renderShapes = (RenderShapeCollection) chart.getChartShapes().
                                             getPlotShapes(0).getAllRenderShapes().get(0) ;
        List dataShapes = renderShapes.getDataItemShapes() ;

        for (int i = 0 ; i < dataShapes.size() ; i++) {
          IChartShape shape = (IChartShape) dataShapes.get(i) ;
          if (shape != null && shape.contains(event.getTrigger().getPoint())) {
            shape.setBorderPaint(Color.yellow) ;
            shape.setDrawBorder(true) ;
          } else {
            shape.setDrawBorder(false) ;
          }
        }
      }

    }) ;
    add(chartpanel) ;
  }

  private static IMatrixDataset createDataset() {
    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addRowKey("Category 1") ;
    dataset.addRowKey("Category 2") ;
    dataset.addRowKey("Category 3") ;
    dataset.addRowKey("Category 4") ;
    dataset.addRowKey("Category 5") ;

    dataset.addColKey("First") ;
    dataset.addColKey("Second") ;
    dataset.addColKey("Third") ;

    double[][] data = { {1.0D, 5D, 4D}, {4D, 7D, 3D}, {3D, 6D, 2D}, {5D, 8D, 3D}, {5D, 4D, 6D}
    } ;
    dataset.addRowData(data.length) ;

    for (int i = 0 ; i < data.length ; i++) {
      for (int j = 0 ; j < 3 ; j++) {
        dataset.addData(i, new NumberChartData(data[i][j])) ;
      }
    }

    return dataset ;
  }

}
