package demo.others.axis;

import java.awt.Color ;
import java.awt.Dimension ;
import java.awt.Graphics ;
import java.awt.Graphics2D ;
import java.awt.GridLayout ;
import java.awt.RenderingHints ;
import java.awt.geom.Rectangle2D ;
import java.util.HashMap ;
import java.util.Map ;

import javax.swing.JPanel ;

import com.ijchart.xychart.axis.category.CategoryAxis ;
import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.axis.value.NumberValueAxis ;
import com.ijchart.xychart.chartinfo.ChartShapeCollection ;
import com.ijchart.xychart.constant.AxisLocation ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.HorizontalAlignment ;
import com.ijchart.xychart.constant.IAngle ;
import com.ijchart.xychart.constant.VerticalAlignment ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.format.NumberChartDataFormatter ;
import com.ijchart.xychart.label.DefaultLabelGenerator ;
import com.ijchart.xychart.marker.CategoryMarker ;
import com.ijchart.xychart.marker.IntervalValueMarker ;
import com.ijchart.xychart.marker.ValueMarker ;
import com.ijchart.xychart.plot.CategoryPlot ;
import com.ijchart.xychart.render.category.Bar2DRender ;
import com.ijchart.xychart.ui.Layer ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class MultipleAxisPanel extends JPanel {
  private Map hashMap = new HashMap() ;

  private IMatrixDataset dataset1 = createDataset1() ;

  private IMatrixDataset dataset2 = createDataset2() ;

  private ChartOrientation orientation = ChartOrientation.VERTICAL ;

  public MultipleAxisPanel() {
    init() ;
  }

  private void init() {
    hashMap.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON) ;
    setLayout(new GridLayout(1, 1)) ;
    addPanel1() ;
  }

  private void addPanel1() {
    JPanel panel = new JPanel() {

      public void paint(Graphics g) {
        super.paintComponent(g) ;
        Graphics2D g2 = (Graphics2D) g ;
        g2.addRenderingHints(hashMap) ;
        setBackground(Color.white) ;
        g2.setPaint(Color.red) ;
        int width = getWidth() ;
        int height = getHeight() ;

//        g2.drawLine(66,6,66,291);

        g2.drawRect(1, 1, width - 2, height - 2) ;
        Rectangle2D plotArea = new Rectangle2D.Double(2, 2, width - 5, height - 5) ;

        CategoryAxis xAxis = createCategoryAxis1() ;
        CategoryMarker marker = new CategoryMarker(dataset1.getRowKey(2)) ;
        marker.setText("Special 1") ;
        marker.setTextVerticalAlign(VerticalAlignment.TOP) ;
        xAxis.addCategoryMarker(Layer.BACKGROUND, marker) ;

        marker = new CategoryMarker(dataset1.getRowKey(4)) ;
        marker.setText("Special 2") ;
        marker.setTextVerticalAlign(VerticalAlignment.TOP) ;
        xAxis.addCategoryMarker(Layer.FOREGROUND, marker) ;
//        xAxis.setAxisVisible(false);


        AbstractValueAxis yAxis = createValueAxis1() ;
        yAxis.setInverted(true) ;

        if (orientation.equals(ChartOrientation.VERTICAL)) {
          ValueMarker valueMarker = new ValueMarker( -4) ;
          valueMarker.setText("Special background value 1") ;
          valueMarker.setTextHorizontalAlign(HorizontalAlignment.RIGHT) ;
          yAxis.addValueMarker(Layer.BACKGROUND, valueMarker) ;

          valueMarker = new ValueMarker( -12) ;
          valueMarker.setText("Special foreground value 2") ;
          valueMarker.setTextHorizontalAlign(HorizontalAlignment.RIGHT) ;
          valueMarker.setTextVerticalAlign(VerticalAlignment.BOTTOM) ;
          yAxis.addValueMarker(Layer.FOREGROUND, valueMarker) ;
        } else if (orientation.equals(ChartOrientation.HORIZONTAL)) {
          ValueMarker valueMarker = new ValueMarker( -4) ;
          valueMarker.setText("Special background value 1") ;
          valueMarker.setTextHorizontalAlign(HorizontalAlignment.LEFT) ;
          valueMarker.setTextVerticalAlign(VerticalAlignment.TOP) ;
          yAxis.addValueMarker(Layer.BACKGROUND, valueMarker) ;

          valueMarker = new ValueMarker( -12) ;
          valueMarker.setText("Special foreground value 2") ;
          valueMarker.setTextHorizontalAlign(HorizontalAlignment.RIGHT) ;
          valueMarker.setTextVerticalAlign(VerticalAlignment.TOP) ;
          yAxis.addValueMarker(Layer.FOREGROUND, valueMarker) ;
        }

        IntervalValueMarker intervalMarker = new IntervalValueMarker( -5, -10) ;
        intervalMarker.setText("special foreground interval marker 1") ;
        yAxis.addIntervalValueMarker(Layer.FOREGROUND, intervalMarker) ;

        intervalMarker = new IntervalValueMarker( -12, -18) ;
        yAxis.addIntervalValueMarker(Layer.BACKGROUND, intervalMarker) ;
        intervalMarker.setText("special background interval marker 2") ;

//        yAxis.setAxisVisible(false);

        Bar2DRender render = new Bar2DRender(dataset1) ;
        render.setLabelGenerator(new DefaultLabelGenerator(new NumberChartDataFormatter("#0.00"))) ;
        render.setMaximumBarWidth(20) ;

        CategoryPlot plot = new CategoryPlot(xAxis, yAxis, render) ;
        plot.setCategoryAxis(1, createCategoryAxis2()) ;
        plot.setCategoryAxisLocation(1, AxisLocation.TOP_OR_RIGHT) ;
        plot.setValueAxis(1, createValueAxis2()) ;
        plot.setValueAxisLocation(1, AxisLocation.BOTTOM_OR_RIGHT) ;

        plot.setCategoryAxis(2, createCategoryAxis3()) ;
        plot.setCategoryAxisLocation(2, AxisLocation.TOP_OR_RIGHT) ;
        plot.setCategoryAxis(3, createCategoryAxis4()) ;
        plot.setCategoryAxisLocation(3, AxisLocation.BOTTOM_OR_LEFT) ;

        plot.setValueAxis(2, createValueAxis3()) ;
        plot.setValueAxisLocation(2, AxisLocation.BOTTOM_OR_RIGHT) ;
        plot.setValueAxis(3, createValueAxis4()) ;
        plot.setValueAxisLocation(3, AxisLocation.TOP_OR_LEFT) ;

        ChartShapeCollection shapes = new ChartShapeCollection() ;
        shapes.setDrawArea(plotArea) ;
        plot.setOrientation(orientation) ;
        plot.draw(g2, plotArea, shapes) ;

        shapes.draw(g2) ;
      }
    } ;

    panel.setPreferredSize(new Dimension(300, 300)) ;
    add(panel) ;
  }

  private CategoryAxis createCategoryAxis1() {
    CategoryAxis axis = new CategoryAxis("category label1", dataset1) ;
    if (orientation.equals(ChartOrientation.HORIZONTAL)) {
      axis.setAxisLabelAngle(IAngle.ANGLE_DOWN_90) ;
    }

    return axis ;
  }

  private CategoryAxis createCategoryAxis2() {
    CategoryAxis axis = new CategoryAxis("category label2", dataset1) ;

    return axis ;
  }

  private CategoryAxis createCategoryAxis3() {
    CategoryAxis axis = new CategoryAxis("category label3", dataset2) ;
    return axis ;
  }

  private CategoryAxis createCategoryAxis4() {
    CategoryAxis axis = new CategoryAxis("category label4", dataset2) ;
    return axis ;
  }

  private AbstractValueAxis createValueAxis1() {
    AbstractValueAxis valueAxis = new NumberValueAxis("value label1", dataset1) ;
    if (orientation.equals(ChartOrientation.VERTICAL)) {
      valueAxis.setAxisLabelAngle(IAngle.ANGLE_DOWN_90) ;
    }
//
//    valueAxis.setAxisLabelAngle(Math.PI ) ;
//    valueAxis.setAxisLabelHorizontalAlign(HorizontalAlignment.RIGHT) ;

    return valueAxis ;
  }

  private AbstractValueAxis createValueAxis2() {
    AbstractValueAxis valueAxis = new NumberValueAxis("value label2", dataset1) ;
    valueAxis.setAxisLabelAngle(IAngle.ANGLE_DOWN_90) ;

//    valueAxis.setAxisLabelAngle(Math.PI ) ;
//    valueAxis.setAxisLabelHorizontalAlign(HorizontalAlignment.CENTER) ;

    return valueAxis ;
  }

  private AbstractValueAxis createValueAxis3() {
    AbstractValueAxis valueAxis = new NumberValueAxis("value label3", dataset2) ;
    valueAxis.setAxisLabelAngle(IAngle.ANGLE_DOWN_90) ;

    return valueAxis ;
  }

  private AbstractValueAxis createValueAxis4() {
    AbstractValueAxis valueAxis = new NumberValueAxis("value label4", dataset2) ;
    valueAxis.setAxisLabelAngle(IAngle.ANGLE_DOWN_90) ;

    return valueAxis ;
  }

  private static IMatrixDataset createDataset1() {
    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addRowKey("123456789011") ;
    dataset.addRowKey("ab\ncd\nef") ;
    dataset.addRowKey("abcde") ;
    dataset.addRowKey("12345678") ;
    dataset.addRowKey("") ;

    for (int i = 0 ; i < 5 ; i++) {
      dataset.addRowData() ;
      for (int j = 0 ; j < 6 ; j++) {
        dataset.addData(i, new NumberChartData((i + 1) * -j + 1 + Math.random())) ;
      }
    }

    return dataset ;
  }

  private static IMatrixDataset createDataset2() {
    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addRowKey("123456789011") ;
    dataset.addRowKey("abncdnef") ;
    dataset.addRowKey("abcde") ;
    dataset.addRowKey("12345678") ;
    dataset.addRowKey("ABCD") ;

    for (int i = 0 ; i < 5 ; i++) {
      dataset.addRowData() ;
      for (int j = 0 ; j < 5 ; j++) {
        dataset.addData(i, new NumberChartData((i + 1) * j * 10 + 1)) ;
      }
    }

    return dataset ;
  }

}
