package demo.piecharts ;

import java.awt.GridLayout ;
import java.awt.event.ActionEvent ;
import java.awt.event.ActionListener ;
import java.text.NumberFormat ;

import javax.swing.JPanel ;
import javax.swing.Timer ;

import com.ijchart.xychart.client.ChartFactory ;
import com.ijchart.xychart.client.ChartPanel ;
import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.dataset.SingleRowDataset ;
import com.ijchart.xychart.label.DefaultLabelGenerator ;
import com.ijchart.xychart.plot.NoAxisPlot ;
import com.ijchart.xychart.render.noaxis.PieRender ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class PieChartDemo7 extends JPanel {
  private NumberFormat format = NumberFormat.getPercentInstance() ;

  private IJChart chart ;

  public PieChartDemo7() {
    format.setMaximumFractionDigits(0) ;
    setLayout(new GridLayout(1, 1)) ;
    addPanel() ;
    new Rotator().start() ;
  }

  class Rotator extends Timer implements ActionListener {

    private int angle ;

    private Rotator() {
      super(200, null) ;
      angle = 270 ;
      addActionListener(this) ;
    }

    public void actionPerformed(ActionEvent actionevent) {
      NoAxisPlot plot = (NoAxisPlot) chart.getPlot() ;
      PieRender render = (PieRender) plot.getRender() ;
      render.setStartAngle(angle) ;

      angle += 1 ;
      if (angle == 360) {
        angle = 0 ;
      }
    }

  }

  private void addPanel() {
    SingleRowDataset dataset = createDataset() ;
    chart = ChartFactory.createPieChart("Pie Chart Demo 7", dataset, true, true) ;
    NoAxisPlot plot = (NoAxisPlot) chart.getPlot() ;
    plot.setNoDataMessage("No data available") ;

    double total = 0.0 ;
    for (int i = 0 ; i < dataset.getColumnCount() ; i++) {
      total += ((NumberChartData) dataset.getData(0, i)).getValue() ;
    }
    final double paraTotal = total ;

    PieRender render = (PieRender) plot.getRender() ;
    render.setCircular(true) ;
    render.setLabelGenerator(new DefaultLabelGenerator() {
      public String generateLabel(IMatrixDataset dataset, int row, int column) {
        Object colKey = dataset.getColKey(column) ;
        NumberChartData data = (NumberChartData) dataset.getData(row, column) ;
        double percent = data.getValue() / paraTotal ;
        return colKey + "=" + format.format(percent) ;
      }
    }) ;

    ChartPanel chartpanel = new ChartPanel(chart) ;
    add(chartpanel) ;
  }

  private static SingleRowDataset createDataset() {
    SingleRowDataset dataset = new SingleRowDataset() ;
    dataset.addRowData() ;

    for (int j = 0 ; j < 14 ; j++) {
      double d = 100D * Math.random() ;
      dataset.addColKey("Section " + j) ;
      dataset.addData(0, new NumberChartData(d)) ;
    }

    return dataset ;
  }

}
