package demo.stackedbarcharts ;

import java.awt.Color ;
import java.awt.Font ;
import java.awt.GradientPaint ;
import java.awt.GridLayout ;

import javax.swing.JPanel ;

import com.ijchart.xychart.axis.value.StackedNumberValueAxis ;
import com.ijchart.xychart.client.ChartFactory ;
import com.ijchart.xychart.client.ChartPanel ;
import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.GradientPaintTransformType ;
import com.ijchart.xychart.constant.HorizontalAlignment ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.plot.CategoryPlot ;
import com.ijchart.xychart.render.category.StackedBarRender ;
import com.ijchart.xychart.tickunit.LongTickUnit ;
import com.ijchart.xychart.title.AbstractTitle ;
import com.ijchart.xychart.title.TextTitle ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.ui.RectangleInsets ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class StackedBarChartDemo2 extends JPanel {
  private ChartOrientation orientation = ChartOrientation.HORIZONTAL ;

  public StackedBarChartDemo2() {
    setLayout(new GridLayout(1, 1)) ;
    addPanel1() ;
  }

  private void addPanel1() {
    IMatrixDataset dataset = createDataset() ;
    IJChart chart = ChartFactory.createStackedBarChart("Public Opinion : Torture of Prisoners",
                                                       "Country", "%", dataset, orientation, true, true) ;
    CategoryPlot plot = (CategoryPlot) chart.getPlot() ;

    ((AbstractTitle) chart.getTitle()).setPadding(new RectangleInsets(2D, 0.0D, 0.0D, 0.0D)) ;
    TextTitle texttitle = new TextTitle("Source: http://news.bbc.co.uk/1/hi/world/6063386.stm",
                                        new Font("Dialog", Font.BOLD, 11)) ;
    texttitle.setPosition(RectangleEdge.BOTTOM) ;
    texttitle.setHorizontalAlignment(HorizontalAlignment.RIGHT) ;
    texttitle.setPadding(new RectangleInsets(0.0D, 0.0D, 4D, 4D)) ;
    chart.addSubtitle(texttitle) ;
    TextTitle texttitle1 = new TextTitle("(*) Across 27,000 respondents in 25 countries",
                                         new Font("Dialog", Font.BOLD, 11)) ;
    texttitle1.setPosition(RectangleEdge.BOTTOM) ;
    texttitle1.setHorizontalAlignment(HorizontalAlignment.RIGHT) ;
    texttitle1.setPadding(new RectangleInsets(4D, 0.0D, 2D, 4D)) ;
    chart.addSubtitle(texttitle1) ;
    chart.getLegendTitle().setBorderVisible(false) ;

    plot.setDomainGridlineVisible(true) ;
    StackedNumberValueAxis numberaxis = (StackedNumberValueAxis) plot.getValueAxis() ;
    numberaxis.setTickUnit(new LongTickUnit()) ;

    StackedBarRender render = (StackedBarRender) plot.getRender() ;
    GradientPaint gradientpaint = new GradientPaint(0.0F, 0.0F, Color.green, 0.0F, 0.0F,
                                                    new Color(0, 64, 0)) ;
    Color color = new Color(0, 0, 0, 0) ;
    GradientPaint gradientpaint1 = new GradientPaint(0.0F, 0.0F, Color.red, 0.0F, 0.0F,
                                                     new Color(64, 0, 0)) ;
    render.setAutoSelectFillPaint(false) ;
    render.setColumnItemFillPaint(0, gradientpaint) ;
    render.setColumnItemFillPaint(1, color) ;
    render.setColumnItemFillPaint(2, gradientpaint1) ;
    render.setBaseItemFillPaintTransformer(GradientPaintTransformType.VERTICAL) ;

    ChartPanel chartpanel = new ChartPanel(chart) ;
    add(chartpanel) ;
  }

  private static IMatrixDataset createDataset() {
    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addColKey("Against all torture") ;
    dataset.addColKey("") ;
    dataset.addColKey("Some degree permissible") ;

    dataset.addRowKey("Italy") ;
    dataset.addRowKey("Great Britain") ;
    dataset.addRowKey("USA") ;
    dataset.addRowKey("Israel") ;
    dataset.addRowKey("Russia") ;
    dataset.addRowKey("India") ;
    dataset.addRowKey("Average (*)") ;

    double data1[] = {81D, 72D, 58D, 48D, 43D, 23D, 59D} ;
    double data2[] = {5D, 4D, 6D, 9D, 20D, 45D, 12D} ;
    double data3[] = {14D, 24D, 36D, 43D, 37D, 32D, 29D} ;

    dataset.addRowData(data1.length) ;
    for (int i = 0 ; i < data1.length ; i++) {
      dataset.addData(i, new NumberChartData(data1[i])) ;
      dataset.addData(i, new NumberChartData(data2[i])) ;
      dataset.addData(i, new NumberChartData(data3[i])) ;
    }

    return dataset ;
  }

}
