package demo.timeseriescharts ;

import java.awt.Graphics2D ;
import java.awt.GridLayout ;
import java.awt.geom.Rectangle2D ;
import java.text.SimpleDateFormat ;
import java.util.Calendar ;
import java.util.Date ;

import javax.swing.JPanel ;

import com.ijchart.xychart.axis.value.XYNumberValueAxis ;
import com.ijchart.xychart.chartinfo.AxisShapeCollection ;
import com.ijchart.xychart.client.ChartFactory ;
import com.ijchart.xychart.client.ChartPanel ;
import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.constant.AxisLabelType ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.DateTickUnitType ;
import com.ijchart.xychart.constant.XYValueAxisType ;
import com.ijchart.xychart.data.XYNumberChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.format.IFormatter ;
import com.ijchart.xychart.plot.XYPlot ;
import com.ijchart.xychart.tickunit.DateTickUnit ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.DateUtil ;
import com.ijchart.xychart.util.StringUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class PeriodAxisDemo2 extends JPanel {

  public PeriodAxisDemo2() {
    setLayout(new GridLayout(1, 1)) ;
    addPanel1() ;
  }

  private void addPanel1() {
    IMatrixDataset dataset = createDataset() ;
    IJChart chart = ChartFactory.createXYLineAndShapeChart("Legal & General Unit Trust Prices",
                                                           "Date", "Price Per Unit", dataset,
                                                           ChartOrientation.VERTICAL, true, true,
                                                           true, false) ;
    XYPlot plot = (XYPlot) chart.getPlot() ;
    plot.setDomainCrosshairVisible(true) ;
    plot.setValueCrosshairVisible(true) ;

    XYNumberValueAxis yAxis = (XYNumberValueAxis) plot.getValueAxis() ;
    yAxis.setAxisRangeIncludeZero(false) ;

    XYNumberValueAxis domainAxis = new XYNumberValueAxis("Date", dataset,
                                                         XYValueAxisType.XY_VALUE_AXIS_X, true) {
      protected double drawTickLabel(Graphics2D g2, Rectangle2D plotArea, Rectangle2D axisArea,
                                     RectangleEdge edge, AxisShapeCollection shapeCollection,
                                     double cursor) {
        cursor = super.drawTickLabel(g2, plotArea, axisArea, edge, shapeCollection, cursor) ;

        SimpleDateFormat dateFormat = new SimpleDateFormat("MMM") ;
        double[] ticks = super.valueTicks ;
        String[] secondLineTicks = new String[ticks.length] ;
        for (int i = 0 ; i < ticks.length ; i++) {
          Date date = new Date((long) ticks[i]) ;
          secondLineTicks[i] = dateFormat.format(date) ;
        }
        secondLineTicks = StringUtil.removeRepeatValues(secondLineTicks) ;
        cursor = drawTickLabelReally(g2, plotArea, axisArea, edge, shapeCollection, cursor,
                                     secondLineTicks) ;

        dateFormat = new SimpleDateFormat("yyyy") ;
        String[] thirdLineTicks = new String[ticks.length] ;
        for (int i = 0 ; i < ticks.length ; i++) {
          Date date = new Date((long) ticks[i]) ;
          thirdLineTicks[i] = dateFormat.format(date) ;
        }
        thirdLineTicks = StringUtil.removeRepeatValues(thirdLineTicks) ;
        cursor = drawTickLabelReally(g2, plotArea, axisArea, edge, shapeCollection, cursor,
                                     thirdLineTicks) ;

        return cursor ;
      }

      protected int getTickLabelLines() {
        return 3 ;
      }

    } ;
    domainAxis.setAxisLabelType(AxisLabelType.COLUMN) ;
    domainAxis.setAxisRangeIncludeZero(false) ;
    domainAxis.setTickValueFormatter(new IFormatter() {
      private Calendar cal = Calendar.getInstance() ;

      private SimpleDateFormat format = new SimpleDateFormat("d") ;

      public String format(Object obj, int index) {
        if (obj != null) {
          String str = obj.toString() ;
          cal.setTimeInMillis(Long.parseLong(str)) ;

          return format.format(cal.getTime()) ;
        }

        return null ;
      }
    }) ;

    DateTickUnit unit = new DateTickUnit(DateTickUnitType.DATE_UNIT_DAY) ;
    unit.setDateFormatter(new SimpleDateFormat("dd")) ;
    domainAxis.setTickUnit(unit) ;
    plot.setDomainAxis(domainAxis) ;

    ChartPanel chartpanel = new ChartPanel(chart) ;
    add(chartpanel) ;
  }

  private static IMatrixDataset createDataset() {
    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addRowKey("L&G European Index Trust") ;
    dataset.addRowData() ;

    dataset.addData(0, createData(24, 1, 2004, 181.80000000000001D)) ;
    dataset.addData(0, createData(25, 1, 2004, 167.30000000000001D)) ;
    dataset.addData(0, createData(26, 1, 2004, 153.80000000000001D)) ;
    dataset.addData(0, createData(27, 1, 2004, 167.59999999999999D)) ;
    dataset.addData(0, createData(28, 1, 2004, 158.80000000000001D)) ;
    dataset.addData(0, createData(29, 1, 2004, 148.30000000000001D)) ;
    dataset.addData(0, createData(30, 1, 2004, 153.90000000000001D)) ;
    dataset.addData(0, createData(31, 1, 2004, 142.69999999999999D)) ;
    dataset.addData(0, createData(1, 2, 2004, 123.2D)) ;
    dataset.addData(0, createData(2, 2, 2004, 131.80000000000001D)) ;
    dataset.addData(0, createData(3, 2, 2004, 139.59999999999999D)) ;
    dataset.addData(0, createData(4, 2, 2004, 142.90000000000001D)) ;
    dataset.addData(0, createData(5, 2, 2004, 138.69999999999999D)) ;
    dataset.addData(0, createData(6, 2, 2004, 137.30000000000001D)) ;
    dataset.addData(0, createData(7, 2, 2004, 143.90000000000001D)) ;
    dataset.addData(0, createData(8, 2, 2004, 139.80000000000001D)) ;
    dataset.addData(0, createData(9, 2, 2004, 137D)) ;
    dataset.addData(0, createData(10, 2, 2004, 132.80000000000001D)) ;

    return dataset ;
  }

  private static XYNumberChartData createData(int day, int month, int year, double value) {
    return new XYNumberChartData(DateUtil.dateToTimeInMillis(year, month, day), value) ;
  }

}
