package demo.timeseriescharts;

import java.awt.Color ;
import java.awt.GridLayout ;
import java.util.Calendar ;

import javax.swing.JPanel ;
import javax.swing.JTabbedPane ;

import com.ijchart.xychart.axis.value.XYNumberValueAxis ;
import com.ijchart.xychart.client.ChartFactory ;
import com.ijchart.xychart.client.ChartPanel ;
import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.DateTickUnitType ;
import com.ijchart.xychart.data.XYNumberChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.format.MillisecondsToDateFormatter ;
import com.ijchart.xychart.plot.XYPlot ;
import com.ijchart.xychart.tickunit.DateTickUnit ;
import com.ijchart.xychart.util.DateUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class TimeSeriesDemo13 extends JPanel {

  public TimeSeriesDemo13() {
    setLayout(new GridLayout(1, 1)) ;
    createPanel() ;
  }

  public void createPanel() {
    IMatrixDataset xydataset = createDataset(26) ;
    JPanel panel1 = createPanel1(xydataset) ;

    IMatrixDataset xydataset1 = createDataset(1) ;
    JPanel panel2 = createPanel1(xydataset1) ;

    JTabbedPane jtabbedpane = new JTabbedPane() ;
    jtabbedpane.add("Chart 1", panel1) ;
    jtabbedpane.add("Chart 2", panel2) ;

    add(jtabbedpane) ;
  }

  private static ChartPanel createPanel1(IMatrixDataset dataset) {
    IJChart chart = ChartFactory.createXYLineAndShapeChart("Weekly Data", "Date", "Price", dataset,
                                                           ChartOrientation.VERTICAL,
                                                           true, true, true, false) ;
    XYPlot xyplot = (XYPlot) chart.getPlot() ;
    xyplot.setDomainGridlinePaint(Color.white) ;
    xyplot.setValueGridlinePaint(Color.white) ;

    XYNumberValueAxis xAxis = (XYNumberValueAxis) xyplot.getDomainAxis() ;
    xAxis.setTickUnit(new DateTickUnit(DateTickUnitType.DATE_UNIT_MONTH)) ;
    xAxis.setTickValueFormatter(new MillisecondsToDateFormatter("MMM ''yy")) ;

    XYNumberValueAxis yAxis = (XYNumberValueAxis) xyplot.getValueAxis() ;
    yAxis.setAxisRangeIncludeZero(false) ;

    ChartPanel chartpanel = new ChartPanel(chart) ;
    return chartpanel ;
  }

  private static IMatrixDataset createDataset(int cnt) {
    Calendar cal = Calendar.getInstance() ;

    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addRowKey("Random Data") ;
    dataset.addRowData() ;

    double val = 100 ;
    for (int i = 0 ; i < cnt ; i++) {
      int year = cal.get(Calendar.YEAR) ;
      int month = cal.get(Calendar.MONTH) + 1 ;
      int day = cal.get(Calendar.DAY_OF_MONTH) ;

      dataset.addData(0, createData(day, month, year, val)) ;
      val *= 1.0D + (Math.random() - 0.495D) / 100D ;
      cal.add(Calendar.DAY_OF_MONTH, 7) ;
    }

    return dataset ;
  }

  private static XYNumberChartData createData(int day, int month, int year, double value) {
    return new XYNumberChartData(DateUtil.dateToTimeInMillis(year, month, day), value) ;
  }

}
