package com.ijchart.xychart.data ;

import java.io.Serializable ;

import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>(startX,endX,baseY,y)ĸֵͼݶ,startXendXʾ,λӦǺ,
 * ⲿǲ޸ĵ.</zh>
 * <en>The (startX,endX,baseY,y) Number ChartData object,if that startX and endX express date,
 * the unit should be in millisecond,this class is immutable.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class IntervalXNumberYBaseChartData extends AbstractChartData implements Cloneable,
  Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -1637325626950943463L ;

  /**
   * <zh>startXֵ.</zh>
   * <en>The startX number data.</en>
   */
  private Number startXData ;

  /**
   * <zh>endXֵ.</zh>
   * <en>The endX number data.</en>
   */
  private Number endXData ;

  /**
   * <zh>yֵ.</zh>
   * <en>The y number data.</en>
   */
  private Number yData ;

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param startX double <zh>startXֵ.</zh>
   * <en>The start x value.</en>
   * @param endX double <zh>endXֵ.</zh>
   * <en>The end x value.</en>
   * @param y double <zh>yֵ.</zh>
   * <en>The y value.</en>
   */
  public IntervalXNumberYBaseChartData(double startX, double endX, double y) {
    this(new Double(startX), new Double(endX), new Double(y)) ;
  }

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param startXData double <zh>startXֵ (Ϊnull).</zh>
   * <en>The start x Number Object (null not permitted).</en>
   * @param endXData double <zh>endXֵ (Ϊnull).</zh>
   * <en>The end x Number Object (null not permitted).</en>
   * @param yData Number <zh>yֵ (Ϊnull).</zh>
   * <en>The y Number Object (null not permitted).</en>
   */
  public IntervalXNumberYBaseChartData(Number startXData, Number endXData, Number yData) {
    if (startXData == null) {
      throw new IllegalArgumentException("Null 'startXData' argument.") ;
    }
    if (endXData == null) {
      throw new IllegalArgumentException("Null 'endXData' argument.") ;
    }
    if (yData == null) {
      throw new IllegalArgumentException("Null 'yData' argument.") ;
    }

    this.startXData = startXData ;
    this.endXData = endXData ;
    this.yData = yData ;
  }

  /**
   * <zh>startXֶ.</zh>
   * <en>Returns the start x Number data.</en>
   *
   * @return Number <zh>Ϊnull.</zh>
   * <en>Never null.</en>
   */
  public Number getStartXData() {
    return this.startXData ;
  }

  /**
   * <zh>endXֶ.</zh>
   * <en>Returns the end x Number data.</en>
   *
   * @return Number <zh>Ϊnull.</zh>
   * <en>Never null.</en>
   */
  public Number getEndXData() {
    return this.endXData ;
  }

  /**
   * <zh>yֶ.</zh>
   * <en>Returns the y Number data.</en>
   *
   * @return Number <zh>Ϊnull.</zh>
   * <en>Never null.</en>
   */
  public Number getYData() {
    return this.yData ;
  }

  /**
   * <zh>õdouble͵startXֵ.</zh>
   * <en>Returns the double start x value.</en>
   *
   * @return double <zh>double͵ֵ.</zh>
   * <en>The double value.</en>
   */
  public double getStartXValue() {
    return this.startXData.doubleValue() ;
  }

  /**
   * <zh>õdouble͵endXֵ.</zh>
   * <en>Returns the double end x value.</en>
   *
   * @return double <zh>double͵ֵ.</zh>
   * <en>The double value.</en>
   */
  public double getEndXValue() {
    return this.endXData.doubleValue() ;
  }

  /**
   * <zh>õdouble͵yֵ.</zh>
   * <en>Returns the double y value.</en>
   *
   * @return double <zh>double͵ֵ.</zh>
   * <en>The double value.</en>
   */
  public double getYValue() {
    return this.yData.doubleValue() ;
  }

  /**
   * <zh>ʵComparableӿڵcompareTo().</zh>
   * <en>Implements the compareTo() method of interface Comparable.</en>
   *
   * @param o Object <zh>ҪȽϵĶ.</zh>
   * <en>The Object to be compared.</en>
   * @return int <zh>ֵΪʾС,0ʾ,ʾ.</zh>
   * <en>A negative integer, zero, or a positive integer as this object	is less than, equal to, or
   * greater than the specified object.</en>
   *
   * @throws ClassCastException if the specified object's type prevents it
   *         from being compared to this Object.
   */
  public int compareTo(Object o) {
    if (o == null) {
      return -1 ;
    } else {
      if (!(o instanceof IntervalXNumberYBaseChartData)) {
        throw new ClassCastException("Must be 'IntervalXNumberYBaseChartData'.") ;
      }

      IntervalXNumberYBaseChartData other = (IntervalXNumberYBaseChartData) o ;
      double diff = this.startXData.doubleValue() - other.getStartXValue() ;
      return (diff < 0) ? -1 : ((diff == 0) ? 0 : 1) ;
    }
  }

  /**
   * <zh>رǩַ.</zh>
   * <en>Returns the label text.</en>
   *
   * @return String <zh>ǩַ (Ϊnull).</zh>
   * <en>The label text (never null).</en>
   */
  public String getLabelText() {
    return "(" + formatNumber(this.startXData) + ","
      + formatNumber(this.endXData) + "," + formatNumber(this.yData) + ")" ;
  }

  /**
   * <zh>ͼȵareaĺ.</zh>
   * <en>Returns the function arguments of hotspot area.</en>
   *
   * @return String <zh> (Ϊnull).</zh>
   * <en>The function arguments (maybe null).</en>
   */
  public String getAreaFunctionArguments() {
    return formatNumberForFunction(this.startXData) + ","
      + formatNumberForFunction(this.endXData) + ","
      + formatNumberForFunction(this.yData) ;
  }

  /**
   * <zh>ͼȵareaurl.</zh>
   * <en>Returns the url for hotspot area.</en>
   *
   * @return String <zh>url (Ϊnull).</zh>
   * <en>The url fragment (maybe null).</en>
   */
  public String getAreaURLFragment() {
    return "startXData=" + formatNumber(this.startXData) + "&endXData="
      + formatNumber(this.endXData) + "&yData=" + formatNumber(this.yData) ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof IntervalXNumberYBaseChartData)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    IntervalXNumberYBaseChartData that = (IntervalXNumberYBaseChartData) obj ;
    if (EqualsUtil.equals(this.startXData, that.startXData)) {
      return false ;
    }
    if (EqualsUtil.equals(this.endXData, that.endXData)) {
      return false ;
    }
    if (EqualsUtil.equals(this.yData, that.yData)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = 37 * result + HashcodeUtil.hashCode(this.startXData) ;
    result = 37 * result + HashcodeUtil.hashCode(this.endXData) ;
    result = 37 * result + HashcodeUtil.hashCode(this.yData) ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[startXdata=" + this.startXData.doubleValue()
      + ",endXData=" + this.endXData.doubleValue()
      + ",yData=" + this.yData.doubleValue()
      + "]" ;
  }

}
