package com.ijchart.xychart.format ;

import java.io.Serializable ;
import java.text.DecimalFormat ;
import java.text.NumberFormat ;

import com.ijchart.xychart.data.IntervalXNumberYBaseChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ʽ{@link com.ijchart.xychart.data.IntervalXNumberYBaseChartData}͵.</zh>
 * <en>A formatter class for {@link com.ijchart.xychart.data.IntervalXNumberYBaseChartData}
 * chart data.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class IntervalXNumberYBaseChartDataFormatter extends AbstractChartDataFormatterAdapter implements
  IChartDataFormatter, Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 5591486090077456413L ;

  /**
   * <zh>ʽstartXֵ.</zh>
   * <en>The startX format object.</en>
   */
  private DecimalFormat startXFormat ;

  /**
   * <zh>ʽendXֵ.</zh>
   * <en>The endX format object.</en>
   */
  private DecimalFormat endXFormat ;

  /**
   * <zh>ʽyֵ.</zh>
   * <en>The y format object.</en>
   */
  private DecimalFormat yFormat ;

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param startXMask String <zh>startXֵ (Ϊnull).</zh>
   * <en>The startX format mask (null not permitted).</en>
   * @param endXMask String <zh>endXֵ (Ϊnull).</zh>
   * <en>The endX format mask (null not permitted).</en>
   * @param yMask String <zh>yֵ (Ϊnull).</zh>
   * <en>The y format mask (null permitted).</en>
   */
  public IntervalXNumberYBaseChartDataFormatter(String startXMask, String endXMask, String yMask) {
    if (startXMask == null) {
      throw new IllegalArgumentException("Null 'startXMask' argument.") ;
    }
    if (endXMask == null) {
      throw new IllegalArgumentException("Null 'endXMask' argument.") ;
    }
    if (yMask == null) {
      throw new IllegalArgumentException("Null 'yMask' argument.") ;
    }

    this.startXFormat = (DecimalFormat) NumberFormat.getInstance() ;
    this.startXFormat.applyPattern(startXMask) ;
    this.endXFormat = (DecimalFormat) NumberFormat.getInstance() ;
    this.endXFormat.applyPattern(endXMask) ;

    this.yFormat = (DecimalFormat) NumberFormat.getInstance() ;
    this.yFormat.applyPattern(yMask) ;
  }

  /**
   * <zh>ʽݼĳ{@link com.ijchart.xychart.data.IChartData}Ϊַ.</zh>
   * <en>Formatted an {@link com.ijchart.xychart.data.IChartData} object to a string.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   * @param row int <zh>ݼ,ڵ0.</zh>
   * <en>The dataset row number,must be greater than or equal to 0.</en>
   * @param column int <zh>ݼ,ڵ0.</zh>
   * <en>The dataset column number,must be greater than or equal to 0.</en>
   *
   * @return String <zh>ַ (Ϊnull).</zh>
   * <en>A string (maybe null).</en>
   */
  public String format(IMatrixDataset dataset, int row, int column) {
    if (dataset == null) {
      throw new IllegalArgumentException("Null 'dataset' argument.") ;
    }
    if (row < 0) {
      throw new IllegalArgumentException("Must be 'row'>=0.") ;
    }
    if (column < 0) {
      throw new IllegalArgumentException("Must be 'column'>=0.") ;
    }

    Object obj = dataset.getData(row, column) ;
    if (obj != null && obj instanceof IntervalXNumberYBaseChartData) {
      IntervalXNumberYBaseChartData data = (IntervalXNumberYBaseChartData) obj ;

      StringBuffer buffer = new StringBuffer(30) ;
      buffer.append("(").append(this.startXFormat.format(data.getStartXValue())) ;
      buffer.append(",").append(this.endXFormat.format(data.getEndXValue())) ;
      buffer.append(",").append(this.yFormat.format(data.getYValue())) ;
      buffer.append(")") ;

      return buffer.toString() ;
    }

    return null ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this object for equality with an arbitrary object.</en>
   *
   * @param obj <zh>ҪȽϵһ (Ϊnull).</zh>
   * <en>The object to test against (null permitted).</en>
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof IntervalXNumberYBaseChartDataFormatter)) {
      return false ;
    }

    IntervalXNumberYBaseChartDataFormatter that = (IntervalXNumberYBaseChartDataFormatter) obj ;
    if (!EqualsUtil.equals(this.startXFormat, that.startXFormat)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.endXFormat, that.endXFormat)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.yFormat, that.yFormat)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>¡.</zh>
   * <en>Creates and returns a copy of this object.</en>
   *
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the <code>Cloneable</code> interface. Subclasses
   *               that override the <code>clone</code> method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  protected Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>hashcode.</zh>
   * <en>Returns a hash code.</en>
   *
   * @return A hash code.
   */
  public int hashCode() {
    int hashCode = HashcodeUtil.hashCode(this.startXFormat) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.endXFormat) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.yFormat) ;

    return hashCode ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of this Range.</en>
   *
   * @return A String.
   */
  public String toString() {
    return "[startXFormat=" + this.startXFormat
      + ",endXFormat=" + this.endXFormat
      + ",startYFormat=" + this.yFormat
      + "]" ;
  }

}
