package com.ijchart.xychart.plot ;

import java.awt.Graphics2D ;
import java.awt.Paint ;
import java.awt.Stroke ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Collection ;
import java.util.Collections ;
import java.util.List ;

import com.ijchart.xychart.axis.AbstractAxis ;
import com.ijchart.xychart.axis.AxisSpace ;
import com.ijchart.xychart.axis.IAxis ;
import com.ijchart.xychart.axis.category.CategoryAxis ;
import com.ijchart.xychart.axis.category.CategoryAxis3D ;
import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.axis.value.NumberValueAxis3D ;
import com.ijchart.xychart.chartinfo.AnnotationShapeCollection ;
import com.ijchart.xychart.chartinfo.AxisShapeCollection ;
import com.ijchart.xychart.chartinfo.ChartShapeCollection ;
import com.ijchart.xychart.chartinfo.PlotShapeCollection ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.AxisLocation ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.PlotType ;
import com.ijchart.xychart.constant.RenderDrawOrder ;
import com.ijchart.xychart.event.PlotChangeEvent ;
import com.ijchart.xychart.render.category.AbstractCategoryRender ;
import com.ijchart.xychart.ui.IEffect3D ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.ui.RectangleInsets ;
import com.ijchart.xychart.ui.supplier.IDrawingSupplier ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ʾֱ{@link com.ijchart.xychart.axis.category.CategoryAxis},
 * {@link com.ijchart.xychart.axis.value.AbstractValueAxis}plot.</zh>
 * <en>Used to display the two coordinate axis are (@Link com.ijchart.xychart.axis.category.CategoryAxis),
 * (@Link com.ijchart.xychart.axis.value.AbstractValueAxis) of the plot.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class CategoryPlot extends AbstractAxisPlot implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -8558646343845309287L ;

  /**
   * <zh>ȱʡcategory᷽ʮǷɼ.</zh>
   * <en>A default flag that controls whether or not a category crosshair is drawn.</en>
   */
  public static final boolean DEFAULT_CATEGORY_CROSSHAIR_VISIBLE = false ;

  /**
   * <zh>ȱʡʮǷԶݵ.</zh>
   * <en>A default flag that controls whether or not the crosshair locks onto
   * actual data points.</en>
   */
  public static final boolean DEFAULT_CATEGORY_CROSSHAIR_LOCKED_ON_DATA = true ;

  /**
   * <zh>ڴ洢categoryλõļ.</zh>
   * <en>Storage for the category axis locations.</en>
   */
  private List categoryAxisLocations ;

  /**
   * <zh>render.</zh>
   * <en>Storage for the renders.</en>
   */
  private List renders ;

  /**
   * <zh>ȱʡĹ.</zh>
   * <en>Default constructor.</en>
   */
  public CategoryPlot() {
    this(null, null, null) ;
  }

  /**
   * <zh>һµĶ.</zh>
   * <en>Creates a new plot.</en>
   *
   * @param categoryAxis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The category axis (null permitted).</en>
   * @param valueAxis AbstractValueAxis <zh>value (Ϊnull).</zh>
   * <en>The value axis (null permitted).</en>
   * @param render AbstractCategoryRender <zh> (Ϊnull).</zh>
   * <en>The item render (null permitted).</en>
   *
   */
  public CategoryPlot(CategoryAxis categoryAxis, AbstractValueAxis valueAxis,
                      AbstractCategoryRender render) {
    this.categoryAxisLocations = new ArrayList() ;
    setListObject(this.domainAxes, 0, categoryAxis) ;
    if (categoryAxis != null) {
      categoryAxis.addChangeListener(this) ;
    }

    setListObject(this.valueAxes, 0, valueAxis) ;
    if (valueAxis != null) {
      valueAxis.addChangeListener(this) ;
    }

    this.renders = new ArrayList() ;
    setListObject(this.renders, 0, render) ;
    if (render != null) {
      render.addChangeListener(this) ;
    }

    setCategoryAxisLocation(AxisLocation.BOTTOM_OR_LEFT) ;
    setValueAxisLocation(AxisLocation.BOTTOM_OR_LEFT) ;
    mapValueAxisToRender(0, 0) ;
  }

  /**
   * <zh>categoryĸ.</zh>
   * <en>Returns the number of category axes.</en>
   *
   * @return <zh>ĸ.</zh>
   * <en>The axis count.</en>
   */
  public int getCategoryAxisCount() {
    return this.domainAxes.size() ;
  }

  /**
   * <zh>ص0CategoryAxis.</zh>
   * <en>Returns the category axis for the plot.</en>
   *
   * @return <zh>0CategoryAxis (Ϊnull).</zh>
   * <en>The category axis (null permitted).</en>
   */
  public CategoryAxis getCategoryAxis() {
    return getCategoryAxis(0) ;
  }

  /**
   * <zh>CategoryAxis.</zh>
   * <en>Returns a category axis.</en>
   *
   * @param index <zh>ֵ.</zh>
   * <en>The axis index.</en>
   *
   * @return <zh>CategoryAxis (Ϊnull).</zh>
   * <en>The axis (null possible).</en>
   */
  public CategoryAxis getCategoryAxis(int index) {
    CategoryAxis result = null ;
    if (index < this.domainAxes.size()) {
      result = (CategoryAxis)this.domainAxes.get(index) ;
    }

    return result ;
  }

  /**
   * <zh>plotcategory,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the category axis for the plot and sends a {@link PlotChangeEvent} to
   * all registered listeners.</en>
   *
   * @param axis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The axis (null not permitted).</en>
   */
  public void setCategoryAxis(CategoryAxis axis) {
    setCategoryAxis(0, axis, true) ;
  }

  /**
   * <zh>category.</zh>
   * <en>Returns a category axis.</en>
   *
   * @param renderIndex <zh>renderֵ.</zh>
   * <en>The index of render.</en>
   *
   * @return <zh>category (Ϊnull).</zh>
   * <en>The category axis (null possible).</en>
   */
  public CategoryAxis getCategoryAxisByRenderIndex(int renderIndex) {
    Integer axisIndex = (Integer)this.renderToDomainAxis.get(new Integer(renderIndex)) ;
    if (axisIndex != null) {
      int idx = axisIndex.intValue() ;
      if (idx < this.domainAxes.size()) {
        return (CategoryAxis)this.domainAxes.get(idx) ;
      } else {
        return (CategoryAxis)this.domainAxes.get(this.domainAxes.size() - 1) ;
      }
    }

    if (renderIndex < this.domainAxes.size()) {
      return (CategoryAxis)this.domainAxes.get(renderIndex) ;
    } else {
      return (CategoryAxis)this.domainAxes.get(this.domainAxes.size() - 1) ;
    }
  }

  /**
   * <zh>category,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets a category axis and sends a {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   * @param axis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The axis (null not permitted).</en>
   */
  public void setCategoryAxis(int index, CategoryAxis axis) {
    setCategoryAxis(index, axis, true) ;
  }

  /**
   * <zh>category,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets a category axis and sends a {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   * @param axis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The axis (null not permitted).</en>
   * @param notify boolean <zh>Ƿ֪ͨ.</zh>
   * <en>Whether or not notify listener.</en>
   */
  public void setCategoryAxis(int index, CategoryAxis axis, boolean notify) {
    if (index < 0) {
      throw new IllegalArgumentException("Must be 'index'>=0.") ;
    }
    if (axis == null) {
      throw new IllegalArgumentException("Null 'axis' argument.") ;
    }

    if (index < this.domainAxes.size()) {
      CategoryAxis existing = (CategoryAxis)this.domainAxes.get(index) ;
      if (existing != null) {
        existing.removeChangeListener(this) ;
      }
    }

    setListObject(domainAxes, index, axis) ;
    axis.addChangeListener(this) ;

    if (notify) {
      notifyListeners(new PlotChangeEvent(this)) ;
    }
  }

  /**
   * <zh>plotcategory,ѾڵĶᱻ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the category axes for this plot and sends a {@link PlotChangeEvent}
   * to all registered listeners.</en>
   *
   * @param axes CategoryAxis[] <zh>category (Ϊnull).</zh>
   * <en>The axes (null not permitted).</en>
   */
  public void setCategoryAxes(CategoryAxis[] axes) {
    setCategoryAxes(axes, 0) ;
  }

  /**
   * <zh>plotcategory,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the category axes for this plot and sends a {@link PlotChangeEvent}
   * to all registered listeners.</en>
   *
   * @param axes CategoryAxis[] <zh>category (Ϊnull).</zh>
   * <en>The axes (null not permitted).</en>
   * @param startIndex int <zh>Ŀʼֵ,startIndex֮ǰĶ󲻻ᱻ,ڵ0.</zh>
   * <en>The beginning of the index value, startIndex before the target will not be covered.
   * must be greater than or equal to 0.</en>
   */
  public void setCategoryAxes(CategoryAxis[] axes, int startIndex) {
    if (axes == null) {
      throw new IllegalArgumentException("Null 'axes' argument.") ;
    }
    if (startIndex < 0) {
      throw new IllegalArgumentException("Must be 'startIndex'>=0.") ;
    }

    if (axes.length > 0) {
      for (int i = 0 ; i < axes.length ; i++) {
        setCategoryAxis(startIndex + i, axes[i], false) ;
      }
      notifyListeners(new PlotChangeEvent(this)) ;
    }
  }

  /**
   * <zh>ָcategory.</zh>
   * <en>Returns the index of the specified axis, or -1 if the axis
   * is not assigned to the plot.</en>
   *
   * @param axis <zh>category (Ϊnull).</zh>
   * <en>The axis (null not permitted).</en>
   *
   * @return <zh>ֵ,ûҵ򷵻-1.</zh>
   * <en>The axis index.</en>
   */
  public int getCategoryAxisIndex(CategoryAxis axis) {
    if (axis == null) {
      throw new IllegalArgumentException("Null 'axis' argument.") ;
    }

    return this.domainAxes.indexOf(axis) ;
  }

  /**
   * <zh>ص0render.</zh>
   * <en>Returns a reference to the render for the plot.</en>
   *
   * @return AbstractCategoryRender <zh>0render</zh>
   * <en>The render.</en>
   */
  public AbstractCategoryRender getRender() {
    return getRender(0) ;
  }

  /**
   * <zh>ֵȡrender.</zh>
   * <en>Returns the render at the given index.</en>
   *
   * @param index <zh>ֵ,ڵ0.</zh>
   * <en>The render index (zero-based).</en>
   *
   * @return AbstractCategoryRender <zh>render(Ϊnull).</zh>
   * <en>The render (possibly null).</en>
   */
  public AbstractCategoryRender getRender(int index) {
    if (index < 0) {
      throw new IllegalArgumentException("Must be 'index'>=0.") ;
    }

    AbstractCategoryRender result = null ;
    if (this.renders.size() > index) {
      result = (AbstractCategoryRender)this.renders.get(index) ;
    }

    return result ;
  }

  /**
   * <zh>render,˼ϲ޸.</zh>
   * <en>Returns the collection of renders,and this collection is unmodifiable.</en>
   *
   * @return Collection <zh>render.</zh>
   * <en>The collection of renders.</en>
   */
  public Collection getRenders() {
    return Collections.unmodifiableCollection(this.renders) ;
  }

  /**
   * <zh>õ0render.</zh>
   * <en>Sets the render at index 0 (sometimes referred to as the "primary" render) and sends
   * a {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param render AbstractCategoryRender <zh>render (Ϊnull).</zh>
   * <en>The render (null not permitted.</en>
   */
  public void setRender(AbstractCategoryRender render) {
    setRender(0, render, true) ;
  }

  /**
   * <zh>õ0render,Ҫм{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the render at index 0 (sometimes referred to as the "primary" render) and,
   * if requested, sends a {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param render  AbstractCategoryRender <zh>render (Ϊnull).</zh>
   * <en>The render (null not permitted).</en>
   * @param notify  boolean <zh>Ƿ֪ͨ.</zh>
   * <en>Whether or not notify listeners.</en>
   */
  public void setRender(AbstractCategoryRender render, boolean notify) {
    setRender(0, render, notify) ;
  }

  /**
   * <zh>趨ֵָrender,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the render at the specified index and sends a
   * {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The index (zero-based).</en>
   * @param render AbstractCategoryRender <zh>render (Ϊnull).</zh>
   * <en>The render (null not permitted).</en>
   */
  public void setRender(int index, AbstractCategoryRender render) {
    setRender(index, render, true) ;
  }

  /**
   * <zh>趨ֵָrender,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets a render.  A {@link PlotChangeEvent} is sent to all registered listeners.</en>
   *
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The dataset index (zero-based).</en>
   * @param render AbstractCategoryRender <zh>render (Ϊnull).</zh>
   * <en>The render (null not permitted).</en>
   * @param notify  boolean <zh>Ƿ֪ͨ.</zh>
   * <en>Whether or not notify listeners.</en>
   */
  public void setRender(int index, AbstractCategoryRender render, boolean notify) {
    if (render == null) {
      throw new IllegalArgumentException("Null 'render' argument.") ;
    }

    // stop listening to the existing render...
    if (index < this.renders.size()) {
      AbstractCategoryRender existing = (AbstractCategoryRender)this.renders.get(index) ;
      if (existing != null) {
        existing.removeChangeListener(this) ;
      }
    }

    // register the new render...
    setListObject(renders, index, render) ;
    render.addChangeListener(this) ;

    if (notify) {
      notifyListeners(new PlotChangeEvent(this)) ;
    }
  }

  /**
   * <zh>render,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the renders for this plot and sends a {@link PlotChangeEvent}
   * to all registered listeners.</en>
   *
   * @param renders AbstractCategoryRender[] <zh>render (Ϊnull).</zh>
   * <en>The renders (null not permitted).</en>
   * @param startIndex int <zh>Ŀʼֵ,startIndex֮ǰĶ󲻻ᱻ,ڵ0.</zh>
   * <en>The beginning of the index value, startIndex before the target will not be covered.
   * must be greater than or equal to 0.</en>
   */
  public void setRenders(AbstractCategoryRender[] renders, int startIndex) {
    if (renders == null) {
      throw new IllegalArgumentException("Null 'renders' argument.") ;
    }
    if (startIndex < 0) {
      throw new IllegalArgumentException("Must be 'startIndex'>=0.") ;
    }

    if (renders.length > 0) {
      for (int i = 0 ; i < renders.length ; i++) {
        setRender(startIndex + i, renders[i], false) ;
      }
      notifyListeners(new PlotChangeEvent(this)) ;
    }
  }

  /**
   * <zh>ָrender,ûҵ򷵻-1.</zh>
   * <en>Returns the index of the specified render, or -1 if the
   * render is not assigned to this plot.</en>
   *
   * @param render AbstractCategoryRender <zh>render (Ϊnull).</zh>
   * <en>The render (null permitted).</en>
   *
   * @return int <zh>ֵ.</zh>
   * <en>The render index.</en>
   */
  public int getRenderIndex(AbstractCategoryRender render) {
    return this.renders.indexOf(render) ;
  }

  /**
   * <zh>ֵΪ0categoryλ.</zh>
   * <en>Returns the category axis location for the primary category axis.</en>
   *
   * @return <zh>λ (Ϊnull).</zh>
   * <en>The location (never null).</en>
   */
  public AxisLocation getCategoryAxisLocation() {
    return getCategoryAxisLocation(0) ;
  }

  /**
   * <zh>ָcategoryλ.</zh>
   * <en>Returns the location for a category axis.</en>
   *
   * @param index <zh>ֵ,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   *
   * @return <zh>λ.</zh>
   * <en>The location.</en>
   */
  public AxisLocation getCategoryAxisLocation(int index) {
    if (index < 0) {
      throw new IllegalArgumentException("Must be 'index'>=0.") ;
    }

    AxisLocation result = null ;
    if (index < this.categoryAxisLocations.size()) {
      result = (AxisLocation)this.categoryAxisLocations.get(index) ;
    }
    if (result == null) {
      result = AxisLocation.getOpposite(getCategoryAxisLocation(0)) ;
    }

    return result ;
  }

  /**
   * <zh>categoryλ.</zh>
   * <en>Sets the location of the category axis and, if requested, sends a
   * {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param location <zh>λ (Ϊnull).</zh>
   * <en>The axis location (null not permitted).</en>
   */
  public void setCategoryAxisLocation(AxisLocation location) {
    setCategoryAxisLocation(0, location) ;
  }

  /**
   * <zh>Ϊindexcategoryλ.</zh>
   * <en>Sets the location for a category axis and sends a {@link PlotChangeEvent}
   * to all registered listeners.</en>
   *
   * @param index <zh>,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   * @param location <zh>λ (Ϊnull).</zh>
   * <en>The location (null not permitted).</en>
   */
  public void setCategoryAxisLocation(int index, AxisLocation location) {
    if (index < 0) {
      throw new IllegalArgumentException("Must be 'index'>=0") ;
    }
    if (location == null) {
      throw new IllegalArgumentException("Null 'location' argument.") ;
    }

    setListObject(categoryAxisLocations, index, location) ;
    notifyListeners(new PlotChangeEvent(this)) ;
  }

  /**
   * <zh>0categoryı߽λ.</zh>
   * <en>Returns the category axis edge.  This is derived from the axis location
   * and the chart orientation.</en>
   *
   * @return <zh>߽λ.</zh>
   * <en>The edge (never null).</en>
   */
  public RectangleEdge getCategoryAxisEdge() {
    return getCategoryAxisEdge(0) ;
  }

  /**
   * <zh>ָcategoryıֵ߽.</zh>
   * <en>Returns the edge for a category axis.</en>
   *
   * @param index <zh>ֵ,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   *
   * @return <zh>ֵ.</zh>
   * <en>The edge (never null).</en>
   */
  public RectangleEdge getCategoryAxisEdge(int index) {
    AxisLocation location = getCategoryAxisLocation(index) ;
    if (location == null) {
      location = AxisLocation.BOTTOM_OR_LEFT ; //default value
    }

    return AbstractPlot.resolveDomainAxisLocation(location, getOrientation()) ;
  }

  /**
   * <zh>domainǷ.</zh>
   * <en>Returns true if the plot's domain axis is zoomable, and false otherwise.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDomainZoomable() {
    return false ;
  }

  /**
   * <zh>ͼplot,Ҫʵִ˷.</zh>
   * <en>Draws the plot within the specified area.Subclasses need to provide an
   * implementation of this method, obviously.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param plotArea Rectangle2D <zh>plot (Ϊnull).</zh>
   * <en>The plot draw area (null not permitted).</en>
   * @param chartShapes ChartShapeCollection <zh>chartͼμ (Ϊnull).</zh>
   * <en>Collects information about the chart (null not permitted).</en>
   */
  public void draw(Graphics2D g2, Rectangle2D plotArea, ChartShapeCollection chartShapes) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (plotArea == null) {
      throw new IllegalArgumentException("Null 'plotArea' argument.") ;
    }
    if (chartShapes == null) {
      throw new IllegalArgumentException("Null 'chartShapes' argument.") ;
    }

    // if the plot area is too small, just return...
    if (plotArea.getWidth() <= DEFAULT_MINIMUM_WIDTH_TO_DRAW ||
        plotArea.getHeight() <= DEFAULT_MINIMUM_HEIGHT_TO_DRAW) {
      return ;
    }

    this.plotShapes = new PlotShapeCollection() ;
    this.plotShapes.setDrawArea(plotArea) ;
    chartShapes.addPlotShapes(this.plotShapes) ;

    ChartOrientation orientation = getOrientation() ;

    // adjust the drawing area for the plot insets (if any)...
    RectangleInsets insets = getInsets() ;
    insets.trim(plotArea) ;

    drawBorder(plotArea, this.plotShapes) ;
    drawBackground(plotArea, orientation, this.plotShapes) ;
    drawBackgroundImage(plotArea, this.plotShapes) ;

    //еݼΪ,ʾݼʱϢ
    boolean isAllDatasetEmpty = true ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      AbstractAxis valueAxis = (AbstractAxis)this.valueAxes.get(i) ;
      if (valueAxis != null) {
        if (!valueAxis.getDataset().isEmpty()) {
          isAllDatasetEmpty = false ;
          break ;
        }
      }
    }
    if (isAllDatasetEmpty) {
      drawNoDataMessage(plotArea, this.plotShapes) ;
      return ;
    }

    allocateRenderShapesList(this.renders.size()) ;
    RenderShapeCollection renderShapes = (RenderShapeCollection)this.renderShapesList.get(0) ;
    this.plotShapes.addRenderShape(renderShapes) ;

    dataArea = drawAxes(g2, plotArea, this.plotShapes) ;
    if (dataArea == null) {
      return ;
    }
    if (this instanceof IEffect3D) {
      IEffect3D effect = (IEffect3D)this ;
      double xOffset = effect.getXOffset() ;
      double yOffset = effect.getYOffset() ;
      Rectangle2D clipArea = new Rectangle2D.Double(this.dataArea.getX() - xOffset,
                                                    this.dataArea.getY(),
                                                    this.dataArea.getWidth() + xOffset,
                                                    this.dataArea.getHeight() + yOffset) ;
      renderShapes.setDrawArea(clipArea) ;
    } else {
      renderShapes.setDrawArea(this.dataArea) ;
    }

    //draw grid lines
    drawCategoryGridlines(g2, this.dataArea, renderShapes) ;
    drawValueGridlines(g2, this.dataArea, renderShapes, this.valueAxes.size()) ;

    if (getRenderDrawOrder().equals(RenderDrawOrder.FORWARD)) {
      IDrawingSupplier supplier = getDrawingSupplier() ;
      supplier.reset() ;
      for (int i = 0 ; i < this.renders.size() ; i++) {
        if (this.renders.get(i) == null
            || !(this.renders.get(i) instanceof AbstractCategoryRender)) {
          continue ;
        }

        AbstractCategoryRender render = (AbstractCategoryRender)this.renders.get(i) ;
        render.setDrawingSupplier(supplier) ;

        if (i == 0) { //0renderƱͼƬ,Ĳ
          render.drawBackground(this.dataArea, orientation, renderShapes) ;
          render.drawDrawAreaBorder(this.dataArea, renderShapes) ;
          render.drawBackgroundImage(this.dataArea, renderShapes) ;

          CategoryAxis categoryAxis = getCategoryAxisByRenderIndex(i) ;
          if (categoryAxis != null) {
            render.drawCategoryMarker(categoryAxis, this.dataArea, orientation, renderShapes) ;
          }

          AbstractValueAxis valueAxis = getValueAxisByRenderIndex(i) ;
          RectangleEdge edge = getValueAxisEdge(i) ;
          if (valueAxis != null) {
            render.drawValueMarker(valueAxis, true, edge, this.dataArea, orientation, renderShapes) ;
            render.drawIntervalValueMarker(valueAxis, true, edge, this.dataArea, orientation,
                                           renderShapes) ;
          }

          //ûжӦѰ
          categoryAxis = (categoryAxis == null) ? getNotNullCategoryAxis(i) : categoryAxis ;
          valueAxis = (valueAxis == null) ? getNotNullValueAxis(i) : valueAxis ;
          edge = (edge == null) ? getNotNullValueAxisEdge(i) : edge ;

          if (domainAxes != null && valueAxis != null) {
            renderShapes.setDataset(render.getDataset()) ;
            render.drawAllItems(categoryAxis, valueAxis, edge, this.dataArea, orientation,
                                renderShapes, getForegroundAlpha()) ;

            AnnotationShapeCollection annotationShapes = new AnnotationShapeCollection() ;
            annotationShapes.setDrawArea(this.dataArea) ;
            this.plotShapes.addAnnotationShape(annotationShapes) ;
            render.drawAllAnnotation(categoryAxis, valueAxis, edge, this.dataArea,
                                     orientation, annotationShapes) ;
          }
        } else {
          RenderShapeCollection tmpRenderShapes = (RenderShapeCollection) renderShapesList.get(i) ;
          if (tmpRenderShapes == null) {
            return ;
          }
          tmpRenderShapes.setDrawArea(this.dataArea) ;
          this.plotShapes.addRenderShape(tmpRenderShapes) ;

          CategoryAxis categoryAxis = getCategoryAxisByRenderIndex(i) ;
          if (categoryAxis != null) {
            render.drawCategoryMarker(categoryAxis, this.dataArea, orientation, tmpRenderShapes) ;
          }

          AbstractValueAxis valueAxis = getValueAxisByRenderIndex(i) ;
          RectangleEdge edge = getValueAxisEdge(i) ;
          if (valueAxis != null) {
            render.drawValueMarker(valueAxis, true, edge, this.dataArea, orientation,
                                   tmpRenderShapes) ;
            render.drawIntervalValueMarker(valueAxis, true, edge, this.dataArea, orientation,
                                           tmpRenderShapes) ;
          }

          //ûжӦѰ
          categoryAxis = (categoryAxis == null) ? getNotNullCategoryAxis(i) : categoryAxis ;
          valueAxis = (valueAxis == null) ? getNotNullValueAxis(i) : valueAxis ;
          edge = (edge == null) ? getNotNullValueAxisEdge(i) : edge ;

          if (domainAxes != null && valueAxis != null) {
            tmpRenderShapes.setDataset(render.getDataset()) ;
            render.drawAllItems(categoryAxis, valueAxis, edge, this.dataArea, orientation,
                                tmpRenderShapes, getForegroundAlpha()) ;

            AnnotationShapeCollection annotationShapes = new AnnotationShapeCollection() ;
            annotationShapes.setDrawArea(this.dataArea) ;
            this.plotShapes.addAnnotationShape(annotationShapes) ;
            render.drawAllAnnotation(categoryAxis, valueAxis, edge, this.dataArea,
                                     orientation, annotationShapes) ;
          }
        }

      }
    } else {
      for (int i = this.renders.size() - 1 ; i >= 0 ; i--) {
        if (this.renders.get(i) == null
            || !(this.renders.get(i) instanceof AbstractCategoryRender)) {
          continue ;
        }

        AbstractCategoryRender render = (AbstractCategoryRender)this.renders.get(i) ;
        IDrawingSupplier supplier = getDrawingSupplier() ;
        supplier.reset() ;
        render.setDrawingSupplier(supplier) ;

        if (i == this.renders.size() - 1) { //0renderƱͼƬ,Ĳ
          render.drawBackground(this.dataArea, orientation, renderShapes) ;
          render.drawDrawAreaBorder(this.dataArea, renderShapes) ;
          render.drawBackgroundImage(this.dataArea, renderShapes) ;

          CategoryAxis categoryAxis = getCategoryAxisByRenderIndex(i) ;
          if (categoryAxis != null) {
            render.drawCategoryMarker(categoryAxis, this.dataArea, orientation, renderShapes) ;
          }

          AbstractValueAxis valueAxis = getValueAxisByRenderIndex(i) ;
          RectangleEdge edge = getValueAxisEdge(i) ;
          if (valueAxis != null) {
            render.drawValueMarker(valueAxis, true, edge, this.dataArea, orientation, renderShapes) ;
            render.drawIntervalValueMarker(valueAxis, true, edge, this.dataArea, orientation,
                                           renderShapes) ;
          }

          //ûжӦѰ
          categoryAxis = (categoryAxis == null) ? getNotNullCategoryAxis(i) : categoryAxis ;
          valueAxis = (valueAxis == null) ? getNotNullValueAxis(i) : valueAxis ;
          edge = (edge == null) ? getNotNullValueAxisEdge(i) : edge ;

          if (domainAxes != null && valueAxis != null) {
            renderShapes.setDataset(render.getDataset()) ;
            render.drawAllItems(categoryAxis, valueAxis, edge, this.dataArea, orientation,
                                renderShapes,
                                getForegroundAlpha()) ;

            AnnotationShapeCollection annotationShapes = new AnnotationShapeCollection() ;
            annotationShapes.setDrawArea(this.dataArea) ;
            this.plotShapes.addAnnotationShape(annotationShapes) ;
            render.drawAllAnnotation(categoryAxis, valueAxis, edge, this.dataArea,
                                     orientation, annotationShapes) ;
          }
        } else {
          RenderShapeCollection tmpRenderShapes = (RenderShapeCollection) renderShapesList.get(i) ;
          if (tmpRenderShapes == null) {
            return ;
          }
          tmpRenderShapes.setDrawArea(this.dataArea) ;
          this.plotShapes.addRenderShape(tmpRenderShapes) ;

          CategoryAxis categoryAxis = getCategoryAxisByRenderIndex(i) ;
          if (categoryAxis != null) {
            render.drawCategoryMarker(categoryAxis, this.dataArea, orientation, tmpRenderShapes) ;
          }

          AbstractValueAxis valueAxis = getValueAxisByRenderIndex(i) ;
          RectangleEdge edge = getValueAxisEdge(i) ;
          if (valueAxis != null) {
            render.drawValueMarker(valueAxis, true, edge, dataArea, orientation, tmpRenderShapes) ;
            render.drawIntervalValueMarker(valueAxis, true, edge, dataArea, orientation,
                                           tmpRenderShapes) ;
          }

          //ûжӦѰ
          categoryAxis = (categoryAxis == null) ? getNotNullCategoryAxis(i) : categoryAxis ;
          valueAxis = (valueAxis == null) ? getNotNullValueAxis(i) : valueAxis ;
          edge = (edge == null) ? getNotNullValueAxisEdge(i) : edge ;

          if (domainAxes != null && valueAxis != null) {
            tmpRenderShapes.setDataset(render.getDataset()) ;
            render.drawAllItems(categoryAxis, valueAxis, edge, this.dataArea, orientation,
                                tmpRenderShapes, getForegroundAlpha()) ;

            AnnotationShapeCollection annotationShapes = new AnnotationShapeCollection() ;
            annotationShapes.setDrawArea(this.dataArea) ;
            this.plotShapes.addAnnotationShape(annotationShapes) ;
            render.drawAllAnnotation(categoryAxis, valueAxis, edge, this.dataArea,
                                     orientation, annotationShapes) ;
          }
        }
      }
    }

    translateMarkerShapes() ;
  }

  /**
   * <zh>е.</zh>
   * <en>A utility method for drawing the plot's axes.</en>
   *
   * @param g2 <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param plotArea <zh>plot (Ϊnull).</zh>
   * <en>The plot draw area (null not permitted).</en>
   * @param shapeCollection PlotShapeCollection <zh>plotͼμ (Ϊnull).</zh>
   * <en>Collects information about the plot (null not permitted).</en>
   *
   * @return Rectangle2D <zh>.</zh>
   * <en>The data area.</en>
   */
  protected Rectangle2D drawAxes(Graphics2D g2, Rectangle2D plotArea,
                                 PlotShapeCollection shapeCollection) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (plotArea == null) {
      throw new IllegalArgumentException("Null 'plotArea' argument.") ;
    }
    if (shapeCollection == null) {
      throw new IllegalArgumentException("Null 'shapeCollection' argument.") ;
    }

    AxisSpace axisSpace = new AxisSpace() ; //Ԥȼspace
    Rectangle2D clonePlotArea = (Rectangle2D) plotArea.clone() ;

    int categoryAxesCount = 0 ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
      if (xAxis != null) {
        categoryAxesCount++ ;
      }
    }

    //calculate data area size
    boolean firstLeftAxis = false ;
    double[] categoryAxisWidthArr = new double[this.domainAxes.size()] ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
      if (xAxis != null) {
        RectangleEdge edge = getCategoryAxisEdge(i) ;
        if (categoryAxesCount <= 1 && !firstLeftAxis
            && (xAxis instanceof IEffect3D) && edge.equals(RectangleEdge.LEFT)) { //3D͵category
          edge = RectangleEdge.RIGHT ;
          firstLeftAxis = true ;
        }
        double fixedSpace = xAxis.getFixedAxisSpace() ;
        double axisWidth = 0 ;
        if (fixedSpace > 0) {
          axisWidth = fixedSpace ;
        } else {
          axisWidth = xAxis.calculateAxisSpace(g2, plotArea, clonePlotArea, edge) ;
        }

        axisSpace.add(axisWidth, edge) ;
        categoryAxisWidthArr[i] = axisWidth ;
        edge.trim(clonePlotArea, axisWidth) ;
      }
    }

    double[] valueAxisWidthArr = new double[this.valueAxes.size()] ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(i) ;
        double fixedSpace = yAxis.getFixedAxisSpace() ;
        double axisWidth = 0 ;
        if (fixedSpace > 0) {
          axisWidth = fixedSpace ;
        } else {
          axisWidth = yAxis.calculateAxisSpace(g2, plotArea, clonePlotArea, edge) ;
        }

        axisSpace.add(axisWidth, edge) ;
        valueAxisWidthArr[i] = axisWidth ;
        edge.trim(clonePlotArea, axisWidth) ;
      }
    }

    RectangleInsets atLeastAxisOffset = getAxisOffset().ensureAtLeast(1) ; //ÿ߽Ϊ1
    atLeastAxisOffset.trim(clonePlotArea) ; //trim data area
    double axisOffsetWidth = atLeastAxisOffset.getLeft() + atLeastAxisOffset.getRight() ;
    double axisOffsetHeight = atLeastAxisOffset.getTop() + atLeastAxisOffset.getBottom() ;

    //Խ
    axisSpace.setTop(Math.min(plotArea.getHeight(), axisSpace.getTop())) ;
    axisSpace.setBottom(Math.min(plotArea.getHeight(), axisSpace.getBottom())) ;
    axisSpace.setLeft(Math.min(plotArea.getWidth(), axisSpace.getLeft())) ;
    axisSpace.setRight(Math.min(plotArea.getWidth(), axisSpace.getRight())) ;

    /////////////////////////// draw top axes ///////////////////////////////
    //draw top category axes
    double topSize = 0.0 ;
    for (int m = this.domainAxes.size() - 1 ; m >= 0 ; m--) {
      IAxis xAxis = (IAxis)this.domainAxes.get(m) ;
      if (xAxis != null) {
        RectangleEdge edge = getCategoryAxisEdge(m) ;
        if (edge.equals(RectangleEdge.TOP)) {
          if (topSize + categoryAxisWidthArr[m] > plotArea.getHeight()) { //Խʾ
            break ;
          }

          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + axisSpace.getLeft(),
                                                        plotArea.getY() + topSize,
                                                        clonePlotArea.getWidth() + axisOffsetWidth,
                                                        categoryAxisWidthArr[m]) ;
          AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
          xAxisShapes.setEdge(edge) ;
          xAxisShapes.setDrawArea(axisArea) ;

          topSize += categoryAxisWidthArr[m] ;
          xAxis.draw(g2, plotArea, axisArea, edge, xAxisShapes, false) ;
          shapeCollection.addXAxisShape(xAxisShapes) ;
        }
      }
    }
    //draw top value axes
    for (int n = this.valueAxes.size() - 1 ; n >= 0 ; n--) {
      IAxis yAxis = (IAxis)this.valueAxes.get(n) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(n) ;
        if (edge.equals(RectangleEdge.TOP)) {
          if (topSize + valueAxisWidthArr[n] > plotArea.getHeight()) { //Խʾ
            break ;
          }

          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + axisSpace.getLeft(),
                                                        plotArea.getY() + topSize,
                                                        clonePlotArea.getWidth() + axisOffsetWidth,
                                                        valueAxisWidthArr[n]) ;
          AxisShapeCollection yAxisShapes = new AxisShapeCollection() ;
          yAxisShapes.setEdge(edge) ;
          yAxisShapes.setDrawArea(axisArea) ;

          topSize += valueAxisWidthArr[n] ;
          yAxis.draw(g2, plotArea, axisArea, edge, yAxisShapes, false) ;
          if (yAxis instanceof AbstractValueAxis) {
            int renderIndex = getRenderIndexForValueAxis(n) ;
            RenderShapeCollection renderShapes = getRenderShapeCollection(renderIndex) ;
            if (renderShapes != null) {
              ((AbstractValueAxis) yAxis).drawValueLine(g2, clonePlotArea, edge, renderShapes) ;
            }
          }
          shapeCollection.addYAxisShape(yAxisShapes) ;
        }
      }
    }

    /////////////////////////// draw bottom axes ///////////////////////////////
    //draw bottom category axes
    //һbottom 3d axisΪ2dʽ
    int firstBottom3DIndex = -1 ;
    int bottom3DAxisCount = 0 ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
      if (xAxis != null && xAxis instanceof CategoryAxis3D) {
        RectangleEdge edge = getCategoryAxisEdge(i) ;
        if (edge.equals(RectangleEdge.BOTTOM)) {
          if (firstBottom3DIndex < 0) {
            firstBottom3DIndex = i ;
          }
          bottom3DAxisCount++ ;
        }
      }
    }

    if (bottom3DAxisCount > 1) {
      for (int i = 0 ; i < this.domainAxes.size() ; i++) {
        IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
        if (xAxis != null && xAxis instanceof CategoryAxis3D) {
          RectangleEdge edge = getCategoryAxisEdge(i) ;
          if (edge.equals(RectangleEdge.BOTTOM)) {
            if (i != firstBottom3DIndex) {
              ((CategoryAxis3D) xAxis).setForce2D(true) ;
            }
          }
        }
      }
    }

    double bottomSize = 0.0 ;
    for (int m = this.domainAxes.size() - 1 ; m >= 0 ; m--) {
      IAxis xAxis = (IAxis)this.domainAxes.get(m) ;
      if (xAxis != null) {
        RectangleEdge edge = getCategoryAxisEdge(m) ;
        if (edge.equals(RectangleEdge.BOTTOM)) {
          if (topSize + bottomSize + categoryAxisWidthArr[m] > plotArea.getHeight()) { //Խʾ
            break ;
          }

          bottomSize += categoryAxisWidthArr[m] ;
          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + axisSpace.getLeft(),
                                                        plotArea.getMaxY() - bottomSize,
                                                        clonePlotArea.getWidth() + axisOffsetWidth,
                                                        categoryAxisWidthArr[m]) ;
          AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
          xAxisShapes.setEdge(edge) ;
          xAxisShapes.setDrawArea(axisArea) ;

          xAxis.draw(g2, plotArea, axisArea, edge, xAxisShapes, false) ;
          shapeCollection.addXAxisShape(xAxisShapes) ;
        }
      }
    }

    //draw bottom value axes
    //һleft 3d axisΪ2dʽ
    int firstBottomValue3DIndex = -1 ;
    int leftBottom3DAxisCount = 0 ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
      if (yAxis != null && yAxis instanceof NumberValueAxis3D) {
        RectangleEdge edge = getValueAxisEdge(i) ;
        if (edge.equals(RectangleEdge.BOTTOM)) {
          if (firstBottomValue3DIndex < 0) {
            firstBottomValue3DIndex = i ;
          }
          leftBottom3DAxisCount++ ;
        }
      }
    }

    if (leftBottom3DAxisCount > 1) {
      for (int i = 0 ; i < this.valueAxes.size() ; i++) {
        IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
        if (yAxis != null && yAxis instanceof NumberValueAxis3D) {
          RectangleEdge edge = getValueAxisEdge(i) ;
          if (edge.equals(RectangleEdge.BOTTOM)) {
            if (i != firstBottom3DIndex) {
              ((NumberValueAxis3D) yAxis).setForce2D(true) ;
            }
          }
        }
      }
    }

    for (int n = this.valueAxes.size() - 1 ; n >= 0 ; n--) {
      IAxis yAxis = (IAxis)this.valueAxes.get(n) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(n) ;
        if (edge.equals(RectangleEdge.BOTTOM)) {
          if (topSize + bottomSize + valueAxisWidthArr[n] > plotArea.getHeight()) { //Խʾ
            break ;
          }

          bottomSize += valueAxisWidthArr[n] ;
          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + axisSpace.getLeft(),
                                                        plotArea.getMaxY() - bottomSize,
                                                        clonePlotArea.getWidth() + axisOffsetWidth,
                                                        valueAxisWidthArr[n]) ;
          AxisShapeCollection yAxisShapes = new AxisShapeCollection() ;
          yAxisShapes.setEdge(edge) ;
          yAxisShapes.setDrawArea(axisArea) ;

          yAxis.draw(g2, plotArea, axisArea, edge, yAxisShapes, false) ;
          if (yAxis instanceof AbstractValueAxis) {
            int renderIndex = getRenderIndexForValueAxis(n) ;
            RenderShapeCollection renderShapes = getRenderShapeCollection(renderIndex) ;
            if (renderShapes != null) {
              ((AbstractValueAxis) yAxis).drawValueLine(g2, clonePlotArea, edge, renderShapes) ;
            }
          }
          shapeCollection.addYAxisShape(yAxisShapes) ;
        }
      }
    }

    /////////////////////////// draw left axes ///////////////////////////////
    //draw left category axes
    //һbottom 3d axisΪ2dʽ
    int firstLeftCategory3DIndex = -1 ;
    int leftCategory3DAxisCount = 0 ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
      if (xAxis != null && xAxis instanceof CategoryAxis3D) {
        RectangleEdge edge = getCategoryAxisEdge(i) ;
        if (edge.equals(RectangleEdge.LEFT)) {
          if (firstLeftCategory3DIndex < 0) {
            firstLeftCategory3DIndex = i ;
          }
          leftCategory3DAxisCount++ ;
        }
      }
    }

    if (leftCategory3DAxisCount > 1) {
      for (int i = 0 ; i < this.domainAxes.size() ; i++) {
        IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
        if (xAxis != null && xAxis instanceof CategoryAxis3D) {
          RectangleEdge edge = getCategoryAxisEdge(i) ;
          if (edge.equals(RectangleEdge.LEFT)) {
            if (i != firstLeftCategory3DIndex) {
              ((CategoryAxis3D) xAxis).setForce2D(true) ;
            }
          }
        }
      }
    }

    double leftSize = 0.0 ;
    double rightSize = 0.0 ;
    boolean leftToRight = false ;
    for (int m = this.domainAxes.size() - 1 ; m >= 0 ; m--) {
      IAxis xAxis = (IAxis)this.domainAxes.get(m) ;
      if (xAxis != null) {
        RectangleEdge edge = getCategoryAxisEdge(m) ;
        if (edge.equals(RectangleEdge.LEFT)) {
          if (xAxis instanceof IEffect3D && !leftToRight && categoryAxesCount <= 1) {
            if (rightSize + leftSize + categoryAxisWidthArr[m] > plotArea.getWidth()) { //Խʾ
              break ;
            }

            rightSize += categoryAxisWidthArr[m] ;
            IEffect3D effect = (IEffect3D) xAxis ;
            Rectangle2D axisArea = new Rectangle2D.
                                   Double(plotArea.getMaxX() - rightSize + effect.getXOffset(),
                                          plotArea.getY() + axisSpace.getTop() - effect.getYOffset(),
                                          categoryAxisWidthArr[m],
                                          clonePlotArea.getHeight() + axisOffsetHeight) ;
            AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
            xAxisShapes.setEdge(edge) ;
            xAxisShapes.setDrawArea(axisArea) ;

            xAxis.draw(g2, plotArea, axisArea, RectangleEdge.RIGHT, xAxisShapes, false) ;
            shapeCollection.addXAxisShape(xAxisShapes) ;
            leftToRight = true ;
          } else {
            if (leftSize + categoryAxisWidthArr[m] > plotArea.getWidth()) { //Խʾ
              break ;
            }

            Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + leftSize,
                                                          plotArea.getY() + axisSpace.getTop(),
                                                          categoryAxisWidthArr[m],
                                                          clonePlotArea.getHeight()
                                                          + axisOffsetHeight) ;
            AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
            xAxisShapes.setEdge(edge) ;
            xAxisShapes.setDrawArea(axisArea) ;

            leftSize += categoryAxisWidthArr[m] ;
            xAxis.draw(g2, plotArea, axisArea, edge, xAxisShapes, false) ;

            shapeCollection.addXAxisShape(xAxisShapes) ;
          }
        }
      }
    }

    //draw left value axes
    //һleft 3d axisΪ2dʽ
    int firstLeftValue3DIndex = -1 ;
    int leftValue3DAxisCount = 0 ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
      if (yAxis != null && yAxis instanceof NumberValueAxis3D) {
        RectangleEdge edge = getValueAxisEdge(i) ;
        if (edge.equals(RectangleEdge.LEFT)) {
          if (firstLeftValue3DIndex < 0) {
            firstLeftValue3DIndex = i ;
          }
          leftValue3DAxisCount++ ;
        }
      }
    }

    if (leftValue3DAxisCount > 1) {
      for (int i = 0 ; i < this.valueAxes.size() ; i++) {
        IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
        if (yAxis != null && yAxis instanceof NumberValueAxis3D) {
          RectangleEdge edge = getValueAxisEdge(i) ;
          if (edge.equals(RectangleEdge.LEFT)) {
            if (i != leftValue3DAxisCount) {
              ((NumberValueAxis3D) yAxis).setForce2D(true) ;
            }
          }
        }
      }
    }

    for (int n = this.valueAxes.size() - 1 ; n >= 0 ; n--) {
      IAxis yAxis = (IAxis)this.valueAxes.get(n) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(n) ;
        if (edge.equals(RectangleEdge.LEFT)) {
          if (leftSize + valueAxisWidthArr[n] > plotArea.getWidth()) { //Խʾ
            break ;
          }

          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + leftSize,
                                                        plotArea.getY() + axisSpace.getTop(),
                                                        valueAxisWidthArr[n],
                                                        clonePlotArea.getHeight()
                                                        + axisOffsetHeight) ;
          AxisShapeCollection yAxisShapes = new AxisShapeCollection() ;
          yAxisShapes.setEdge(edge) ;
          yAxisShapes.setDrawArea(axisArea) ;

          leftSize += valueAxisWidthArr[n] ;
          yAxis.draw(g2, plotArea, axisArea, edge, yAxisShapes, false) ;
          if (yAxis instanceof AbstractValueAxis) {
            int renderIndex = getRenderIndexForValueAxis(n) ;
            RenderShapeCollection renderShapes = getRenderShapeCollection(renderIndex) ;
            if (renderShapes != null) {
              ((AbstractValueAxis) yAxis).drawValueLine(g2, clonePlotArea, edge, renderShapes) ;
            }
          }
          shapeCollection.addYAxisShape(yAxisShapes) ;
        }
      }
    }

    /////////////////////////// draw right axes ///////////////////////////////
    //draw right category axes
    for (int m = this.domainAxes.size() - 1 ; m >= 0 ; m--) {
      IAxis xAxis = (IAxis)this.domainAxes.get(m) ;
      if (xAxis != null) {
        RectangleEdge edge = getCategoryAxisEdge(m) ;
        if (edge.equals(RectangleEdge.RIGHT)) {
          if (rightSize + leftSize + categoryAxisWidthArr[m] > plotArea.getWidth()) { //Խʾ
            break ;
          }

          rightSize += categoryAxisWidthArr[m] ;
          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getMaxX() - rightSize,
                                                        plotArea.getY() + axisSpace.getTop(),
                                                        categoryAxisWidthArr[m],
                                                        clonePlotArea.getHeight()
                                                        + axisOffsetHeight) ;
          AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
          xAxisShapes.setEdge(edge) ;
          xAxisShapes.setDrawArea(axisArea) ;

          xAxis.draw(g2, plotArea, axisArea, edge, xAxisShapes, false) ;
          shapeCollection.addXAxisShape(xAxisShapes) ;
        }
      }
    }
    //draw right value axes
    for (int n = this.valueAxes.size() - 1 ; n >= 0 ; n--) {
      IAxis yAxis = (IAxis)this.valueAxes.get(n) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(n) ;
        if (edge.equals(RectangleEdge.RIGHT)) {
          if (rightSize + leftSize + valueAxisWidthArr[n] > plotArea.getWidth()) { //Խʾ
            break ;
          }

          rightSize += valueAxisWidthArr[n] ;
          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getMaxX() - rightSize,
                                                        plotArea.getY() + axisSpace.getTop(),
                                                        valueAxisWidthArr[n],
                                                        clonePlotArea.getHeight()
                                                        + axisOffsetHeight) ;
          AxisShapeCollection yAxisShapes = new AxisShapeCollection() ;
          yAxisShapes.setEdge(edge) ;
          yAxisShapes.setDrawArea(axisArea) ;

          yAxis.draw(g2, plotArea, axisArea, edge, yAxisShapes, false) ;
          if (yAxis instanceof AbstractValueAxis) {
            int renderIndex = getRenderIndexForValueAxis(n) ;
            RenderShapeCollection renderShapes = getRenderShapeCollection(renderIndex) ;
            if (renderShapes != null) {
              ((AbstractValueAxis) yAxis).drawValueLine(g2, clonePlotArea, edge, renderShapes) ;
            }
          }
          shapeCollection.addYAxisShape(yAxisShapes) ;
        }
      }
    }

    return clonePlotArea ;
  }

  /**
   * <zh>ѰҲΪnullCategory,ѰҷΪӵǰֵʼǰ,ҲѰ.</zh>
   * <en>Search the category axis at the specified position,first search previous,if not found
   *  search next.</en>
   *
   * @param index int <zh>ֵ.</zh>
   * <en>The index.</en>
   * @return CategoryAxis <zh>Category.</zh>
   * <en>The category axis.</en>
   */
  protected CategoryAxis getNotNullCategoryAxis(int index) {
    Object result = null ;

    int size = this.domainAxes.size() ;
    int start = index - 1 ;
    for (int i = start ; i >= 0 ; i--) {
      if (i < size) {
        Object obj = this.domainAxes.get(i) ;
        if (obj != null) {
          result = obj ;
          break ;
        }
      }
    }

    start = index + 1 ;
    for (int i = start ; i < size ; i++) {
      if (i >= 0) {
        Object obj = this.domainAxes.get(i) ;
        if (obj != null) {
          result = obj ;
          break ;
        }
      }
    }

    return (CategoryAxis) result ;
  }

  /**
   * <zh>ѰҲΪnullValue,ѰҷΪӵǰֵʼǰ,ҲѰ.</zh>
   * <en>Search the value axis at the specified position,first search previous,if not found
   *  search next.</en>
   *
   * @param index int <zh>ֵ.</zh>
   * <en>The index.</en>
   * @return AbstractValueAxis <zh>value.</zh>
   * <en>The value axis.</en>
   */
  protected AbstractValueAxis getNotNullValueAxis(int index) {
    Object result = null ;

    int size = this.valueAxes.size() ;
    int start = index - 1 ;
    for (int i = start ; i >= 0 ; i--) {
      if (i < size) {
        Object obj = this.valueAxes.get(i) ;
        if (obj != null) {
          result = obj ;
          break ;
        }
      }
    }

    start = index + 1 ;
    for (int i = start ; i < size ; i++) {
      if (i >= 0) {
        Object obj = this.valueAxes.get(i) ;
        if (obj != null) {
          result = obj ;
          break ;
        }
      }
    }

    return (AbstractValueAxis) result ;
  }

  /**
   * <zh>ѰҲΪnullvalueı߽,ѰҷΪӵǰֵʼǰ,
   * ҲѰ.</zh>
   * <en>Search the value axis edge at the specified position,first search previous,if not found
   *  search next.</en>
   *
   * @param index int <zh>ֵ.</zh>
   * <en>The index.</en>
   * @return RectangleEdge <zh>value߽.</zh>
   * <en>The edge of value axis.</en>
   */
  protected RectangleEdge getNotNullValueAxisEdge(int index) {
    int find = -1 ;

    int size = this.valueAxes.size() ;
    int start = index - 1 ;
    for (int i = start ; i >= 0 ; i--) {
      if (i < size) {
        if (this.valueAxes.get(i) != null) {
          find = i ;
          break ;
        }
      }
    }

    start = index + 1 ;
    for (int i = start ; i < size ; i++) {
      if (i >= 0) {
        if (this.valueAxes.get(i) != null) {
          find = i ;
          break ;
        }
      }
    }

    RectangleEdge result = null ;
    if (find >= 0) {
      result = getValueAxisEdge(find) ;
    }

    return result ;
  }

  /**
   * <zh>Ƶ0category᷽.</zh>
   * <en>Draws the first category axis gridlines for the plot.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param dataArea Rectangle2D <zh>ͼdata() (Ϊnull).</zh>
   * <en>The area within which the data should be drawn (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   */
  private void drawCategoryGridlines(Graphics2D g2, Rectangle2D dataArea,
                                     RenderShapeCollection renderShapes) {
    if (isDomainGridlineVisible() && this.domainAxes.size() > 0) {
      RectangleEdge axisEdge = getCategoryAxisEdge() ;
      Stroke gridStroke = getDomainGridlineStroke() ;
      Paint gridPaint = getDomainGridlinePaint() ;

      AbstractAxis axis = (AbstractAxis)this.domainAxes.get(0) ;
      if (axis != null) {
        axis.drawGridLine(g2, dataArea, axisEdge, renderShapes, gridStroke, gridPaint) ;
      }
    }
  }

  /**
   * <zh>plotַ.</zh>
   * <en>Returns a short string describing the plot type.</en>
   *
   * @return <zh>plot.</zh>
   * <en>The plot type.</en>
   */
  public PlotType getPlotType() {
    return PlotType.PLOT_TYPE_CATEGORY ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests the plot for equality with an arbitrary object.</en>
   *
   * @param obj <zh>ҪȽϵĶ (Ϊnull).</zh>
   * <en>The object to test against (null permitted).</en>
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof CategoryPlot)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    CategoryPlot that = (CategoryPlot) obj ;
    if (!EqualsUtil.equals(this.renders, that.renders)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.categoryAxisLocations, that.categoryAxisLocations)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcode.</zh>
   * <en>Returns a hash code for this object.</en>
   *
   * @return <zh>hashcode.</zh>
   * <en>A hash code.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = 29 * result + HashcodeUtil.hashCode(this.renders) ;
    result = 29 * result + HashcodeUtil.hashCode(this.categoryAxisLocations) ;

    return result ;
  }

  /**
   * <zh>¡.</zh>
   * <en>Returns a clone of the plot.</en>
   *
   * @return A clone.
   *
   * @throws CloneNotSupportedException  if the cloning is not supported.
   */
  public Object clone() throws CloneNotSupportedException {
    CategoryPlot clone = (CategoryPlot)super.clone() ;

    clone.domainAxes = new ArrayList() ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      CategoryAxis xAxis = (CategoryAxis)this.domainAxes.get(i) ;
      if (xAxis != null) {
        xAxis.removeAllChangeListener() ;
        CategoryAxis clonedAxis = (CategoryAxis) xAxis.clone() ;
        clone.setCategoryAxis(i, clonedAxis) ;
      }
    }

    clone.valueAxes = new ArrayList() ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      AbstractValueAxis yAxis = (AbstractValueAxis)this.valueAxes.get(i) ;
      if (yAxis != null) {
        yAxis.removeAllChangeListener() ;
        AbstractValueAxis clonedAxis = (AbstractValueAxis) yAxis.clone() ;
        clone.setValueAxis(i, clonedAxis) ;
      }
    }

    return clone ;
  }

  /**
   * <zh>toString().</zh>
   * <en>Override toString method.</en>
   *
   * @return String
   */
  public String toString() {
    return "CategoryPlot" ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;

    int renderCount = this.renders.size() ;
    for (int i = 0 ; i < renderCount ; i++) {
      AbstractCategoryRender render = (AbstractCategoryRender)this.renders.get(i) ;
      if (render != null) {
        render.removeChangeListener(this) ;
        render.addChangeListener(this) ;
      }
    }
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
  }

}
