package com.ijchart.xychart.range ;

import java.io.Serializable ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ʾһݷΧ,һСһֵ.</zh>
 * <en>Represents an immutable range of values.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public strictfp class NumberRange implements Serializable, Cloneable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 3112994441705322924L ;

  /**
   * <zh>lowerֵupperֵΪ0NumberRange.</zh>
   * <en>The NumberRange object with lower and upper value both 0.</en>
   */
  public static final NumberRange RANGE_ZERO = new NumberRange(0, 0) ;

  /**
   * <zh>Сֵ.</zh>
   * <en>The lower bound of the range.</en>
   */
  private double lower ;

  /**
   * <zh>ֵ.</zh>
   * <en>The upper bound of the range.</en>
   */
  private double upper ;

  /**
   * <zh>һСһֵ,lower<=upper.</zh>
   * <en>Creates a new range.</en>
   *
   * @param lower <zh>Сֵ,<=upper.</zh>
   * <en>The lower bound (must be <= upper bound).</en>
   * @param upper <zh>Сֵ,>=lower.</zh>
   * <en>The upper bound (must be >= lower bound).</en>
   */
  public NumberRange(double lower, double upper) {
    if (lower > upper) {
      String msg = "NumberRange(double, double): require lower (" + lower
                   + ") <= upper (" + upper + ")." ;
      throw new IllegalArgumentException(msg) ;
    }

    this.lower = lower ;
    this.upper = upper ;
  }

  /**
   * <zh>Сֵ.</zh>
   * <en>Returns the lower bound for the range.</en>
   *
   * @return <zh>Сֵ.</zh>
   * <en>The lower bound.</en>
   */
  public double getLowerBound() {
    return this.lower ;
  }

  /**
   * <zh>شֵ.</zh>
   * <en>Returns the upper bound for the range.</en>
   *
   * @return <zh>ֵ.</zh>
   * <en>The upper bound.</en>
   */
  public double getUpperBound() {
    return this.upper ;
  }

  /**
   * <zh>ݷΧĳ.</zh>
   * <en>Returns the length of the range.</en>
   *
   * @return <zh>.</zh>
   * <en>The length.</en>
   */
  public double getLength() {
    return this.upper - this.lower ;
  }

  /**
   * <zh>мֵ,(lower+upper)/2.</zh>
   * <en>Returns the central value for the range.</en>
   *
   * @return <zh>мֵ.</zh>
   * <en>The central value.</en>
   */
  public double getCentralValue() {
    return (this.lower + this.upper) / 2.0 ;
  }

  /**
   * <zh>жϸֵǷڴ˷Χ֮.</zh>
   * <en>Returns true if the range contains the specified value and false otherwise.</en>
   *
   * @param value double <zh>ҪȽϵֵ.</zh>
   * <en>The value.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean contains(double value) {
    return (value >= this.lower && value <= this.upper) ;
  }

  /**
   * <zh>ӻСݷΧ.</zh>
   * <en>Increases or decreases the range by the specified percentage about the specified anchor
   * value.
   * <P>
   * To double the length of the axis range, use 200% (2.0).
   * To halve the length of the axis range, use 50% (0.5).</en>
   *
   * @param percent double <zh>.</zh>
   * <en>The resize factor.</en>
   *
   * @return NumberRange <zh>µݷΧ,percent<=0,null.</zh>
   * <en>Returns the new NumberRange,if percent<=0 return null.</en>
   */
  public NumberRange resizeRange(double percent) {
    return resizeRange(percent, getCentralValue()) ;
  }

  /**
   * <zh>ӻСݷΧ.</zh>
   * <en>Increases or decreases the range by the specified percentage about the specified anchor
   * value.
   * <P>
   * To double the length of the axis range, use 200% (2.0).
   * To halve the length of the axis range, use 50% (0.5).</en>
   *
   * @param percent double <zh>.</zh>
   * <en>The resize factor.</en>
   * @param anchorValue double <zh>µеֵ.</zh>
   * <en>the new central value after the resize.</en>
   *
   * @return NumberRange <zh>µݷΧ,percent<=0,null.</zh>
   * <en>Returns the new NumberRange,if percent<=0 return null.</en>
   */
  public NumberRange resizeRange(double percent, double anchorValue) {
    if (percent > 0.0) {
      double halfLength = getLength() * percent / 2 ;
      return new NumberRange(anchorValue - halfLength, anchorValue + halfLength) ;
    } else {
      return null ;
    }
  }

  /**
   * <zh>ӻСݷΧ.</zh>
   * <en>Increases or decreases the range by the specified percentage about the specified anchor
   * value.
   * <P>
   * To double the length of the axis range, use 200% (2.0).
   * To halve the length of the axis range, use 50% (0.5).</en>
   *
   * @param percent double <zh>.</zh>
   * <en>The resize factor.</en>
   * @param anchorValue double <zh>µеֵ.</zh>
   * <en>the new central value after the resize.</en>
   *
   * @return NumberRange <zh>µݷΧ,percent<=0,null.</zh>
   * <en>The new NumberRange,if percent<=0 return null.</en>
   */
  public NumberRange resizeRange2(double percent, double anchorValue) {
    if (percent > 0.0) {
      double left = anchorValue - getLowerBound() ;
      double right = getUpperBound() - anchorValue ;

      return new NumberRange(anchorValue - left * percent, anchorValue + right * percent) ;
    } else {
      return null ;
    }
  }

  /**
   * <zh>ŴݷΧ.</zh>
   * <en>Zooms in on the current range.</en>
   *
   * @param lowerPercent double <zh>µСֵ.</zh>
   * <en>The new lower bound.</en>
   * @param upperPercent double <zh>µĴֵ.</zh>
   * <en>The new upper bound.</en>
   * @param invert boolean <zh>Ƿת.</zh>
   * <en>Whether or not the axis is invert.</en>
   *
   * @return NumberRange <zh>µݷΧ,percent<=0,null.</zh>
   * <en>The new NumberRange,if percent<=0 return null.</en>
   */
  public NumberRange zoomRange(double lowerPercent, double upperPercent, boolean invert) {
    double start = getLowerBound() ;
    double end = getUpperBound() ;
    double length = getLength() ;
    if (invert) {
      return new NumberRange(start + length * upperPercent, end - length * lowerPercent) ;
    } else {
      return new NumberRange(start + length * lowerPercent, end - length * upperPercent) ;
    }
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this object for equality with an arbitrary object.</en>
   *
   * @param obj <zh>ҪȽϵһ (Ϊnull).</zh>
   * <en>The object to test against (null permitted).</en>
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }

    if (!(obj instanceof NumberRange)) {
      return false ;
    }

    NumberRange range = (NumberRange) obj ;
    if (!(this.lower == range.lower)) {
      return false ;
    }
    if (!(this.upper == range.upper)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcode.</zh>
   * <en>Returns a hash code.</en>
   *
   * @return A hash code.
   */
  public int hashCode() {
    long temp = Double.doubleToLongBits(this.lower) ;
    int result = (int) (temp ^ (temp >>> 32)) ;
    temp = Double.doubleToLongBits(this.upper) ;
    result = 29 * result + (int) (temp ^ (temp >>> 32)) ;

    return result ;
  }

  /**
   * <zh>¡.</zh>
   * <en>Creates and returns a copy of this object.</en>
   *
   * @throws java.lang.CloneNotSupportedException
   * @return Object
   */
  protected Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of this Range.</en>
   *
   * @return A String.
   */
  public String toString() {
    return "[lower=" + this.lower + ",upper=" + this.upper + "]" ;
  }

}
