package com.ijchart.xychart.render.category ;

import java.io.Serializable ;

import com.ijchart.xychart.dataset.IMatrixDataset ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ṩ(category)renderĳʵ.</zh>
 * <en>Base class providing common services for category renders.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public abstract class AbstractBaseValueCategoryRender extends AbstractCategoryRender implements
  Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -8111708088439649571L ;

  /**
   * <zh>ȱʡڻ״ͼĵĻֵ,̶ֵֵСֵֵ֮.</zh>
   * <en>The default base value for the bars,and the value must between the minimum and
   * maximum value of axis tick mark.</en>
   */
  public static final double DEFAULT_BASE_VALUE = 0.0 ;

  /**
   * <zh>ڻ״ͼĵĻֵ,ֵݼСֵֵ֮.</zh>
   * <en>The base value for the bars.</en>
   */
  private double baseValue ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  protected AbstractBaseValueCategoryRender(IMatrixDataset dataset) {
    super(dataset) ;

    this.baseValue = DEFAULT_BASE_VALUE ;
  }

  /**
   * <zh>ڻ״ͼĵĻֵ,ֵݼСֵֵ֮.</zh>
   * <en>The base value for the bars.</en>
   *
   * @return <zh>ڻ״ͼĵĻֵ.</zh>
   * <en>The base value for the bars.</en>
   */
  public double getBaseValue() {
    return this.baseValue ;
  }

  /**
   * <zh>ڻ״ͼĵĻֵ,ֵݼСֵֵ֮,
   * м{@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the base value for the bars and sends a
   *  {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param value double <zh>ڻ״ͼĵĻֵ.</zh>
   * <en>The new base value.</en>
   */
  public void setBaseValue(double value) {
    if (this.baseValue != value) {
      this.baseValue = value ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates if some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof AbstractBaseValueCategoryRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    AbstractBaseValueCategoryRender that = (AbstractBaseValueCategoryRender) obj ;
    if (this.baseValue != that.baseValue) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = 37 * result + (int) (this.baseValue * 100) ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[baseValue=" + this.baseValue + "]" ;
  }

}
