package com.ijchart.xychart.render.category ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.shape.ParallelepipedTextChartShape ;
import com.ijchart.xychart.ui.IEffect3D ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>3D͵״ͼ.</zh>
 * <en>Draw 3D bar shapes.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class Bar3DRender extends Bar2DRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 1889234991548074717L ;

  /**
   * <zh>3DʾЧ.</zh>
   * <en>The IEffect3D object.</en>
   */
  private IEffect3D effect ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param effect3D IEffect3D <zh>3DʾЧ (Ϊnull).</zh>
   * <en>The IEffect3D object (null not permitted).</en>
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public Bar3DRender(IEffect3D effect3D, IMatrixDataset dataset) {
    super(dataset) ;

    if (effect3D == null) {
      throw new IllegalArgumentException("Null 'effect3D' argument.") ;
    }
    this.effect = effect3D ;
  }

  /**
   * <zh>3DʾЧ.</zh>
   * <en>The IEffect3D object.</en>
   *
   * @return IEffect3D
   */
  public IEffect3D getEffect3D() {
    return this.effect ;
  }

  /**
   * <zh>ο3DʾЧ,
   * м{@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the IEffect3D object and sends a
   *  {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param effect3D IEffect3D <zh>3DʾЧ.</zh>
   * <en>The IEffect3D object.</en>
   */
  public void setEffect3D(IEffect3D effect3D) {
    if (effect3D == null) {
      throw new IllegalArgumentException("Null 'effect3D' argument.") ;
    }

    this.effect = effect3D ;
    fireRenderChanged();
  }

  /**
    * <zh>ֻͼ,ͼε.</zh>
    * <en>Create a shape with no attribute specified.</en>
    *
    * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
    * <en>The bar rectangle (null not permitted).</en>
    * @param text String <zh>Ҫʾ (Ϊnull).</zh>
    * <en>The text (null permitted).</en>
    * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
    * <en>The append parameter,used for special propose (null permitted).</en>
    *
    * @return IChartShape <zh>ͼ.</zh>
    * <en>The shape.</en>
    */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    //ƫx,y offset
    Rectangle2D cloneRect = (Rectangle2D) rect.clone() ;
    cloneRect.setRect(cloneRect.getX() - this.effect.getXOffset(),
                      cloneRect.getY() + this.effect.getYOffset(),
                      cloneRect.getWidth(), cloneRect.getHeight()) ;

    ParallelepipedTextChartShape shape = new ParallelepipedTextChartShape(cloneRect, text) ;
    shape.setXOffset(this.effect.getXOffset()) ;
    shape.setYOffset(this.effect.getYOffset()) ;

    return shape ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof Bar3DRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    Bar3DRender that = (Bar3DRender) obj ;
    if (this.effect != that.effect) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = 37 * result + this.effect.hashCode() ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[effect=" + this.effect + "]" ;
  }
}
