package com.ijchart.xychart.render.category ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;
import java.util.List ;

import com.ijchart.xychart.axis.category.CategoryAxis ;
import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.LegendType ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.data.group.NumberGroupChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.exception.InvalidChartDataException ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ƶջʽ״ͼ.</zh>
 * <en>Draw stacked bar shapes.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class GroupStackedBarRender extends StackedBarRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 1618799877383395810L ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public GroupStackedBarRender(IMatrixDataset dataset) {
    this(dataset, false) ;
  }

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   * @param renderAsPercentages boolean <zh>Ƿ԰ٷֱʽʾ.</zh>
   * <en>A flag that controls whether the data values are rendered as percentages.</en>
   */
  public GroupStackedBarRender(IMatrixDataset dataset, boolean renderAsPercentages) {
    super(dataset, renderAsPercentages) ;
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param categoryAxis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The category axis (null not permitted).</en>
   * @param valueAxis AbstractValueAxis <zh>value (Ϊnull).</zh>
   * <en>The value axis (null not permitted).</en>
   * @param valueAxisEdge RectangleEdge <zh>valueλ (Ϊnull).</zh>
   * <en>The value axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(CategoryAxis categoryAxis, AbstractValueAxis valueAxis,
                           RectangleEdge valueAxisEdge, Rectangle2D dataArea,
                           ChartOrientation orientation, RenderShapeCollection renderShapes,
                           float alpha) {
    if (categoryAxis == null) {
      throw new IllegalArgumentException("Null 'categoryAxis' argument.") ;
    }
    if (valueAxis == null) {
      throw new IllegalArgumentException("Null 'valueAxis' argument.") ;
    }
    if (valueAxisEdge == null) {
      throw new IllegalArgumentException("Null 'valueAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }

    int columns = dataset.getColumnCount() ;
    int rows = dataset.getRowCount() ;

    double basePosition = valueAxis.valueToJava2D(dataArea, valueAxisEdge, 0F, false) ;
    if (basePosition == AbstractValueAxis.INVALID_COORDINATE_VALUE) { //ֵ᷶Χ
      return ;
    }
    double maxPointTotal = getDatasetMaxPointTotal() ;
    boolean renderAsPercentages = isRenderAsPercentages() ;

    int maxDatasCount = getDatasetMaxPointDatas() ;
    this.fillPaints = getFillPaints(maxDatasCount) ;
    this.borderPaints = getBorderPaints(maxDatasCount) ;
    this.borderStrokes = getBorderStrokes(maxDatasCount) ;

    //״ͼ
    if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
      double gapSize = 0 ; //С
      if (rows > 1) {
        gapSize = dataArea.getWidth() * categoryAxis.getCategoryMargin() / (rows - 1) ;
      }
      double axisMargin = categoryAxis.getLowerMargin() + categoryAxis.getUpperMargin()
                          + ((rows == 1) ? 0 : categoryAxis.getCategoryMargin()) ; //еĿհ
      double categoryWidth = dataArea.getWidth() * (1 - axisMargin) / rows ; //ÿϵеĿ
      double categoryOneMargin = (dataArea.getWidth()
                                  * (1 - categoryAxis.getLowerMargin()
                                     - categoryAxis.getUpperMargin())
                                  * getItemMargin()) / rows ; //ÿcategoryĿհ
      double calculateBarWidth = (categoryWidth - categoryOneMargin) / columns ; //Ŀ
      double reallyBarWidth = Math.min(calculateBarWidth, getMaximumBarWidth()) ; //ʵʿ
      if (reallyBarWidth < getMinimumBarWidth()) { //ĿССƿ
        return ;
      }
      double barItemMargin = categoryOneMargin / columns ; //ÿ֮Ŀ

      for (int i = 0 ; i < rows ; i++) {
        double startX = dataArea.getX() + dataArea.getWidth() * categoryAxis.getLowerMargin()
                        + i * (categoryWidth + gapSize) ;
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof NumberGroupChartData) {
            NumberGroupChartData groupData = (NumberGroupChartData) data ;
            List datas = groupData.getDatas() ;
            if (datas != null) {
              double positiveY = 0, negativeY = 0 ; //ֵyƫֵ
              double x11 = startX + 0.5 * barItemMargin + 0.5 * (calculateBarWidth - reallyBarWidth)
                           + (calculateBarWidth + barItemMargin) * j ;
              for (int m = 0 ; m < datas.size() ; m++) {
                Number number = (Number) datas.get(m) ;
                if (number != null) {
                  double value = number.doubleValue() ;
                  value = (renderAsPercentages && maxPointTotal != 0) ? 100 * value / maxPointTotal
                          : value ;
                  double y = valueAxis.valueToJava2D(dataArea, valueAxisEdge, value, false) ;
                  //ֵ᷶Χֵ
                  if (y == AbstractValueAxis.INVALID_COORDINATE_VALUE || y == basePosition) {
                    renderShapes.addDataItemShape(null) ;
                    continue ;
                  }

                  double rectY = Math.min(y, basePosition) ;
                  if (y < basePosition) { //λڻֵ֮
                    rectY -= positiveY ;
                    positiveY += basePosition - y ;
                  } else {
                    rectY += negativeY ;
                    negativeY += y - basePosition ;
                  }
                  Rectangle2D rect = new Rectangle2D.Double(x11, rectY, reallyBarWidth,
                                                            Math.abs(basePosition - y)) ;
                  IChartShape shape = createItemShape(orientation, valueAxis, i, j, rect,
                                                      fillPaints[m], borderPaints[m],
                                                      borderStrokes[m], alpha, new Integer(m)) ;
                  renderShapes.addDataItemShape(shape) ;
                } else {
                  renderShapes.addDataItemShape(null) ;
                }
              }
            }
          } else {
            renderShapes.addDataItemShape(null) ;
          }
        }
      }
    } else if (orientation.equals(ChartOrientation.HORIZONTAL)) {
      double gapSize = 0 ; //С
      if (rows > 1) {
        gapSize = dataArea.getHeight() * categoryAxis.getCategoryMargin() / (rows - 1) ;
      }
      double axisMargin = categoryAxis.getLowerMargin() + categoryAxis.getUpperMargin()
                          + ((rows == 1) ? 0 : categoryAxis.getCategoryMargin()) ; //еĿհ
      double categoryWidth = dataArea.getHeight() * (1 - axisMargin) / rows ; //ÿϵеĿ
      double categoryOneMargin = (dataArea.getHeight()
                                  * (1 - categoryAxis.getLowerMargin()
                                     - categoryAxis.getUpperMargin())
                                  * getItemMargin()) / rows ; //ÿcategoryĿհ
      double calculateBarWidth = (categoryWidth - categoryOneMargin) / columns ; //Ŀ
      double reallyBarWidth = Math.min(calculateBarWidth, getMaximumBarWidth()) ; //ʵʿ
      if (reallyBarWidth < getMinimumBarWidth()) { //ĿССƿ
        return ;
      }
      double barItemMargin = categoryOneMargin / columns ; //ÿ֮Ŀ

      for (int i = 0 ; i < rows ; i++) {
        double startY = dataArea.getY() + dataArea.getHeight() * categoryAxis.getLowerMargin()
                        + i * (categoryWidth + gapSize) ;
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof NumberGroupChartData) {
            NumberGroupChartData groupData = (NumberGroupChartData) data ;
            List datas = groupData.getDatas() ;
            if (datas != null) {
              double positiveX = 0, negativeX = 0 ; //ֵxƫֵ
              double y11 = startY + 0.5 * barItemMargin + 0.5 * (calculateBarWidth - reallyBarWidth)
                           + (calculateBarWidth + barItemMargin) * j ;
              for (int m = 0 ; m < datas.size() ; m++) {
                Number number = (Number) datas.get(m) ;
                if (number != null) {
                  double value = number.doubleValue() ;
                  value = (renderAsPercentages && maxPointTotal != 0) ? 100 * value / maxPointTotal
                          : value ;
                  double x = valueAxis.valueToJava2D(dataArea, valueAxisEdge, value, false) ;
                  //ֵ᷶Χֵ
                  if (x == AbstractValueAxis.INVALID_COORDINATE_VALUE || x == basePosition) {
                    renderShapes.addDataItemShape(null) ;
                    continue ;
                  }

                  double rectX = Math.min(x, basePosition) ; //x
                  if (x < basePosition) { //λڻֵ֮
                    rectX -= positiveX ;
                    positiveX += basePosition - x ;
                  } else {
                    rectX += negativeX ;
                    negativeX += x - basePosition ;
                  }

                  Rectangle2D rect = new Rectangle2D.Double(rectX, y11, Math.abs(basePosition - x),
                                                            reallyBarWidth) ;
                  IChartShape shape = createItemShape(orientation, valueAxis, i, j, rect,
                                                      fillPaints[m], borderPaints[m],
                                                      borderStrokes[m], alpha, new Integer(m)) ;
                  renderShapes.addDataItemShape(shape) ;
                } else {
                  renderShapes.addDataItemShape(null) ;
                }
              }
            }
          } else {
            renderShapes.addDataItemShape(null) ;
          }
        }
      }
    }

  }

  /**
   * <zh>ͼ.</zh>
   * <en>Return the legend type.</en>
   *
   * @return LegendType <zh>ͼ.</zh>
   * <en>The legend type.</en>
   */
  public LegendType getLegendType() {
    return LegendType.ROW ;
  }

  /**
   * <zh>ͼĸ.</zh>
   * <en>Returns the count of legends.</en>
   *
   * @return int <zh>ͼĸ.</zh>
   * <en>The count of legends.</en>
   */
  protected int getLegendItemCount() {
    IMatrixDataset dataset = getDataset() ;
    if (dataset != null) {
      return getDatasetMaxPointDatas() ;
    }

    return 0 ;
  }

  /**
   * <zh>ݼһݵݵ.</zh>
   * <en>Returns the max data count of a dataset point.</en>
   *
   * @return int <zh>.</zh>
   * <en>The count.</en>
   */
  private int getDatasetMaxPointDatas() {
    int maxCount = 0 ;

    IMatrixDataset dataset = getDataset() ;
    int rows = dataset.getRowCount() ;

    if (dataset != null) {
      for (int i = 0 ; i < rows ; i++) {
        int currRowCols = dataset.getColumnCount(i) ;
        for (int j = 0 ; j < currRowCols ; j++) {
          Object data = dataset.getData(i, j) ;
          if (data != null) {
            if (!(data instanceof NumberGroupChartData)) {
              throw new InvalidChartDataException(
                "Invalid chart data type,musts be 'NumberGroupChartData'.") ;
            } else { //Сֵиֵ֮,ֵֵ֮
              NumberGroupChartData numberData = (NumberGroupChartData) data ;
              List datas = numberData.getDatas() ;
              if (datas != null) {
                maxCount = Math.max(maxCount, datas.size()) ;
              }
            }
          }
        }
      }
    }

    return maxCount ;
  }

  /**
   * <zh>ݼһ֮͵ֵ.</zh>
   * <en>Calculate the max total value of a dataset value.</en>
   *
   * @return double
   */
  private double getDatasetMaxPointTotal() {
    double maxPointTotal = 0 ;

    IMatrixDataset dataset = getDataset() ;
    int rows = dataset.getRowCount() ;

    if (dataset != null && isRenderAsPercentages()) {
      for (int i = 0 ; i < rows ; i++) {
        int currRowCols = dataset.getColumnCount(i) ;
        for (int j = 0 ; j < currRowCols ; j++) {
          Object data = dataset.getData(i, j) ;
          if (data != null) {
            if (!(data instanceof NumberGroupChartData)) {
              throw new InvalidChartDataException(
                "Invalid chart data type,musts be 'NumberGroupChartData'.") ;
            } else { //Сֵиֵ֮,ֵֵ֮
              double totalPoint = 0 ;
              NumberGroupChartData numberData = (NumberGroupChartData) data ;
              List datas = numberData.getDatas() ;

              if (datas != null) {
                for (int m = 0 ; m < datas.size() ; m++) {
                  Number number = (Number) datas.get(m) ;
                  if (number != null) {
                    totalPoint += Math.abs(number.doubleValue()) ;
                  }
                }
              }

              maxPointTotal = Math.max(maxPointTotal, totalPoint) ;
            }
          }
        }

      }
    }

    return maxPointTotal ;
  }

}
