package com.ijchart.xychart.render.category.outlier ;

import java.awt.geom.Point2D ;
import java.io.Serializable ;

import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 * <zh>ʾBoxAndWhisker(ͼ)Outlier.</zh>
 * <en>Represents one outlier in the box and whisker plot.
 * All the coordinates in this class are in Java2D space.</en>
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class Outlier implements Comparable, Serializable, Cloneable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -4767944879956037020L ;

  /**
   * <zh>outlierԲεľε.</zh>
   * <en>The xy coordinates of the bounding box containing the outlier ellipse.</en>
   */
  private Point2D point ;

  /**
   * <zh>Բεİ뾶.</zh>
   * <en>The radius of the ellipse.</en>
   */
  private double radius ;

  /**
   * <zh>(x,y)Ͱ뾶ȹ.</zh>
   * <en>Constructs an outlier item consisting of a point and the radius of the
   * outlier ellipse.</en>
   *
   * @param xCoord double <zh>x.</zh>
   * <en>The x coordinate of the point.</en>
   * @param yCoord double <zh>y.</zh>
   * <en>The y coordinate of the point.</en>
   * @param radius double <zh>뾶ĳ.</zh>
   * <en>The radius of the ellipse.</en>
   */
  public Outlier(double xCoord, double yCoord, double radius) {
    this.point = new Point2D.Double(xCoord - radius, yCoord - radius) ;
    this.radius = radius ;
  }

  /**
   * <zh>ذoutlierԲεľε.</zh>
   * <en>Returns the xy coordinates of the bounding box containing the outlier ellipse.</en>
   *
   * @return Point2D <zh>outlierԲεľε</zh>
   * <en>The location of the outlier ellipse.</en>
   */
  public Point2D getPoint() {
    return this.point ;
  }

  /**
   * <zh>ðoutlierԲεľε.</zh>
   * <en>Sets the xy coordinates of the bounding box containing the outlier ellipse.</en>
   *
   * @param point Point2D <zh>.</zh>
   * <en>The location.</en>
   */
  public void setPoint(Point2D point) {
    this.point = point ;
  }

  /**
   * <zh>ذoutlierԲεľεx.</zh>
   * <en>Returns the x coordinate of the bounding box containing the outlier ellipse.</en>
   *
   * @return double <zh>x.</zh>
   * <en>The x coordinate.</en>
   */
  public double getX() {
    return getPoint().getX() ;
  }

  /**
   * <zh>ذoutlierԲεľεy.</zh>
   * <en>Returns the y coordinate of the bounding box containing the outlier ellipse.</en>
   *
   * @return double <zh>y.</zh>
   * <en>The y coordinate.</en>
   */
  public double getY() {
    return getPoint().getY() ;
  }

  /**
   * <zh>Բεİ뾶.</zh>
   * <en>Returns the radius of the outlier ellipse.</en>
   *
   * @return double <zh>뾶.</zh>
   * <en>The radius.</en>
   */
  public double getRadius() {
    return this.radius ;
  }

  /**
   * <zh>Բεİ뾶.</zh>
   * <en>Sets the radius of the outlier ellipse.</en>
   *
   * @param radius <zh>뾶.</zh>
   * <en>The new radius.</en>
   */
  public void setRadius(double radius) {
    this.radius = radius ;
  }

  /**
   * <zh>ȽĴС.</zh>
   * <en>Compares this object with the specified object for order, based on
   * the outlier's point.</en>
   *
   * @param   o the Object to be compared.
   * @return A negative integer, zero, or a positive integer as this object
   *      is less than, equal to, or greater than the specified object.
   *
   */
  public int compareTo(Object o) {
    Outlier outlier = (Outlier) o ;
    Point2D p1 = getPoint() ;
    Point2D p2 = outlier.getPoint() ;

    if (p1.equals(p2)) {
      return 0 ;
    } else if ((p1.getX() < p2.getX()) || (p1.getY() < p2.getY())) {
      return -1 ;
    } else {
      return 1 ;
    }
  }

  /**
   * <zh>жǷص.</zh>
   * <en>Returns a true if outlier is overlapped and false if it is not.
   * Overlapping is determined by the respective bounding boxes plus
   * a small margin.</en>
   *
   * @param other  the other outlier.
   *
   * @return A boolean indicating whether or not an overlap has
   *         occurred.
   */
  public boolean overlaps(Outlier other) {
    return ((other.getX() >= getX() - (this.radius * 1.1))
            && (other.getX() <= getX() + (this.radius * 1.1))
            && (other.getY() >= getY() - (this.radius * 1.1))
            && (other.getY() <= getY() + (this.radius * 1.1))) ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof Outlier)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    Outlier that = (Outlier) obj ;
    if (!EqualsUtil.equals(this.point, that.point)) {
      return false ;
    }
    if (this.radius != that.radius) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() + HashcodeUtil.hashCode(this.point) ;
    result = 37 * result + this.radius ;

    return (int) result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a textual representation of the outlier.</en>
   *
   * @return A String representing the outlier.
   */
  public String toString() {
    return "[" + getX() + "," + getY() + "]" ;
  }

}
