package com.ijchart.xychart.render.noaxis ;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Font ;
import java.awt.Paint ;
import java.awt.Stroke ;
import java.awt.geom.Ellipse2D ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.Line2D ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;

import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.DataBackgroundShapeDrawType ;
import com.ijchart.xychart.constant.HorizontalAlignment ;
import com.ijchart.xychart.constant.LegendType ;
import com.ijchart.xychart.constant.Rotation ;
import com.ijchart.xychart.constant.TextPosition ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.render.INoAxisRender ;
import com.ijchart.xychart.shape.EllipseTextChartShape ;
import com.ijchart.xychart.shape.GeneralPathChartShape ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.shape.LineAndShapeChartShape ;
import com.ijchart.xychart.shape.LineChartShape ;
import com.ijchart.xychart.shape.RectangleTextChartShape ;
import com.ijchart.xychart.util.AngleUtil ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ͼ.</zh>
 * <en>Draw spider web chart shapes.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class SpiderWebRender extends AbstractNoAxisRender implements INoAxisRender, Cloneable,
  Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -3314147370814917609L ;

  /**
   * <zh>ȱʡͼδС,λǰٷֱ.</zh>
   * <en>The default head radius percent (currently 1%).</en>
   */
  public static final double DEFAULT_HEAD_SIZE = 0.01 ;

  /**
   * <zh>ȱʡĿհ׵ĴС.</zh>
   * <en>The default space left around the outside of the plot as a percentage.</en>
   */
  public static final double DEFAULT_INTERIOR_GAP = 0.25 ;

  /**
   * <zh>ȱʡĻߵpaint.</zh>
   * <en>The default paint used to draw the axis lines.</en>
   */
  public static final Paint DEFAULT_AXIS_LINE_PAINT = Color.black ;

  /**
   * <zh>ȱʡĻߵstroke.</zh>
   * <en>The default stroke used to draw the axis lines.</en>
   */
  public static final Stroke DEFAULT_AXIS_LINE_STROKE = new BasicStroke(1.0f) ;

  /**
   * <zh>ȱʡߵĻƷ.</zh>
   * <en>The default rotation of draw axis.</en>
   */
  public static final Rotation DEFAULT_DIRECTION = Rotation.CLOCKWISE ;

  /**
   * <zh>ȱʡĻƵʼǶ.</zh>
   * <en>The default starting angle.</en>
   */
  public static final double DEFAULT_START_ANGLE = 75.0 ;

  /**
   * <zh>ȱʡĻlabel.</zh>
   * <en>The default font used to display the labels.</en>
   */
  public static final Font DEFAULT_LABEL_FONT = new Font("SansSerif", Font.PLAIN, 10) ;

  /**
   * <zh>ȱʡĻlabelpaint.</zh>
   * <en>The default color used to draw the category labels.</en>
   */
  public static final Paint DEFAULT_LABEL_PAINT = Color.black ;

  /**
   * <zh>ͼδС,λǰٷֱ.</zh>
   * <en>The head radius percent.</en>
   */
  private double headSize ;

  /**
   * <zh>հ׵ĴС.</zh>
   * <en>The space left around the outside of the plot as a percentage.</en>
   */
  private double interiorGap ;

  /**
   * <zh>ߵpaint.</zh>
   * <en>The paint used to draw the axis lines.</en>
   */
  private transient Paint axisLinePaint ;

  /**
   * <zh>ߵstroke.</zh>
   * <en>The stroke used to draw the axis lines.</en>
   */
  private transient Stroke axisLineStroke ;

  /**
   * <zh>ߵĻƷ.</zh>
   * <en>The rotation of draw axis.</en>
   */
  private Rotation direction ;

  /**
   * <zh>ƵʼǶ.</zh>
   * <en>The starting angle.</en>
   */
  private double startAngle ;

  /**
   * <zh>label.</zh>
   * <en>The font used to display the labels.</en>
   */
  private Font labelFont ;

  /**
   * <zh>labelpaint.</zh>
   * <en>The color used to draw the category labels.</en>
   */
  private transient Paint labelPaint ;

  /**
   * <zh>ͼ.</zh>
   * <en>The head shape.</en>
   */
  private transient Ellipse2D headShape ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public SpiderWebRender(IMatrixDataset dataset) {
    super(dataset) ;

    this.headSize = DEFAULT_HEAD_SIZE ;
    this.interiorGap = DEFAULT_INTERIOR_GAP ;
    this.axisLinePaint = DEFAULT_AXIS_LINE_PAINT ;
    this.axisLineStroke = DEFAULT_AXIS_LINE_STROKE ;

    this.direction = DEFAULT_DIRECTION ;
    this.startAngle = DEFAULT_START_ANGLE ;

    this.labelFont = DEFAULT_LABEL_FONT ;
    this.labelPaint = DEFAULT_LABEL_PAINT ;

    setAutoSelectBorderPaint(true) ;
    setAutoSelectBorderStroke(true) ;
    setBaseItemDrawBorder(true) ;
  }

  /**
   * <zh>ͼδС.</zh>
   * <en>Returns the head size.</en>
   *
   * @return double <zh>ͼδС.</zh>
   * <en>The head size.</en>
   */
  public double getHeadSize() {
    return this.headSize ;
  }

  /**
   * <zh>ͼδС,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the head size and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param size double <zh>ͼδС(ٷֱ).</zh>
   * <en>The head size,in percent.</en>
   */
  public void setHeadSize(double size) {
    if (size <= 0) {
      throw new IllegalArgumentException("Argument size must be >0.") ;
    }
    if (this.headSize != size) {
      this.headSize = size ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ؿհ׵ĴС.</zh>
   * <en>Returns the space left around the outside of the plot as a percentage.</en>
   *
   * @return double <zh>հ׵ĴС.</zh>
   * <en>The space size.</en>
   */
  public double getInteriorGap() {
    return this.interiorGap ;
  }

  /**
   * <zh>ÿհ׵ĴС,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the space left around the outside of the plot as a percentage and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param gap double <zh>հ׵ĴС.</zh>
   * <en>The space size.</en>
   */
  public void setInteriorGap(double gap) {
    if (this.interiorGap != gap) {
      this.interiorGap = gap ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ػߵpaint.</zh>
   * <en>Returns the paint used to draw the axis lines.</en>
   *
   * @return Paint <zh>paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getAxisLinePaint() {
    return this.axisLinePaint ;
  }

  /**
   * <zh>ûߵpaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint used to draw the axis lines and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>ߵpaint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setAxisLinePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument") ;
    }

    this.axisLinePaint = paint ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ػߵstroke.</zh>
   * <en>Returns the stroke used to draw the axis lines.</en>
   *
   * @return Stroke <zh>stroke (Ϊnull).</zh>
   * <en>The stroke (never null).</en>
   */
  public Stroke getAxisLineStroke() {
    return this.axisLineStroke ;
  }

  /**
   * <zh>ûߵstroke,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the stroke used to draw the axis lines and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>stroke (Ϊnull).</zh>
   * <en>The stroke (null not permitted).</en>
   */
  public void setAxisLineStroke(Stroke stroke) {
    if (stroke == null) {
      throw new IllegalArgumentException("Null 'stroke' argument") ;
    }

    this.axisLineStroke = stroke ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ߵĻƷ.</zh>
   * <en>Returns the rotation of draw axis.</en>
   *
   * @return Rotation <zh>Ʒ (Ϊnull).</zh>
   * <en>The rotation (never null).</en>
   */
  public Rotation getDirection() {
    return this.direction ;
  }

  /**
   * <zh>ߵĻƷ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the rotation of draw axis and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param direction Rotation <zh>Ʒ (Ϊnull).</zh>
   * <en>The rotation (null not permitted).</en>
   */
  public void setDirection(Rotation direction) {
    if (direction == null) {
      throw new IllegalArgumentException("Null 'direction' argument") ;
    }

    if (!this.direction.equals(direction)) {
      this.direction = direction ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ػƵʼǶ.</zh>
   * <en>Returns the starting angle.</en>
   *
   * @return double <zh>ʼǶ.</zh>
   * <en>The starting angle.</en>
   */
  public double getStartAngle() {
    return this.startAngle ;
  }

  /**
   * <zh>ûƵʼǶ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the starting angle and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param angle double <zh>ƵʼǶ.</zh>
   * <en>The starting angle.</en>
   */
  public void setStartAngle(double angle) {
    if (this.startAngle != angle) {
      this.startAngle = angle ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ػlabel.</zh>
   * <en>Returns the font used to display the labels.</en>
   *
   * @return Font <zh> (Ϊnull).</zh>
   * <en>The font (never null).</en>
   */
  public Font getLabelFont() {
    return this.labelFont ;
  }

  /**
   * <zh>ûlabel,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the font used to display the labels and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param font Font <zh> (Ϊnull).</zh>
   * <en>The font (null not permitted).</en>
   */
  public void setLabelFont(Font font) {
    if (font == null) {
      throw new IllegalArgumentException("Null 'font' argument") ;
    }

    this.labelFont = font ;
    fireRenderChanged() ;
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(Rectangle2D dataArea, RenderShapeCollection renderShapes, float alpha) {
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }

    int rows = dataset.getRowCount() ;
    int cols = dataset.getColumnCount() ;
    this.fillPaints = getFillPaints(rows) ;
    this.borderPaints = getBorderPaints(rows) ;
    this.borderStrokes = getBorderStrokes(rows) ;
    renderShapes.setDataBackgroundShapeDrawType(DataBackgroundShapeDrawType.ONE_BY_ROW) ;

    double gapHorizontal = dataArea.getWidth() * this.interiorGap ;
    double gapVertical = dataArea.getHeight() * this.interiorGap ;

    double drawX = dataArea.getX() + gapHorizontal * 0.5 ;
    double drawY = dataArea.getY() + gapVertical * 0.5 ;
    double drawWidth = dataArea.getWidth() - gapHorizontal ;
    double drawHeight = dataArea.getHeight() - gapVertical ;

    double headWidth = dataArea.getWidth() * this.headSize ;
    double headHeight = dataArea.getHeight() * this.headSize ;
    headShape = new Ellipse2D.Double( -headWidth / 2, -headHeight / 2, headWidth, headHeight) ;

    // make the chart area a square
    double minAreaSize = Math.min(drawWidth, drawHeight) * 0.5 ;
    drawX = drawX + drawWidth * 0.5 - minAreaSize ;
    drawY = drawY + drawHeight * 0.5 - minAreaSize ;
    drawWidth = drawHeight = 2 * minAreaSize ;

    double centerX = drawX + minAreaSize ;
    double centerY = drawY + minAreaSize ;

    String[] ticks = new String[cols] ;
    for (int i = 0 ; i < cols ; i++) {
      Object colKey = dataset.getColKey(i) ;
      if (colKey != null) {
        ticks[i] = colKey.toString() ;
      }
    }

    for (int i = 0 ; i < cols ; i++) {
      double angle = -this.startAngle - this.direction.getFactor() * i * 360 / cols ;
      double x = centerX + minAreaSize * Math.cos(angle / 180.0 * Math.PI) ;
      double y = centerY + minAreaSize * Math.sin(angle / 180.0 * Math.PI) ;

      LineChartShape lineShape = new LineChartShape(centerX, centerY, x, y) ;
      lineShape.setDrawBorder(true) ;
      lineShape.setBorderPaint(this.axisLinePaint) ;
      lineShape.setBorderStroke(this.axisLineStroke) ;
      renderShapes.addAngleGridlineShape(lineShape) ;

      Rectangle2D textArea = new Rectangle2D.Double(x, y, 1, 1) ;
      RectangleTextChartShape textShape = new RectangleTextChartShape(textArea, ticks[i]) ;
      textShape.setFillShape(false) ;
      textShape.setDrawBorder(false) ;
      textShape.setTextFont(this.labelFont) ;
      textShape.setTextPaint(this.labelPaint) ;

      double modAngle = angle % 360.0 ;
      if (modAngle == -90.0) {
        textShape.setTextPosition(TextPosition.OUTSIDE_TOP) ;
        textShape.setTextHorizontalAlign(HorizontalAlignment.CENTER) ;
      } else if (modAngle == 0.0) {
        textShape.setTextPosition(TextPosition.OUTSIDE_RIGHT) ;
        textShape.translate(5, -3) ;
      } else if (modAngle == 90.0) {
        textShape.setTextPosition(TextPosition.OUTSIDE_BOTTOM) ;
        textShape.setTextHorizontalAlign(HorizontalAlignment.CENTER) ;
      } else if (modAngle == 180.0) {
        textShape.setTextPosition(TextPosition.OUTSIDE_LEFT) ;
      } else {
        int quadrant = AngleUtil.getQuadrantByDegreeAngle(angle) ;
        if (quadrant == 1 || quadrant == 4) {
          textShape.setTextPosition(TextPosition.OUTSIDE_RIGHT) ;
        } else if (quadrant == 2 || quadrant == 3) {
          textShape.setTextPosition(TextPosition.OUTSIDE_LEFT) ;
        }
      }
      renderShapes.addPieLabelShape(textShape) ;
    }

    double maxValue = 0.0D ;
    for (int i = 0 ; i < rows ; i++) {
      int currRowCols = dataset.getColumnCount(i) ;
      for (int j = 0 ; j < currRowCols ; j++) {
        IChartData data = dataset.getData(i, j) ;
        if (data != null && data instanceof NumberChartData) {
          NumberChartData numberData = (NumberChartData) data ;
          double value = numberData.getValue() ;
          if (!Double.isNaN(value)) {
            maxValue = Math.max(maxValue, value) ;
          }
        }
      }
    }

    //draw data shapes
    for (int i = 0 ; i < rows ; i++) {
      int currRowCols = dataset.getColumnCount(i) ;
      Point2D[] points = new Point2D.Double[currRowCols] ;
      for (int j = 0 ; j < currRowCols ; j++) {
        IChartData data = dataset.getData(i, j) ;
        if (data != null && data instanceof NumberChartData) {
          NumberChartData numberData = (NumberChartData) data ;
          double value = numberData.getValue() ;
          if (Double.isNaN(value) || value < 0) {
            points[j] = null ;
          } else {
            double angle = -this.startAngle - this.direction.getFactor() * j * 360 / cols ;
            double xPos = centerX
                          + minAreaSize * Math.cos(angle / 180.0 * Math.PI) * value / maxValue ;
            double yPos = centerY
                          + minAreaSize * Math.sin(angle / 180.0 * Math.PI) * value / maxValue ;
            points[j] = new Point2D.Double(xPos, yPos) ;
          }
        } else {
          points[j] = null ;
        }
      }

      for (int m = 0 ; m < currRowCols ; m++) {
        if (points[m] == null) {
          renderShapes.addDataItemShape(null) ;
        } else {
          LineAndShapeChartShape lineShape = null ;
          boolean findNotNull = false ;
          for (int n = m + 1 ; n < currRowCols ; n++) {
            if (points[n] != null) {
              Line2D line = new Line2D.Double(0, 0, points[n].getX() - points[m].getX(),
                                              points[n].getY() - points[m].getY()) ;
              lineShape = new LineAndShapeChartShape(points[m].getX(), points[m].getY(),
                                                     headShape, line) ;
              findNotNull = true ;
              break ;
            }
          }

          if (!findNotNull) {
            for (int n = 0 ; n < m ; n++) {
              if (points[n] != null) {
                Line2D line = new Line2D.Double(0, 0, points[n].getX() - points[m].getX(),
                                                points[n].getY() - points[m].getY()) ;
                lineShape = new LineAndShapeChartShape(points[m].getX(), points[m].getY(),
                                                       headShape, line) ;
                break ;
              }
            }
          }

          if (lineShape != null) {
            IChartShape shape = createItemShape(i, m, fillPaints[i], fillPaints[i], borderStrokes[i],
                                                alpha, lineShape) ;
            renderShapes.addDataItemShape(shape) ;
          } else {
            renderShapes.addDataItemShape(null) ;
          }
        }
      }

      Boolean fill = isRowItemFill(i) ;
      if ((fill != null && fill.booleanValue()) || isBaseItemFill()) {
        GeneralPath path = new GeneralPath() ;
        boolean hasMove = false ;
        for (int m = 0 ; m < currRowCols ; m++) {
          if (points[m] != null) {
            if (!hasMove) {
              path.moveTo((float) points[m].getX(), (float) points[m].getY()) ;
              hasMove = true ;
            } else {
              path.lineTo((float) points[m].getX(), (float) points[m].getY()) ;
            }
          }
        }

        path.closePath() ;
        GeneralPathChartShape pathShape = new GeneralPathChartShape(path) ;
        pathShape.setFillAlpha(0.1F) ;
        pathShape.setFillShape(true) ;
        pathShape.setDrawBorder(false) ;
        pathShape.setFillPaint(fillPaints[i]) ;
        renderShapes.addDataItemBackgroundShape(pathShape) ;
      } else {
        renderShapes.addDataItemBackgroundShape(null) ;
      }
    }
  }

  /**
   * <zh>ÿͼͼ.</zh>
   * <en>Returns the shape array of legend items.</en>
   *
   * @param legendCount int <zh>ͼĸ.</zh>
   * <en>The legend items count.</en>
   * @return IChartShape[] <zh>ͼ.</zh>
   * <en>The shapes array.</en>
   */
  protected IChartShape[] getLegendItemShapes(int legendCount) {
    if (legendCount > 0) {
      final int legendWidth = 8, legendHeight = 8 ; //ͼ,߶
      IChartShape[] shapes = new IChartShape[legendCount] ;

      for (int i = 0 ; i < legendCount ; i++) {
        Rectangle2D rect = new Rectangle2D.Double(0, 0, legendWidth, legendHeight) ;
        EllipseTextChartShape ellipseShape = new EllipseTextChartShape(rect, null) ;
        ellipseShape.setTextPosition(TextPosition.OUTSIDE_RIGHT) ;
        shapes[i] = ellipseShape ;
      }

      return shapes ;
    }

    return null ;
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Return the legend type.</en>
   *
   * @return LegendType <zh>ͼ.</zh>
   * <en>The legend type.</en>
   */
  public LegendType getLegendType() {
    return LegendType.ROW ;
  }

  /**
   * <zh>ֻͼ,ͼε.</zh>
   * <en>Create a shape with no attribute specified.</en>
   *
   * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
   * <en>The bar rectangle (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
   * <en>The append parameter,used for special propose (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    if (appendParam == null) {
      throw new IllegalArgumentException("Null 'appendParam' argument.") ;
    }

    return (IChartShape) appendParam ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof SpiderWebRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    SpiderWebRender that = (SpiderWebRender) obj ;
    if (this.headSize != that.headSize) {
      return false ;
    }
    if (this.interiorGap != that.interiorGap) {
      return false ;
    }
    if (EqualsUtil.equals(this.axisLinePaint, that.axisLinePaint)) {
      return false ;
    }
    if (EqualsUtil.equals(this.axisLineStroke, that.axisLineStroke)) {
      return false ;
    }
    if (EqualsUtil.equals(this.direction, that.direction)) {
      return false ;
    }
    if (this.startAngle != that.startAngle) {
      return false ;
    }
    if (EqualsUtil.equals(this.labelFont, that.labelFont)) {
      return false ;
    }
    if (EqualsUtil.equals(this.labelPaint, that.labelPaint)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() ;
    result = 37 * result + this.headSize ;
    result = 37 * result + this.interiorGap ;
    result = 37 * result + HashcodeUtil.hashCode(this.axisLinePaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.axisLineStroke) ;
    result = 37 * result + HashcodeUtil.hashCode(this.direction) ;
    result = 37 * result + this.startAngle ;
    result = 37 * result + HashcodeUtil.hashCode(this.labelFont) ;
    result = 37 * result + HashcodeUtil.hashCode(this.labelPaint) ;

    return (int) result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "SpiderWebRender" ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;

    SerialUtil.writePaint(stream, this.axisLinePaint) ;
    SerialUtil.writeStroke(stream, this.axisLineStroke) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;

    this.axisLinePaint = SerialUtil.readPaint(stream) ;
    this.axisLineStroke = SerialUtil.readStroke(stream) ;
  }

}
