package com.ijchart.xychart.render.noaxis.pie ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 *
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ͼⲿǩĽṹ.</zh>
 * <en>A structure that retains information about the label for a section in a pie chart.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class PieLabelRecord implements Comparable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -1149375063383524323L ;

  /**
   * <zh>ͼǩļֵ.</zh>
   * <en>The section key.</en>
   */
  private Comparable key ;

  /**
   * <zh>ͼĵĽǶ(λǻ).</zh>
   * <en>The angle of the centre of the section (in radians).</en>
   */
  private double angle ;

  /**
   * <zh>y.</zh>
   * <en>The base y-coordinate.</en>
   */
  private double baseY ;

  /**
   * <zh>y.</zh>
   * <en>The allocated y-coordinate.</en>
   */
  private double allocatedY ;

  /**
   * <zh>ǩľ.</zh>
   * <en>Rectangle containing the label.</en>
   */
  private Rectangle2D labelRect ;

  /**
   * <zh>ǩ.</zh>
   * <en>The label.</en>
   */
  private String label ;

  /**
   * <zh>ǩĸ߶.</zh>
   * <en>The label height.</en>
   */
  private double labelHeight ;

  /**
   * <zh>.</zh>
   * <en>The gap.</en>
   */
  private double gap ;

  /**
   * <zh>ӵİٷֱ.</zh>
   * <en>The link percent.</en>
   */
  private double linkPercent ;

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param key Comparable <zh>ͼǩļֵ.</zh>
   * <en>The section key.</en>
   * @param angle double <zh>ͼĵĽǶ(λǻ).</zh>
   * <en>The angle of the centre of the section (in radians).</en>
   * @param baseY double <zh>y.</zh>
   * <en>The base y-coordinate.</en>
   * @param labelRect Rectangle2D <zh>ǩľ.</zh>
   * <en>Rectangle containing the label.</en>
   * @param label String <zh>ǩ.</zh>
   * <en>The label.</en>
   * @param labelHeight double <zh>ǩĸ߶.</zh>
   * <en>The label height.</en>
   * @param gap double <zh>.</zh>
   * <en>The gap.</en>
   * @param linkPercent double <zh>ӵİٷֱ.</zh>
   * <en>The link percent.</en>
   */
  public PieLabelRecord(Comparable key, double angle, double baseY, Rectangle2D labelRect,
                        String label, double labelHeight, double gap, double linkPercent) {
    if (key == null) {
      throw new IllegalArgumentException("Null 'key' argument") ;
    }

    this.key = key ;
    this.angle = angle ;
    this.baseY = baseY ;
    this.allocatedY = baseY ;
    this.labelRect = labelRect ;
    this.label = label ;
    this.labelHeight = labelHeight ;
    this.gap = gap ;
    this.linkPercent = linkPercent ;
  }

  /**
   * <zh>رͼǩļֵ.</zh>
   * <en>Returns the section key.</en>
   *
   * @return Comparable <zh>ֵ.</zh>
   * <en>The key.</en>
   */
  public Comparable getKey() {
    return this.key ;
  }

  /**
   * <zh>ñͼǩļֵ.</zh>
   * <en>Sets the section key.</en>
   *
   * @param key Comparable <zh>ֵ.</zh>
   * <en>The key.</en>
   */
  public void setKey(Comparable key) {
    if (key == null) {
      throw new IllegalArgumentException("Null 'key' argument") ;
    }

    this.key = key ;
  }

  /**
   * <zh>رͼĵĽǶ(λǻ).</zh>
   * <en>Returns the angle of the centre of the section (in radians).</en>
   *
   * @return double <zh>Ƕ.</zh>
   * <en>The angle.</en>
   */
  public double getAngle() {
    return this.angle ;
  }

  /**
   * <zh>ñͼĵĽǶ(λǻ).</zh>
   * <en>Sets the angle of the centre of the section (in radians).</en>
   *
   * @param angle double <zh>Ƕ.</zh>
   * <en>The angle.</en>
   */
  public void setAngle(double angle) {
    this.angle = angle ;
  }

  /**
   * <zh>y.</zh>
   * <en>Returns the base y-coordinate.</en>
   *
   * @return double <zh>y.</zh>
   * <en>The base y-coordinate.</en>
   */
  public double getBaseY() {
    return this.baseY ;
  }

  /**
   * <zh>y.</zh>
   * <en>Returns the base y-coordinate.</en>
   *
   * @param y double <zh>y.</zh>
   * <en>The base y-coordinate.</en>
   */
  public void setBaseY(double y) {
    this.baseY = y ;
  }

  /**
   * <zh>طy.</zh>
   * <en>Returns the allocated y-coordinate.</en>
   *
   * @return double <zh>y.</zh>
   * <en>The base y-coordinate.</en>
   */
  public double getAllocatedY() {
    return this.allocatedY ;
  }

  /**
   * <zh>÷y.</zh>
   * <en>Sets the allocated y-coordinate.</en>
   *
   * @param y double <zh>y.</zh>
   * <en>The base y-coordinate.</en>
   */
  public void setAllocatedY(double y) {
    this.allocatedY = y ;
  }

  /**
   * <zh>ذǩľ.</zh>
   * <en>Returns the rectangle containing the label.</en>
   *
   * @return TextBox <zh>ǩ.</zh>
   * <en>The label.</en>
   */
  public Rectangle2D getLabelRect() {
    return this.labelRect ;
  }

  /**
   * <zh>ðǩľ.</zh>
   * <en>Sets the rectangle containing the label.</en>
   *
   * @param labelRect Rectangle2D <zh>ǩľ.</zh>
   * <en>Rectangle containing the label.</en>
   */
  public void setLabelRect(Rectangle2D labelRect) {
    this.labelRect = labelRect ;
  }

  /**
   * <zh>ǩ.</zh>
   * <en>The label.</en>
   *
   * @param label String <zh>ǩ (Ϊnull).</zh>
   * <en>The label (null permitted).</en>
   */
  public void setLabel(String label) {
    this.label = label ;
  }

  /**
   * <zh>ǩ.</zh>
   * <en>The label.</en>
   *
   * @return String <zh>ǩ.</zh>
   * <en>The label.</en>
   */
  public String getLabel() {
    return this.label ;
  }

  /**
   * <zh>رǩĸ߶.</zh>
   * <en>Returns the label height.</en>
   *
   * @return double <zh>߶.</zh>
   * <en>The height.</en>
   */
  public double getLabelHeight() {
    return this.labelHeight ;
  }

  /**
   * <zh>ñǩĸ߶.</zh>
   * <en>Sets the label height.</en>
   *
   * @param height double <zh>߶.</zh>
   * <en>The height.</en>
   */
  public void setLabelHeight(double height) {
    this.labelHeight = height ;
  }

  /**
   * <zh>ؼ.</zh>
   * <en>Returns the gap.</en>
   *
   * @return double <zh>.</zh>
   * <en>The gap.</en>
   */
  public double getGap() {
    return this.gap ;
  }

  /**
   * <zh>ü.</zh>
   * <en>Sets the gap.</en>
   *
   * @param gap double <zh>.</zh>
   * <en>The gap.</en>
   */
  public void setGap(double gap) {
    this.gap = gap ;
  }

  /**
   * <zh>ӵİٷֱ.</zh>
   * <en>Returns the link percent.</en>
   *
   * @return double <zh>ٷֱ.</zh>
   * <en>The percent.</en>
   */
  public double getLinkPercent() {
    return this.linkPercent ;
  }

  /**
   * <zh>ӵİٷֱ.</zh>
   * <en>Sets the link percent.</en>
   *
   * @param percent double <zh>ٷֱ.</zh>
   * <en>The percent.</en>
   */
  public void setLinkPercent(double percent) {
    this.linkPercent = percent ;
  }

  /**
   * <zh>رǩ±߽yֵ.</zh>
   * <en>Returns the lower bound of the label.</en>
   *
   * @return double <zh>yֵ.</zh>
   * <en>The lower bound.</en>
   */
  public double getLowerY() {
    return this.allocatedY - this.labelHeight * 0.5 ;
  }

  /**
   * <zh>رǩϱ߽yֵ.</zh>
   * <en>Returns the upper bound of the label.</en>
   *
   * @return double <zh>yֵ.</zh>
   * <en>The upper bound.</en>
   */
  public double getUpperY() {
    return this.allocatedY + this.labelHeight * 0.5 ;
  }

  /**
   * <zh>ȽĴС.</zh>
   * <en>Compares this object to an arbitrary object.</en>
   *
   * @param obj  The object to compare against.
   *
   * @return An integer that specifies the relative order of the two objects.
   */
  public int compareTo(Object obj) {
    int result = 0 ;
    if (obj instanceof PieLabelRecord) {
      PieLabelRecord plr = (PieLabelRecord) obj ;
      if (this.baseY < plr.baseY) {
        result = -1 ;
      } else if (this.baseY > plr.baseY) {
        result = 1 ;
      }
    }

    return result ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof PieLabelRecord)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    PieLabelRecord that = (PieLabelRecord) obj ;
    if (!EqualsUtil.equals(this.key, that.key)) {
      return false ;
    }
    if (this.angle != that.angle) {
      return false ;
    }
    if (this.baseY != that.baseY) {
      return false ;
    }
    if (this.allocatedY != that.allocatedY) {
      return false ;
    }
    if (!EqualsUtil.equals(this.labelRect, that.labelRect)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.label, that.label)) {
      return false ;
    }
    if (this.labelHeight != that.labelHeight) {
      return false ;
    }
    if (this.gap != that.gap) {
      return false ;
    }
    if (this.linkPercent != that.linkPercent) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() ;
    result = 37 * result + HashcodeUtil.hashCode(this.key) ;
    result = 37 * result + this.angle ;
    result = 37 * result + this.baseY ;
    result = 37 * result + this.allocatedY ;
    result = 37 * result + HashcodeUtil.hashCode(this.labelRect) ;
    result = 37 * result + HashcodeUtil.hashCode(this.label) ;
    result = 37 * result + this.labelHeight ;
    result = 37 * result + this.gap ;
    result = 37 * result + this.linkPercent ;

    return (int) result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return this.label ;
  }

}
