package com.ijchart.xychart.render.xy ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.List ;

import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.LegendType ;
import com.ijchart.xychart.constant.SortType ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.data.IXYChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.exception.RowXValueDuplicateException ;
import com.ijchart.xychart.range.NumberRange ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.DatasetUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ƶջʽXY״ͼ.</zh>
 * <en>Draw xy stacked bar shapes.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYStackedBarRender extends AbstractXYRender implements Cloneable,
  Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -2566429058476610401L ;

  /**
   * <zh>ȱʡǷԶȵı־.</zh>
   * <en>A flag to indicate whether the width should be calculated automatically.</en>
   */
  public static final boolean DEFAULT_AUTO_WIDTH = true ;

  /**
   * <zh>ÿbarλõȱʡֵ.</zh>
   * <en>The default value between 0.0 and 1.0 that indicates the position of the x-value
   * within the interval.</en>
   */
  public static final double DEFAULT_INTERVAL_POSITION_FACTOR = 0.5D ;

  /**
   * <zh>̶ÿbarĿֵ.</zh>
   * <en>The fixed interval width.</en>
   */
  public static final double DEFAULT_FIXED_INTERVAL_WIDTH = 1.0D ;

  /**
   * <zh>ȱʡĿհ,λΪٷֱ,˲barĿ.</zh>
   * <en>The default percentage margin (to reduce the width of bars).</en>
   */
  public static final double DEFAULT_BAR_MARGIN = 0.1D ;

  /**
   * <zh>ȱʡǷظxֵ.</zh>
   * <en>A default flag that controls whether or not duplicate x-values are allowed.</en>
   */
  public static final boolean DEFAULT_ALLOW_DUPLICATE_X_VALUE = false ;

  /**
   * <zh>Ƿ԰ٷֱʽʾ.</zh>
   * <en>A flag that controls whether the bars display values or percentages.</en>
   */
  private boolean renderAsPercentages ;

  /**
   * <zh>ǷԶbarĿ.</zh>
   * <en>A flag to indicate whether the width should be calculated automatically.</en>
   */
  private boolean autoWidth ;

  /**
   * <zh>ÿbarλõĲֵ,[0.0,1.0]֮.</zh>
   * <en>A value between 0.0 and 1.0 that indicates the position of the x-value
   * within the interval.</en>
   */
  private double intervalPositionFactor ;

  /**
   * <zh>̶ÿbarĿֵ.</zh>
   * <en>The fixed interval width.</en>
   */
  private double fixedIntervalWidth ;

  /**
   * <zh>հ,λΪٷֱ,˲barĿ.</zh>
   * <en>Percentage margin (to reduce the width of bars).</en>
   */
  private double margin ;

  /**
   * <zh>Ƿظxֵ.</zh>
   * <en>A flag that controls whether or not duplicate x-values are allowed.</en>
   */
  private boolean allowDuplicateXValues ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public XYStackedBarRender(IMatrixDataset dataset) {
    this(dataset, false) ;
  }

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   * @param renderAsPercentages boolean <zh>Ƿ԰ٷֱʽʾ.</zh>
   * <en>A flag that controls whether the data values are rendered as percentages.</en>
   */
  public XYStackedBarRender(IMatrixDataset dataset, boolean renderAsPercentages) {
    super(dataset) ;

    this.renderAsPercentages = renderAsPercentages ;

    this.autoWidth = true ;
    this.intervalPositionFactor = DEFAULT_INTERVAL_POSITION_FACTOR ;
    this.fixedIntervalWidth = DEFAULT_FIXED_INTERVAL_WIDTH ;
    this.margin = DEFAULT_BAR_MARGIN ;
    this.allowDuplicateXValues = DEFAULT_ALLOW_DUPLICATE_X_VALUE ;
  }

  /**
   * <zh>԰ٷֱʽʾ,true,򷵻false.</zh>
   * <en>Returns true if the render displays each item value as a percentage
   * (so that the stacked bars add to 100%), and false otherwise.</en>
   *
   * @return A boolean. <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isRenderAsPercentages() {
    return this.renderAsPercentages ;
  }

  /**
   * <zh>Ƿ԰ٷֱʽʾ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the flag that controls whether the render displays each item
   * value as a percentage (so that the stacked bars add to 100%), and sends
   * a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param asPercentages <zh>ֵ.</zh>
   * <en>The flag.</en>
   */
  public void setRenderAsPercentages(boolean asPercentages) {
    if (this.renderAsPercentages != asPercentages) {
      this.renderAsPercentages = asPercentages ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ǷԶbarĿ.</zh>
   * <en>Returns whether or not the width should be calculated automatically.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isAutoWidth() {
    return this.autoWidth ;
  }

  /**
   * <zh>ǷԶbarĿ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the flag to indicate whether the width should be calculated automatically,and sends
   * a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param autoWidth boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setAutoWidth(boolean autoWidth) {
    if (this.autoWidth != autoWidth) {
      this.autoWidth = autoWidth ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ؼÿbarλõĲֵ.</zh>
   * <en>Returns the value that indicates the position of the x-value within the interval.</en>
   *
   * @return double <zh>ֵ.</zh>
   * <en>The factor.</en>
   */
  public double getIntervalPositionFactor() {
    return this.intervalPositionFactor ;
  }

  /**
   * <zh>üÿbarλõĲֵ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the value that indicates the position of the x-value within the interval,and sends
   * a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param factor double <zh>ֵ,[0.0,1.0]֮.</zh>
   * <en>The factor,must be in interval [0.0,1.0].</en>
   */
  public void setIntervalPositionFactor(double factor) {
    if (factor < 0 || factor > 1.0) {
      throw new IllegalArgumentException("Argument 'factor' must be in interval [0.0,1.0]") ;
    }

    if (this.intervalPositionFactor != factor) {
      this.intervalPositionFactor = factor ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ع̶ÿbarĿֵ.</zh>
   * <en>Returns the fixed interval width.</en>
   *
   * @return double <zh>ֵ.</zh>
   * <en>The width.</en>
   */
  public double getFixedIntervalWidth() {
    return this.fixedIntervalWidth ;
  }

  /**
   * <zh>ù̶ÿbarĿֵ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the fixed interval width,and sends
   * a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param fixedIntervalWidth double <zh>ֵ,0.</zh>
   * <en>The width,must be greater than 0.</en>
   */
  public void setFixedIntervalWidth(double fixedIntervalWidth) {
    if (this.fixedIntervalWidth != fixedIntervalWidth) {
      this.fixedIntervalWidth = fixedIntervalWidth ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ÿbarĿֵ.</zh>
   * <en>Returns the interval width.</en>
   *
   * @return double <zh>ֵ.</zh>
   * <en>The width.</en>
   */
  public double getIntervalWidth() {
    double intervalWidth = fixedIntervalWidth ;
    if (this.autoWidth) {
      intervalWidth = calculateInterval() ; //ÿϵеĿ
    }

    return intervalWidth ;
  }

  /**
   * <zh>ؿհֵ.</zh>
   * <en>Returns the margin which is a percentage amount by which the bars are trimmed.</en>
   *
   * @return double <zh>հֵ.</zh>
   * <en>The margin. </en>
   */
  public double getMargin() {
    return this.margin ;
  }

  /**
   * <zh>ÿհֵ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the percentage amount by which the bars are trimmed and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param margin double <zh>հֵ,ֵӦ[0.0,1.0).</zh>
   * <en>The new margin,this value should belong to the interval [0.0,1.0). </en>
   */
  public void setMargin(double margin) {
    if (margin < 0 || margin >= 1.0) {
      throw new IllegalArgumentException("Argument 'margin' must be in interval [0.0,1.0).") ;
    }

    if (this.margin != margin) {
      this.margin = margin ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>Ƿظxֵ.</zh>
   * <en>Returns the flag that controls whether or not duplicate x-values are allowed.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isAllowDuplicateXValues() {
    return this.allowDuplicateXValues ;
  }

  /**
   * <zh>Ƿظxֵ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the flag that controls whether or not duplicate x-values are allowed and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setAllowDuplicateXValues(boolean flag) {
    if (this.allowDuplicateXValues != flag) {
      this.allowDuplicateXValues = flag ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param xAxis AbstractValueAxis <zh>xֵ (Ϊnull).</zh>
   * <en>The x value axis (null not permitted).</en>
   * @param xAxisEdge RectangleEdge <zh>xλ(Ϊnull).</zh>
   * <en>The x axis edge.</en>
   * @param yAxis AbstractValueAxis <zh>yֵ (Ϊnull).</zh>
   * <en>The y value axis (null not permitted).</en>
   * @param yAxisEdge RectangleEdge <zh>yλ (Ϊnull).</zh>
   * <en>The y axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(AbstractValueAxis xAxis, RectangleEdge xAxisEdge,
                           AbstractValueAxis yAxis, RectangleEdge yAxisEdge,
                           Rectangle2D dataArea, ChartOrientation orientation,
                           RenderShapeCollection renderShapes, float alpha) {
    if (xAxis == null) {
      throw new IllegalArgumentException("Null 'xAxis' argument.") ;
    }
    if (xAxisEdge == null) {
      throw new IllegalArgumentException("Null 'xAxisEdge' argument.") ;
    }
    if (yAxis == null) {
      throw new IllegalArgumentException("Null 'yAxis' argument.") ;
    }
    if (yAxisEdge == null) {
      throw new IllegalArgumentException("Null 'yAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }

    if (!allowDuplicateXValues) {
      checkRowDataDuplicate() ; //ǷظXֵ
    }

    double intervalWidth = this.fixedIntervalWidth ;
    if (this.autoWidth) {
      intervalWidth = calculateInterval() ; //ÿϵеĿ
    }

    int rows = dataset.getRowCount() ;
    this.fillPaints = getFillPaints(rows) ;
    this.borderPaints = getBorderPaints(rows) ;
    this.borderStrokes = getBorderStrokes(rows) ;

    double rangeValue = 0 ;
    if (this.renderAsPercentages) {
      NumberRange range = DatasetUtil.getDatasetYRangeForStackBarRender(dataset) ;
      rangeValue = Math.abs(range.getUpperBound() - range.getLowerBound()) ;
      if (rangeValue == 0) {
        rangeValue = 1 ;
      }
    }

    //״ͼ
    for (int i = 0 ; i < rows ; i++) {
      int currRowCols = dataset.getColumnCount(i) ;
      for (int j = 0 ; j < currRowCols ; j++) {
        IChartData data = dataset.getData(i, j) ;
        if (data != null && data instanceof IXYChartData) {
          IXYChartData xyData = (IXYChartData) data ;
          double xValue = xyData.getX() ;
          double yValue = xyData.getY() ;
          if (Double.isNaN(xValue) || Double.isNaN(yValue)) {
            renderShapes.addDataItemShape(null) ;
            continue ;
          }

          double startXValue = xValue - this.intervalPositionFactor * intervalWidth ;
          double endXValue = xValue + (1.0 - this.intervalPositionFactor) * intervalWidth ;
          double startX = xAxis.valueToJava2D(dataArea, xAxisEdge, startXValue, false) ;
          double endX = xAxis.valueToJava2D(dataArea, xAxisEdge, endXValue, false) ;

          double yTotal = getYValueTotal(dataset, i, xValue, yValue) ;
          double startYValue = yTotal ;
          double endYValue = yTotal + yValue ;
          if (this.renderAsPercentages) {
            startYValue /= rangeValue ;
            startYValue *= 100 ;
            endYValue /= rangeValue ;
            endYValue *= 100 ;
          }

          double startY = yAxis.valueToJava2D(dataArea, yAxisEdge, startYValue, false) ;
          double endY = yAxis.valueToJava2D(dataArea, yAxisEdge, endYValue, false) ;

          //ֵ᷶Χֵ
          if (startX == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || endX == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || startY == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || endY == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
            renderShapes.addDataItemShape(null) ;
            continue ;
          }

          Rectangle2D rect = null ;
          if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
            double min = (yValue == 0) ? 0 : 1 ;
            double rectWidth = Math.abs(startX - endX) ; //ο
            double offsetX = this.margin * rectWidth * 0.5 ; //xƫֵ

            rect = new Rectangle2D.Double(Math.min(startX, endX) + offsetX,
                                          Math.min(startY, endY),
                                          Math.max(min, rectWidth - offsetX * 2),
                                          Math.abs(startY - endY)) ;
          } else if (orientation.equals(ChartOrientation.HORIZONTAL)) { //ˮƽ
            double min = (yValue == 0) ? 0 : 1 ;
            double rectHeight = Math.abs(startX - endX) ; //θ߶
            double offsetX = this.margin * rectHeight * 0.5 ; //xƫֵ

            rect = new Rectangle2D.Double(Math.min(startY, endY),
                                          Math.min(startX, endX) + offsetX,
                                          Math.abs(startY - endY),
                                          Math.max(min, rectHeight - offsetX * 2)) ;
          }

          IChartShape shape = createItemShape(orientation, yAxis, i, j, rect, fillPaints[i],
                                              borderPaints[i], borderStrokes[i], alpha, null) ;
          renderShapes.addDataItemShape(shape) ;
        } else {
          renderShapes.addDataItemShape(null) ;
        }
      }
    }

  }

  /**
   * <zh>ݼĳrow֮ǰyĺ,ЩݵxֵҪxValueͬ.</zh>
   *
   * @param dataset IMatrixDataset
   * @param row int
   * @param xValue double
   * @param yValue double
   * @return double
   */
  private static double getYValueTotal(IMatrixDataset dataset, int row, double xValue,
                                       double yValue) {
    double totalY = 0 ;
    int yValueSign = sign(yValue) ;

    for (int i = 0 ; i < row ; i++) {
      int currRowCols = dataset.getColumnCount(i) ;
      for (int j = 0 ; j < currRowCols ; j++) {
        IXYChartData xyData = (IXYChartData) dataset.getData(i, j) ;

        if (xyData != null) {
          double x = xyData.getX() ;
          double y = xyData.getY() ;
          if (!Double.isNaN(x) && !Double.isNaN(y) && x == xValue) {
            if (sign(y) == yValueSign) {
              totalY += y ;
            }
          }
        }

      }
    }

    return totalY ;
  }

  /**
   * <zh>һֵķ,ֵΪ-101,ֱʾС0,0,0.</zh>
   * @param value double
   * @return int
   */
  private static int sign(double value) {
    return (value < 0) ? -1 : ((value == 0) ? 0 : 1) ;
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Return the legend type.</en>
   *
   * @return LegendType <zh>ͼ.</zh>
   * <en>The legend type.</en>
   */
  public LegendType getLegendType() {
    return LegendType.ROW ;
  }

  /**
   * <zh>¼intervalֵ</zh>
   * <en>Recalculate the minimum width of interval.</en>
   *
   * @return double <zh>ֵ.</zh>
   * <en>The minimum width.</en>
   */
  private double calculateInterval() {
    double result = Double.POSITIVE_INFINITY ;
    IMatrixDataset dataset = getDataset() ;
    int rowCount = dataset.getRowCount() ;

    for (int i = 0 ; i < rowCount ; i++) {
      double minRowInterval = Double.POSITIVE_INFINITY ;
      List rowData = new ArrayList() ;
      int currRowCols = dataset.getColumnCount(i) ;

      for (int j = 0 ; j < currRowCols ; j++) {
        rowData.add(dataset.getData(i, j)) ;
      }

      dataset.sortRowData(i, SortType.ASCENDING) ;
      int itemCount = rowData.size() ;

      //Ѱҵһǿֵ
      double prevValue = Double.POSITIVE_INFINITY ;
      int prevPos = -1 ;
      for (int j = 0 ; j < itemCount ; j++) {
        IXYChartData data = (IXYChartData) rowData.get(j) ;
        if (data != null && !Double.isNaN(data.getX())) {
          prevValue = data.getX() ;
          prevPos = j ;
          break ;
        }
      }

      for (int j = prevPos + 1 ; j < itemCount ; j++) {
        IXYChartData data = (IXYChartData) rowData.get(j) ;
        if (data != null && !Double.isNaN(data.getX())) {
          minRowInterval = Math.min(minRowInterval, data.getX() - prevValue) ;
          prevValue = data.getX() ;
        }
      }

      result = Math.min(result, minRowInterval) ;
    }

    return result ;
  }

  /**
   * <zh>ݼеÿһǷظ.</zh>
   * <en>Check whether or not the x values are duplicate of a series.</en>
   */
  private void checkRowDataDuplicate() {
    IMatrixDataset dataset = getDataset() ;
    int rowCount = dataset.getRowCount() ;

    for (int i = 0 ; i < rowCount ; i++) {
      dataset.sortRowData(i, SortType.ASCENDING) ;
      double prevValue = Double.NaN ;
      int currRowCols = dataset.getColumnCount(i) ;

      for (int j = 0 ; j < currRowCols ; j++) {
        IXYChartData data = (IXYChartData) dataset.getData(i, j) ;
        if (data == null) {
          continue ;
        }

        double currValue = data.getX() ;
        if (!Double.isNaN(prevValue) && !Double.isNaN(currValue)) {
          if (prevValue == currValue) {
            String msg = "Dataset row " + i + " column " + j + " x value duplicated." ;
            throw new RowXValueDuplicateException(msg) ;
          }
        }
        prevValue = currValue ;
      }
    }
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates if some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof XYStackedBarRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    XYStackedBarRender that = (XYStackedBarRender) obj ;
    if (this.renderAsPercentages != that.renderAsPercentages) {
      return false ;
    }
    if (this.autoWidth != that.autoWidth) {
      return false ;
    }
    if (this.intervalPositionFactor != that.intervalPositionFactor) {
      return false ;
    }
    if (this.fixedIntervalWidth != that.fixedIntervalWidth) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double hashcode = super.hashCode() * 37 + HashcodeUtil.hashCode(this.renderAsPercentages) ;
    hashcode = hashcode * 37 + HashcodeUtil.hashCode(this.autoWidth) ;
    hashcode = hashcode * 37 + this.intervalPositionFactor * 10 ;
    hashcode = hashcode * 37 + this.fixedIntervalWidth * 10 ;

    return (int) hashcode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[renderAsPercentages=" + this.renderAsPercentages
      + ",autoWidth=" + this.autoWidth
      + ",intervalPositionFactor=" + this.intervalPositionFactor
      + ",fixedIntervalWidth=" + this.fixedIntervalWidth
      + ",allowDuplicateXValues=" + this.allowDuplicateXValues
      + "]" ;
  }

}
