package com.ijchart.xychart.shape ;

import java.awt.AlphaComposite ;
import java.awt.GradientPaint ;
import java.awt.Graphics2D ;
import java.awt.Paint ;
import java.awt.Shape ;
import java.awt.Stroke ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.Area ;
import java.awt.geom.PathIterator ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.Arrays ;

import javax.swing.event.EventListenerList ;

import com.ijchart.xychart.constant.ChartShapeDrawType ;
import com.ijchart.xychart.constant.GradientPaintTransformType ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.event.ChartShapeChangeEvent ;
import com.ijchart.xychart.listener.IChartShapeChangeListener ;
import com.ijchart.xychart.shape.change.IShapeChange ;
import com.ijchart.xychart.ui.DefaultGradientPaintTransformer ;
import com.ijchart.xychart.ui.Graphics2DState ;
import com.ijchart.xychart.ui.RectangleInsets ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.GraphicsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ṩ{@link IChartShape}ĳʵ.</zh>
 * <en>A base class for handling the drawing of {@link IChartShape}.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public abstract class AbstractChartShape implements IChartShape, Cloneable, Serializable,
  IChartShapeChangeListener {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 7385508126626790866L ;

  /**
   * <zh>ȱʡͼεк.</zh>
   * <en>The default shape id.</en>
   */
  public static final int DEFAULT_SHAPE_ID = -1 ;

  /**
   * <zh>ȱʡͼchartеĻ.</zh>
   * <en>The default chart shape draw type.</en>
   */
  public static final ChartShapeDrawType DEFAULT_DRAW_TYPE = ChartShapeDrawType.ANY_SHAPE ;

  /**
   * <zh>ȱʡǷɼ.</zh>
   * <en>A default flag controls whether or not the ChartShape can be selected.</en>
   */
  public static final boolean DEFAULT_VISIBLE = true ;

  /**
   * <zh>ڼǷĳͼ.</zh>
   * <en>The transformed shape for calculate whether or not contains a point.</en>
   */
  protected transient Shape transformedShape ;

  /**
   * <zh>Gradient Paint 任.</zh>
   * <en>Gradient Paint transform type.</en>
   */
  private GradientPaintTransformType gradientPaintTransformType ;

  /**
   * <zh>ʹõPaint.</zh>
   * <en>The fill Paint.</en>
   */
  private transient Paint fillPaint ;

  /**
   * <zh>䱳alphaֵ.</zh>
   * <en>The alpha value for fill background.</en>
   */
  private float fillAlpha ;

  /**
   * <zh>ʹõPaint.</zh>
   * <en>The border Paint.</en>
   */
  private transient Paint borderPaint ;

  /**
   * <zh>ͼѡʱêPaint.</zh>
   * <en>The anchor Paint of selected ChartShape. .</en>
   */
  private transient Paint selectAnchorPaint ;

  /**
   * <zh>ʹõStroke.</zh>
   * <en>The border Stroke.</en>
   */
  private transient Stroke borderStroke ;

  /**
   * <zh>Ƿ.</zh>
   * <en>Whether or not draw border.</en>
   */
  private boolean drawBorder ;

  /**
   * <zh>Ƿͼڲ.</zh>
   * <en>Whether or not fill ChartShape.</en>
   */
  private boolean fillShape ;

  /**
   * <zh>,ɱл.</zh>
   * <en>The listener collection,not serializable.</en>
   */
  private transient EventListenerList listenerList ;

  /**
   * <zh>תĽǶ(λΪ).</zh>
   * <en>The angle of rotation in radians.</en>
   */
  private double rotate ;

  /**
   * <zh>ͼǷɱѡ.</zh>
   * <en>A flag controls whether or not the ChartShape can be selected.</en>
   */
  private boolean selected ;

  /**
   * <zh>ԭʼtooltipͼ.</zh>
   * <en>The original tooltip shape.</en>
   */
  private transient Shape originTooltipShape ;

  /**
   * <zh>ѡͼʱêĿ.</zh>
   * <en>the anchor rectangle width of selected ChartShape.</en>
   */
  private double selectRectWidth ;

  /**
   * <zh>ѡͼʱêAlphaֵ.</zh>
   * <en>the anchor rectangle alpha value of selected ChartShape.</en>
   */
  private float selectRectAlpha ;

  /**
   * <zh>߿֮Ŀհ.</zh>
   * <en>The margin around the inside of the draw area.</en>
   */
  private RectangleInsets margin ;

  /**
   * <zh>tooltipַ.</zh>
   * <en>The tooltip text.</en>
   */
  private String tooltipText ;

  /**
   * <zh>URLַ.</zh>
   * <en>The URL text.</en>
   */
  private String url ;

  /**
   * <zh>ͼεı.</zh>
   * <en>The shape id.</en>
   */
  private int id ;

  /**
   * <zh>IChartData.</zh>
   * <en>The instance of IChartData.</en>
   */
  private IChartData data ;

  /**
   * <zh>ͼchartеĻ.</zh>
   * <en>The chart shape draw type.</en>
   */
  private ChartShapeDrawType drawType ;

  /**
   * <zh>Ƿɼ.</zh>
   * <en>A flag controls whether or not the ChartShape can be selected.</en>
   */
  private boolean visible ;

  /**
   * <zh>IShapeChange.</zh>
   * <en>The IShapeChange object.</en>
   */
  private IShapeChange shapeChange ;

  /**
   * <zh>ûвĹ.</zh>
   * <en>No argument constructor.</en>
   */
  protected AbstractChartShape() {
    this.fillPaint = DEFAULT_FILL_PAINT ;
    this.fillAlpha = DEFAULT_FILL_ALPHA ;
    this.gradientPaintTransformType = DEFAULT_GRADIENT_PAINT_TRANSFORM_TYPE ;
    this.borderPaint = DEFAULT_BORDER_PAINT ;
    this.borderStroke = DEFAULT_BORDER_STROKE ;
    this.selectAnchorPaint = DEFAULT_SELECT_FILL_PAINT ;
    this.drawBorder = DEFAULT_DRAW_BORDER ;
    this.fillShape = DEFAULT_FILL_SHAPE ;

    this.rotate = 0.0 ;
    this.selected = DEFAULT_SELECTED ;
    this.originTooltipShape = null ;
    this.selectRectWidth = DEFAULT_SELECT_RECT_WIDTH ;
    this.selectRectAlpha = DEFAULT_SELECT_RECT_ALPHA ;
    this.margin = DEFAULT_MARGIN ;
    this.id = DEFAULT_SHAPE_ID ;
    this.data = null ;
    this.drawType = DEFAULT_DRAW_TYPE ;
    this.visible = DEFAULT_VISIBLE ;

    this.shapeChange = null ;

    this.listenerList = new EventListenerList() ;
  }

  /**
   * <zh>ͼεı,һchart,ͼεIDΨһ.</zh>
   * <en>Sets an ID for this shape,the shape ID is unique and continuous in a chart.</en>
   * @param shapeId int
   */
  public void setId(int shapeId) {
    this.id = shapeId ;
  }

  /**
   * <zh>ͼεı.</zh>
   * <en>Returns the ID.</en>
   *
   * @return int <zh>ͼεı.</zh>
   * <en>The shape ID.</en>
   */
  public int getId() {
    return this.id ;
  }

  /**
   * <zh>ͼڲɫ.</zh>
   * <en>Get the ChartShape fill paint.</en>
   *
   * @return Paint <zh>ͼڲɫ.</zh>
   * <en>The ChartShape fill paint.</en>
   */
  public Paint getFillPaint() {
    return this.fillPaint ;
  }

  /**
   * <zh>ͼڲɫչ.</zh>
   * <en>Get the ChartShape GradientPaintTransformType type.</en>
   *
   * @return GradientPaintTransformType
   */
  public GradientPaintTransformType getGradientPaintTransformType() {
    return this.gradientPaintTransformType ;
  }

  /**
   * <zh>Paint.</zh>
   * <en>Get the ChartShape border Paint.</en>
   *
   * @return Paint <zh>Paint.</zh>
   * <en>The border Paint.</en>
   */
  public Paint getBorderPaint() {
    return this.borderPaint ;
  }

  /**
   * <zh>߻.</zh>
   * <en>Get the ChartShape border Stroke.</en>
   *
   * @return Stroke <zh>߻.</zh>
   * <en>The border Stroke.</en>
   */
  public Stroke getBorderStroke() {
    return this.borderStroke ;
  }

  /**
   * <zh>ǷƵı־.</zh>
   * <en>A flag that controls whether or not the shape border is visible.</en>
   *
   * @return boolean <zh>Ƿtrue,򷵻false.</zh>
   * <en>A boolean flag.</en>
   */
  public boolean isDrawBorder() {
    return this.drawBorder ;
  }

  /**
   * <zh>ǷƵı־.</zh>
   * <en>A flag that controls whether or not the shape is filled.</en>
   *
   * @return boolean <zh>Ƿtrue,򷵻false.</zh>
   * <en>A boolean flag.</en>
   */
  public boolean isFillShape() {
    return this.fillShape ;
  }

  /**
   * <zh>Ƿͼε,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets a flag that controls whether or not the shape border is visible and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>true,false.</zh>
   * <en>true,draw the border of the shape;false otherwise.</en>
   */
  public void setDrawBorder(boolean flag) {
    if (this.drawBorder != flag) {
      this.drawBorder = flag ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼڲɫ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the shape fill Paint and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>ͼڲõPaint(Ϊnull).</zh>
   * <en>The ChartShape fill Paint(null not permitted).</en>
   */
  public void setFillPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.fillPaint = paint ;
    fireChartShapeChanged() ;
  }

  /**
   * <zh>Ƿͼڲ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets a flag that controls whether or not the shape is filled,and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param bFillShape boolean <zh>trueͼڲ,falseͼڲ.</zh>
   * <en>true,fill the shape;false otherwise.</en>
   */
  public void setFillShape(boolean bFillShape) {
    if (this.fillShape != bFillShape) {
      this.fillShape = bFillShape ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼڲɫչ,PaintΪGradientPaint,
   * м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the ChartShape GradientPaintTransformType type,use only when fill Paint is
   *  GradientPaint,and sends a {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param transformType GradientPaintTransformType <zh>任 (Ϊnull).</zh>
   * <en>The gradient paint transform type (null permitted).</en>
   */
  public void setGradientPaintTransformType(GradientPaintTransformType transformType) {
    this.gradientPaintTransformType = transformType ;
    fireChartShapeChanged() ;
  }

  /**
   * <zh>ͼߵPaint,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the shape border Paint and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>ͼPaint(Ϊnull).</zh>
   * <en>The ChartShape border Paint(null not permitted).</en>
   */
  public void setBorderPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.borderPaint = paint ;
    fireChartShapeChanged() ;
  }

  /**
   * <zh>ͼߵĻ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the shape border Stroke and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>ͼ߻(Ϊnull).</zh>
   * <en>The ChartShape border Stroke(null not permitted).</en>
   */
  public void setBorderStroke(Stroke stroke) {
    if (stroke == null) {
      throw new IllegalArgumentException("Null 'stroke' argument.") ;
    }

    this.borderStroke = stroke ;
    fireChartShapeChanged() ;
  }

  /**
   * <zh>ͼλתǶ,λΪ,м{@link ChartShapeChangeEvent}¼.
   * 任ľʾ:
   * <pre>
   *		[   cos(theta)    -sin(theta)    0   ]
   *		[   sin(theta)     cos(theta)    0   ]
   *		[       0              0         1   ]
   * </pre>
   * .</zh>
   * <en>Concatenates the current Graphics2D Transform with a rotation
   *  transform,and sends a {@link ChartShapeChangeEvent} to all registered listeners.
   *  Subsequent rendering is rotated by the specified radians relative to the previous
   *  origin.This is equivalent to calling transform(R), where R is an
   *  AffineTransform represented by the following matrix:
   * <pre>
   *		[   cos(theta)    -sin(theta)    0   ]
   *		[   sin(theta)     cos(theta)    0   ]
   *		[       0              0         1   ]
   * </pre>
   * Rotating with a positive angle theta rotates points on the positive
   * x axis toward the positive y axis.</en>
   *
   * @param theta <zh>תĽǶ,λΪ.</zh>
   * <en>the angle of rotation in radians.</en>
   */
  public void setRotate(double theta) {
    if (this.rotate != theta) {
      this.rotate = theta ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼǷѡ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets a flag controls whether or not the shape is selected,and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setSelected(boolean flag) {
    if (this.selected != flag) {
      this.selected = flag ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼλתǶ,λΪ.</zh>
   * <en>Return the angle of rotation in radians.</en>
   *
   * @return double <zh>ͼλתǶ,λΪ.</zh>
   * <en>The angle of rotation in radians.</en>
   */
  public double getRotate() {
    return this.rotate ;
  }

  /**
   * <zh>ͼǷѡ.</zh>
   * <en>Returns whether or not this shape is selected.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isSelected() {
    return this.selected ;
  }

  /**
   * <zh>ԭʼtooltipͼ,ͼûпתضͼεӰ.</zh>
   * <en>Returns the original tooltip shape.</en>
   *
   * @return Shape <zh>ͼ (Ϊnull).</zh>
   * <en>A shape (null permitted).</en>
   */
  public Shape getOriginTooltipShape() {
    return this.originTooltipShape ;
  }

  /**
   * <zh>ԭʼtooltipͼ.</zh>
   * <en>Sets the original tooltip shape.</en>
   *
   * @param shape Shape <zh>ͼ (Ϊnull).</zh>
   * <en>A shape (null permitted).</en>
   */
  public void setOriginTooltipShape(Shape shape) {
    this.originTooltipShape = shape ;
  }

  /**
   * <zh>tooltipͼ,ͼҪתضͼεӰ.</zh>
   * <en>Returns the tooltip shape.</en>
   *
   * @return Shape <zh>ͼ (Ϊnull).</zh>
   * <en>A shape (null permitted).</en>
   */
  public Shape getTooltipShape() {
    if (this.originTooltipShape != null) {
      AffineTransform transform = new AffineTransform() ;

      Point2D translatePoint = getTooltipShapeTranslatePoint() ;
      float translateX = 0, translateY = 0 ;
      if (translatePoint != null) {
        translateX = (float) translatePoint.getX() ;
        translateY = (float) translatePoint.getY() ;
        transform.translate(translateX, translateY) ;
      }
      transform.rotate(getRotate()) ;

      Shape newShape = GraphicsUtil.translateShapeByPath(originTooltipShape, translateX, translateY) ;
      return transform.createTransformedShape(newShape) ;
    }

    return null ;
  }

  /**
   * <zh>tooltipͼӦƫƵλ.</zh>
   * <en>Returns the translate point of tooltip shape.</en>
   *
   * @return Point2D <zh> (Ϊnull).</zh>
   * <en>The translate point (null permitted).</en>
   */
  protected Point2D getTooltipShapeTranslatePoint() {
    return null ;
  }

  /**
   * <zh>tooltipͼӦתĽǶ.</zh>
   * <en>Returns the angle of rotation of the tooltip shape.</en>
   *
   * @return double <zh>Ƕ(λǻ).</zh>
   * <en>The angle of rotation (in radians).</en>
   */
  protected double getTooltipShapeRotate() {
    return 0 ;
  }

  /**
   * <zh>ѡͼʱêĿ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the anchor rectangle width of selected ChartShape,and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param width double <zh>,0.</zh>
   * <en>The width,must greater than 0.</en>
   */
  public void setSelectRectWidth(double width) {
    if (width <= 0) {
      throw new IllegalArgumentException("ARgument 'width' must be greater than 0.") ;
    }

    this.selectRectWidth = width ;
  }

  /**
   * <zh>ͼѡʱêPaint,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the anchor Paint of selected ChartShape,and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param selectAnchorPaint Paint <zh>Ϊnull.</zh>
   * <en>null not permitted.</en>
   */
  public void setSelectAnchorPaint(Paint selectAnchorPaint) {
    if (selectAnchorPaint == null) {
      throw new IllegalArgumentException("Null 'selectAnchorPaint' argument.") ;
    }
    this.selectAnchorPaint = selectAnchorPaint ;
  }

  /**
   * <zh>ѡͼʱêĿ.</zh>
   * <en>Returns the anchor rectangle width of selected ChartShape.</en>
   *
   * @return double <zh>,0.</zh>
   * <en>The width,must greater than 0.</en>
   */
  public double getSelectRectWidth() {
    return this.selectRectWidth ;
  }

  /**
   * <zh>ͼѡʱêPaint.</zh>
   * <en>Returns the anchor Paint of selected ChartShape.</en>
   *
   * @return Paint
   */
  public Paint getSelectAnchorPaint() {
    return this.selectAnchorPaint ;
  }

  /**
   * <zh>صpointǷͼ.</zh>
   * <en>Tests if a specified Point2D is inside the boundary of the ChartShape.</en>
   *
   * @param point Point2D <zh>ֵ,.(Ϊnull).</zh>
   * <en>The position,maybe a mouse point (null not permitted).</en>
   * @return boolean <zh>򷵻true,򷵻false.</zh>
   * <en>Returns true if this shape contains point(x,y),otherwise return false.</en>
   */
  public boolean contains(Point2D point) {
    if (point == null) {
      throw new IllegalArgumentException("Null 'point' argument.") ;
    }
    return contains(point.getX(), point.getY()) ;
  }

  /**
   * <zh>ص(x,y)Ƿͼ,߿С3,߿ᱻ3.</zh>
   * <en>Tests if a specified Point2D is inside the boundary of the ChartShape.</en>
   *
   * @param x double <zh>(xλ).</zh>
   * <en>the x position.</en>
   * @param y double <zh>(yλ).</zh>
   * <en>the y position.</en>
   * @return boolean <zh>򷵻true,򷵻false.</zh>
   * <en>Returns true if this shape contains point(x,y),otherwise return false.</en>
   */
  public boolean contains(double x, double y) {
    calculateTransformedShape() ;
    if (this.transformedShape != null) {
      return this.transformedShape.contains(x, y) ;
    }

    return false ;
  }

  /**
   * <zh>任ͼ.</zh>
   * <en>Calculate the transformed shape.</en>
   */
  protected abstract void calculateTransformedShape() ;

  /**
   * <zh>ַı߿֮Ŀհ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the margin around the inside of the block,and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param margin RectangleInsets <zh>ַı߿֮Ŀհ (Ϊnull).</zh>
   * <en>The margin around the inside of the block (null permitted).</en>
   */
  public void setMargin(RectangleInsets margin) {
    if (margin == null) {
      throw new IllegalArgumentException("Null 'margin' argument.") ;
    }
    this.margin = margin ;
    fireChartShapeChanged() ;
  }

  /**
   * <zh>ַı߿֮Ŀհ.</zh>
   * <en>Returns the margin around the inside of the block.</en>
   *
   * @return RectangleInsets <zh>ַı߿֮Ŀհ.</zh>
   * <en>The margin around the inside of the block.</en>
   */
  public RectangleInsets getMargin() {
    return this.margin ;
  }

  /**
   * <zh>ѡͼʱêAlphaֵ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the anchor rectangle alpha value of selected ChartShape,and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param selectRectAlpha float <zh>Alphaֵ,ڱ[0.0,1.0]ĸ.</zh>
   * <en>The alpha value,must be a floating point number in the inclusive range [0.0,1.0].</en>
   */
  public void setSelectRectAlpha(float selectRectAlpha) {
    if (!(selectRectAlpha >= 0 && selectRectAlpha <= 1.0)) {
      throw new IllegalArgumentException("Argument 'selectRectAlpha' must be between 0.0 and 1.0 .") ;
    }
    this.selectRectAlpha = selectRectAlpha ;
    fireChartShapeChanged() ;
  }

  /**
   * <zh>ѡͼʱêAlphaֵ.</zh>
   * <en>Returns the anchor rectangle alpha value of selected ChartShape.</en>
   *
   * @return float <zh>Alphaֵ,ڱ[0.0,1.0]ĸ.</zh>
   * <en>The alpha valuea floating point number in the inclusive range [0.0,1.0].</en>
   */
  public float getSelectRectAlpha() {
    return this.selectRectAlpha ;
  }

  /**
   * <zh>䱳alphaֵ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>The alpha value for fill background and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void setFillAlpha(float alpha) {
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException(
        "The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    if (this.fillAlpha != alpha) {
      this.fillAlpha = alpha ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>䱳alphaֵ.</zh>
   * <en>Returns the alpha value for fill background.</en>
   * @return float <zh>alphaֵ.</zh>
   * <en>The alpha value.</en>
   */
  public float getFillAlpha() {
    return this.fillAlpha ;
  }

  /**
   * <zh>tooltipַ.</zh>
   * <en>Sets the tooltip text.</en>
   *
   * @param tooltip String <zh>tooltipַ (Ϊnull).</zh>
   * <en>The tooltip text (null permitted).</en>
   */
  public void setTooltipText(String tooltip) {
    this.tooltipText = tooltip ;
  }

  /**
   * <zh>tooltipַ.</zh>
   * <en>Returns the tooltip text.</en>
   *
   * @return String <zh>tooltipַ (Ϊnull).</zh>
   * <en>The tooltip text (maybe null).</en>
   */
  public String getTooltipText() {
    return this.tooltipText ;
  }

  /**
   * <zh>URLַ.</zh>
   * <en>Sets the URL text.</en>
   *
   * @param url String <zh>URLַ (Ϊnull).</zh>
   * <en>The URL text (null permitted).</en>
   */
  public void setURL(String url) {
    this.url = url ;
  }

  /**
   * <zh>URLַ.</zh>
   * <en>Returns the URL text.</en>
   *
   * @return String <zh>URLַ (Ϊnull).</zh>
   * <en>The URL text (maybe null).</en>
   */
  public String getURL() {
    return this.url ;
  }

  /**
   * <zh>ͼε߿.</zh>
   * <en>Returns a Rectangle2D that completely encloses the Shape.</en>
   *
   * @return Rectangle2D <zh>߿.</zh>
   * <en>A Rectangle2D that completely encloses the Shape.</en>
   */
  public Rectangle2D getBound() {
    calculateTransformedShape() ;
    if (transformedShape != null) {
      return transformedShape.getBounds2D() ;
    }

    return null ;
  }

  /**
   * <zh>Ʊ.</zh>
   * <en>Draw background color.</en>
   *
   * @param g2 Graphics2D <zh>ͼλ.</zh>
   * <en>The specified Graphics context.</en>
   * @param drawArea Shape <zh>ƷΧ.</zh>
   * <en>The draw area.</en>
   */
  protected void drawBackground(Graphics2D g2, Shape drawArea) {
    if (isFillShape() && drawArea != null) {
      Paint fill = getFillPaint() ;
      if (fill instanceof GradientPaint && getGradientPaintTransformType() != null) {
        DefaultGradientPaintTransformer transformer = new DefaultGradientPaintTransformer(
          getGradientPaintTransformType()) ;
        fill = transformer.transform((GradientPaint) fill, drawArea) ;
      }

      Graphics2DState state = new Graphics2DState(g2) ;
      state.saveOldState() ;

      g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getFillAlpha())) ;
      g2.setPaint(fill) ;
      g2.fill(drawArea) ;

      state.restoreOldState() ;
    }
  }

  /**
   * <zh>߿.</zh>
   * <en>Draw border.</en>
   *
   * @param g2 Graphics2D <zh>ͼλ.</zh>
   * <en>The specified Graphics context.</en>
   * @param drawArea Shape <zh>ƷΧ.</zh>
   * <en>The draw area.</en>
   */
  protected void drawBorder(Graphics2D g2, Shape drawArea) {
    if (isDrawBorder() && drawArea != null) { //
      Graphics2DState state = new Graphics2DState(g2) ;
      state.saveOldState() ;

      if (getBorderStroke() != null) {
        g2.setStroke(getBorderStroke()) ;
      }

      g2.setPaint(getBorderPaint()) ;
      g2.draw(drawArea) ;

      state.restoreOldState() ;
    }
  }

  /**
   * <zh>ʱͼεĵ.</zh>
   * <en>Returns the lock coordinates of this shape.</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getLockPoint() {
    return null ;
  }

  /**
   * <zh>IChartData.</zh>
   * <en>Returns the IChartData object.</en>
   *
   * @return IChartData <zh>(Ϊnull).</zh>
   * <en>(maybe null).</en>
   */
  public IChartData getData() {
    return this.data ;
  }

  /**
   * <zh>IChartData,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the IChartData object and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param data IChartData <zh>(Ϊnull).</zh>
   * <en>(null permitted).</en>
   */
  public void setData(IChartData data) {
    if (!(this.data == null && data == null)) {
      this.data = data ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼǷɼ.</zh>
   * <en>Returns whether or not the shape is visible.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isVisible() {
    return this.visible ;
  }

  /**
   * <zh>ͼǷɼ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets whether or not the shape is visible and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setVisible(boolean flag) {
    if (this.visible != flag) {
      this.visible = flag ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼchartеĻ.</zh>
   * <en>Returns the chart shape draw type.</en>
   *
   * @return ChartShapeDrawType <zh>(Ϊnull).</zh>
   * <en>(never null).</en>
   */
  public ChartShapeDrawType getShapeDrawType() {
    return this.drawType ;
  }

  /**
   * <zh>ͼchartеĻ,м{@link ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the chart shape draw type and sends a
   * {@link ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param drawType ChartShapeDrawType <zh>(null not permitted).</zh>
   * <en>(null not permitted).</en>
   */
  public void setShapeDrawType(ChartShapeDrawType drawType) {
    if (drawType == null) {
      throw new IllegalArgumentException("Null 'drawType' argument") ;
    }

    this.drawType = drawType ;
  }

  /**
   * <zh>עͼ{@link IChartShape}仯ļ߶.</zh>
   * <en>Registers an object for notification of changes to the {@link IChartShape}.</en>
   *
   * @param listener <zh>עļ.</zh>
   * <en>the object to register.</en>
   */
  public void addChangeListener(IChartShapeChangeListener listener) {
    listenerList.add(IChartShapeChangeListener.class, listener) ;
  }

  /**
   * <zh>ɾעͼ{@link IChartShape}仯߶.</zh>
   * <en>Deregisters an object for notification of changes to the {@link IChartShape}.</en>
   *
   * @param listener <zh>Ҫɾļ߶.</zh>
   * <en>the object to deregister.</en>
   */
  public void removeChangeListener(IChartShapeChangeListener listener) {
    listenerList.remove(IChartShapeChangeListener.class, listener) ;
  }

  /**
   * <zh>еͼ{@link IChartShape}仯߶.</zh>
   * <en>Deregisters all object for notification of changes to the {@link IChartShape}.</en>
   */
  public void removeAllChangeListener() {
    Object[] listeners = this.listenerList.getListenerList() ;

    for (int i = listeners.length - 2 ; i >= 0 ; i -= 2) {
      this.listenerList.remove(IChartShapeChangeListener.class,
                               (IChartShapeChangeListener) listeners[i + 1]) ;
    }
  }

  /**
   * <zh>ָlistenerǷ,ҪڵԪ.</zh>
   * <en>Returns true if the specified object is registered with
   * the {@link IChartShape} as a listener.  Most applications won't need to call this
   * method, it exists mainly for use by unit testing code.</en>
   *
   * @param listener <zh>ҪԵļ.</zh>
   * <en>the listener.</en>
   * @return A boolean.
   */
  public boolean hasListener(IChartShapeChangeListener listener) {
    return Arrays.asList(this.listenerList.getListenerList()).contains(listener) ;
  }

  /**
   * <zh>ͼζڲı,ֲܴ¼,ͨ˷֪ͨͼζı.</zh>
   * <en>If the data object of {@link com.ijchart.xychart.shape.IChartShape} changed and no event
   * triggered,set {@link com.ijchart.xychart.shape.IChartShape} changed by this method.</en>
   */
  public void forceChartShapeChanged() {
    fireChartShapeChanged() ;
  }

  /**
   * <zh>յͼζ仯֪ͨ.</zh>
   * <en>Receives notification of an {@link com.ijchart.xychart.shape.IChartShape} change event.</en>
   *
   * @param event <zh>ͼζ仯Ķ.</zh>
   * <en>Information about the event.</en>
   */
  public void chartShapeChanged(ChartShapeChangeEvent event) {
    this.transformedShape = null ;
    notifyListeners(event) ;
  }

  /**
   * <zh>{@link IShapeChange}.</zh>
   * <en>Sets the {@link IShapeChange} object.</en>
   *
   * @param change IShapeChange <zh>{@link IShapeChange} (Ϊnull).</zh>
   * <en>{@link IShapeChange} object (null permitted).</en>
   */
  public void setShapeChange(IShapeChange change) {
    if (!(this.shapeChange == null && change == null)) {
      this.shapeChange = change ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>γȵ.</zh>
   * <en>Returns a string containing the coordinates for a given shape.This
   * string is intended for use in an image map.</en>
   *
   * @param parentClipArea Shape <zh>ͼμ (Ϊnull).</zh>
   * <en>The clip area.</en>
   *
   * @return String <zh>ȵ (Ϊnull).</zh>
   * <en>The coordinates of hotspot (maybe null).</en>
   */
  public String getHotspotPolyCoords(Shape parentClipArea) {
    Shape shape = getHotspotShape() ;
    if (shape == null) {
      return "" ;
    }

    if (shape instanceof Rectangle2D && parentClipArea == null) {
      Rectangle2D rect = (Rectangle2D) shape ;
      int x1 = (int) rect.getX() ;
      int y1 = (int) rect.getY() ;
      int x2 = x1 + (int) rect.getWidth() ;
      int y2 = y1 + (int) rect.getHeight() ;
      if (x2 == x1) {
        x2++ ;
      }
      if (y2 == y1) {
        y2++ ;
      }

      return x1 + "," + y1 + "," + x2 + "," + y2 ;
    }

    Area hotspotArea = new Area(shape) ;
    if (parentClipArea != null) {
      hotspotArea.intersect(new Area(parentClipArea)) ;
    }

    float lastX = 0, lastY = 0 ;
    StringBuffer result = new StringBuffer() ;
    boolean first = true ;
    float[] coords = new float[6] ;

    PathIterator pi = hotspotArea.getPathIterator(null) ;
    while (!pi.isDone()) {
      pi.currentSegment(coords) ;
      if (first) {
        first = false ;
        result.append((int) coords[0]) ;
        result.append(",").append((int) coords[1]) ;

        lastX = coords[0] ;
        lastY = coords[1] ;
      } else {
        if (!(lastX == coords[0] && lastY == coords[1])) {
          result.append(",") ;
          result.append((int) coords[0]) ;
          result.append(",") ;
          result.append((int) coords[1]) ;

          lastX = coords[0] ;
          lastY = coords[1] ;
        }
      }
      pi.next() ;
    }

    return result.toString() ;
  }

  /**
   * <zh>γȵͼ.</zh>
   * <en>Returns the shape type of hotspot.</en>
   *
   * @return String <zh>ͼ,ֵpolyrect.</zh>
   * <en>The shape type of hotspot,the value can be poly or rect.</en>
   */
  public String getHotspotType() {
    return HOTSPOT_TYPE_POLYGON ;
  }

  /**
   * <zh>ȵͼ.</zh>
   * <en>Returns the hotspot shape.</en>
   *
   * @return Shape <zh>ȵͼ (Ϊnull).</zh>
   * <en>The hotspot shape (maybe null).</en>
   */
  protected abstract Shape getHotspotShape() ;

  /**
   * <zh>{@link IShapeChange}.</zh>
   * <en>Returns the {@link IShapeChange} object.</en>
   *
   * @return IShapeChange <zh>{@link IShapeChange} (Ϊnull).</zh>
   * <en>{@link IShapeChange} object (maybe null).</en>
   */
  public IShapeChange getShapeChange() {
    return this.shapeChange ;
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻͼεѡв,Ϊͼοܻص,ԻͼʱҪȻеͼκ,
   * ٵô˷ѡв.</zh>
   * <en>Draw the selected ChartShape on a Java 2d graphics device(such as the screen or the
   * printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>the graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  protected abstract void drawSelect(Graphics2D g2) ;

  /**
   * <zh>ͨмͼη˱仯.</zh>
   * <en>Notifies all registered listeners that the chart shape has changed.</en>
   */
  protected void fireChartShapeChanged() {
    this.transformedShape = null ;
    notifyListeners(new ChartShapeChangeEvent(this, this)) ;
  }

  /**
   * <zh>֪ͨעļͼη˱仯.</zh>
   * <en>Notifies all registered listeners that the chart shape has changed.</en>
   *
   * @param event <zh>¼.</zh>
   * <en>contains information about the event that triggered the notification.</en>
   */
  protected void notifyListeners(ChartShapeChangeEvent event) {
    Object[] listeners = this.listenerList.getListenerList() ;
    for (int i = listeners.length - 2 ; i >= 0 ; i -= 2) {
      if (listeners[i] == IChartShapeChangeListener.class) {
        ((IChartShapeChangeListener) listeners[i + 1]).chartShapeChanged(event) ;
      }
    }
  }

  /**
   * <zh>ݵê.</zh>
   * <en>Draw anchor by the coordinates of points.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull).</zh>
   * <en>the graphics device (such as the screen or the printer) (null not permitted).</en>
   * @param translateX double <zh>xƫֵ,ĵ㶼.</zh>
   * <en>The x offset value, the other point coordinates are relative to this coordinate.</en>
   * @param translateY double <zh>yƫֵ,ĵ㶼.</zh>
   * <en>The y offset value, the other point coordinates are relative to this coordinate.</en>
   * @param points Double[] <zh>êֵ.</zh>
   * <en>The coordinates of draw anchor.</en>
   */
  protected void drawAnchorByPoints(Graphics2D g2, double translateX, double translateY,
                                    Point2D.Double[] points) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }

    if (isSelected() && points != null) {
      Graphics2DState state = new Graphics2DState(g2) ;
      state.saveOldState() ;
      g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getSelectRectAlpha())) ;
      g2.translate(translateX, translateY) ;
      g2.rotate(getRotate()) ;
      g2.setPaint(getSelectAnchorPaint()) ;

      double rectWidth = getSelectRectWidth() ;
      double halfRect = 0.5 * rectWidth ;
      for (int i = 0 ; i < points.length ; i++) {
        if (points[i] != null) {
          double pointX = points[i].getX() ;
          double pointY = points[i].getY() ;
          Rectangle2D anchor = new Rectangle2D.Double(pointX - halfRect, pointY - halfRect,
                                                      rectWidth, rectWidth) ;
          g2.fill(anchor) ;
        }
      }

      g2.rotate( -getRotate()) ;
      state.restoreOldState() ;
    }

  }

  /**
   * <zh>ͼ.</zh>
   * <en>Fill this shape.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull).</zh>
   * <en>the graphics device (such as the screen or the printer) (null not permitted).</en>
   * @param fillPaint Paint <zh>õpaint (Ϊnull).</zh>
   * <en>The fill paint (null not permitted).</en>
   * @param shape Shape <zh>Ҫͼ (Ϊnull).</zh>
   * <en>The fill shape (null not permitted).</en>
   */
  protected void fillShape(Graphics2D g2, Paint fillPaint, Shape shape) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (fillPaint == null) {
      throw new IllegalArgumentException("Null 'fillPaint' argument.") ;
    }
    if (shape == null) {
      throw new IllegalArgumentException("Null 'shape' argument.") ;
    }

    if (isFillShape()) {
      Graphics2DState state = new Graphics2DState(g2) ;
      state.saveOldState() ;

      g2.setPaint(fillPaint) ;
      g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getFillAlpha())) ;
      g2.fill(shape) ;

      state.restoreOldState() ;
    }
  }

  /**
   * <zh>ͼα߿.</zh>
   * <en>Outline this shape.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull).</zh>
   * <en>the graphics device (such as the screen or the printer) (null not permitted).</en>
   * @param shape Shape <zh>Ҫͼ (Ϊnull).</zh>
   * <en>The fill shape (null not permitted).</en>
   */
  protected void drawShapeBorder(Graphics2D g2, Shape shape) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (shape == null) {
      throw new IllegalArgumentException("Null 'shape' argument.") ;
    }

    if (isDrawBorder()) {
      Graphics2DState state = new Graphics2DState(g2) ;
      state.saveOldState() ;

      g2.setPaint(getBorderPaint()) ;
      g2.draw(shape) ;
      state.restoreOldState() ;
    }
  }

  /**
   * <zh>{@link IShapeChange}beforeDraw.</zh>
   * <en>Call the beforeDraw method of class {@link IShapeChange}.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull).</zh>
   * <en>The graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  protected void beforeDraw(Graphics2D g2) {
    if (this.shapeChange != null) {
      this.shapeChange.beforeDraw(this, g2) ;
    }
  }

  /**
   * <zh>{@link IShapeChange}afterDraw.</zh>
   * <en>Call the afterDraw method of class {@link IShapeChange}.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull).</zh>
   * <en>The graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  protected void afterDraw(Graphics2D g2) {
    if (this.shapeChange != null) {
      this.shapeChange.afterDraw(this, g2) ;
    }
  }

  /**
   * <zh>γȵͼ.</zh>
   * <en>Returns the shape type of hotspot.</en>
   *
   * @param shape Shape <zh>ͼ (Ϊnull).</zh>
   * <en>The shape (null permitted).</en>
   *
   * @return String <zh>ͼ,ֵcircpolyrect.</zh>
   * <en>The shape type of hotspot,the value can be circ or poly or rect.</en>
   */
  protected String getHotspotType(Shape shape) {
    if (shape == null) {
      return "" ;
    }

    if (shape instanceof Rectangle2D) {
      return "rect" ;
    } else {
      return "poly" ;
    }
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates if some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ.</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }

    if (!(obj instanceof AbstractChartShape)) {
      return false ;
    }

    AbstractChartShape that = (AbstractChartShape) obj ;

    if (!EqualsUtil.equals(this.gradientPaintTransformType, that.gradientPaintTransformType)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.fillPaint, that.fillPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.borderPaint, that.borderPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.borderStroke, that.borderStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.selectAnchorPaint, that.selectAnchorPaint)) {
      return false ;
    }
    if (!(this.drawBorder == that.drawBorder)) {
      return false ;
    }
    if (!(this.fillShape == that.fillShape)) {
      return false ;
    }
    if (!(this.rotate == that.rotate)) {
      return false ;
    }
    if (!(this.selected == that.selected)) {
      return false ;
    }
    if (this.selectRectWidth != that.selectRectWidth) {
      return false ;
    }
    if (!EqualsUtil.equals(this.margin, that.margin)) {
      return false ;
    }
    if (this.selectRectAlpha != that.selectRectAlpha) {
      return false ;
    }
    if (this.fillAlpha != that.fillAlpha) {
      return false ;
    }
    if (!EqualsUtil.equals(this.data, that.data)) {
      return false ;
    }
    if (this.visible != that.visible) {
      return false ;
    }
    if (!EqualsUtil.equals(this.shapeChange, that.shapeChange)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = HashcodeUtil.hashCode(this.gradientPaintTransformType) ;
    result = 37 * result + HashcodeUtil.hashCode(this.fillPaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.borderPaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.borderStroke) ;
    result = 37 * result + HashcodeUtil.hashCode(this.drawBorder) ;
    result = 37 * result + HashcodeUtil.hashCode(this.fillShape) ;
    result = 37 * result + HashcodeUtil.hashCode(this.selected) ;
    result = 37 * result + HashcodeUtil.hashCode(this.selectAnchorPaint) ;
    result = 37 * result + (int)this.rotate * 10 ;
    result = 37 * result + (int)this.selectRectWidth * 10 ;
    result = 37 * result + HashcodeUtil.hashCode(this.margin) ;
    result = 37 * result + (int)this.selectRectAlpha * 100 ;
    result = 37 * result + (int)this.fillAlpha * 100 ;
    result = 37 * result + HashcodeUtil.hashCode(this.data) ;
    result = 37 * result + HashcodeUtil.hashCode(this.visible) ;
    result = 37 * result + HashcodeUtil.hashCode(this.shapeChange) ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    AbstractChartShape clone = (AbstractChartShape)super.clone() ;
    clone.listenerList = new EventListenerList() ;

    return clone ;
  }

  /**
   * <zh>л.</zh>
   * <en>Provides serialization support.</en>
   * @param stream <zh>.</zh>
   * <en>the output stream.</en>
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writePaint(stream, this.fillPaint) ;
    SerialUtil.writePaint(stream, this.borderPaint) ;
    SerialUtil.writeStroke(stream, this.borderStroke) ;
    SerialUtil.writePaint(stream, this.selectAnchorPaint) ;
    SerialUtil.writeShape(stream, this.originTooltipShape) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.fillPaint = SerialUtil.readPaint(stream) ;
    this.borderPaint = SerialUtil.readPaint(stream) ;
    this.borderStroke = SerialUtil.readStroke(stream) ;
    this.selectAnchorPaint = SerialUtil.readPaint(stream) ;
    this.originTooltipShape = SerialUtil.readShape(stream) ;

    this.listenerList = new EventListenerList() ;
    this.transformedShape = null ;
  }

}
