package com.ijchart.xychart.shape ;

import java.awt.Graphics2D ;
import java.awt.Shape ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.constant.ShapeType ;
import com.ijchart.xychart.ui.Graphics2DState ;
import com.ijchart.xychart.util.EqualsUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ƹͼ.</zh>
 * <en>Draw a I-shaped {@link com.ijchart.xychart.shape.IChartShape}.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class IShapedChartShape extends AbstractChartShape implements Serializable, Cloneable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -1992891407282426360L ;

  /**
   * <zh>½ǵĵX.</zh>
   * <en>The x coordinate of left-hand corner.</en>
   */
  private double leftX ;

  /**
   * <zh>½ǵĵY.</zh>
   * <en>The y coordinate of left-hand corner.</en>
   */
  private double bottomY ;

  /**
   * <zh>εĿ.</zh>
   * <en>The width of I-shaped shape.</en>
   */
  private double width ;

  /**
   * <zh>εĸ߶.</zh>
   * <en>The height of I-shaped shape.</en>
   */
  private double height ;

  /**
   * <zh>ƹεͼ.</zh>
   * <en>Draw a I-shaped chart shape.</en>
   *
   * @param leftX double <zh>½ǵĵX.</zh>
   * <en>The x coordinate of left-hand corner.</en>
   * @param bottomY double <zh>½ǵĵY.</zh>
   * <en>The y coordinate of left-hand corner.</en>
   * @param width double <zh>εĿ.</zh>
   * <en>The width of I-shaped shape.</en>
   * @param height double <zh>εĸ߶.</zh>
   * <en>The height of I-shaped shape.</en>
   */
  public IShapedChartShape(double leftX, double bottomY, double width, double height) {
    this.leftX = leftX ;
    this.bottomY = bottomY ;
    this.width = width ;
    this.height = height ;
  }

  /**
   * <zh>ƹεͼ.</zh>
   * <en>Draw a I-shaped chart shape.</en>
   *
   * @param bottomLeft Point2D <zh>½ (Ϊnull).</zh>
   * <en>The point of left-hand corner (null not permitted).</en>
   * @param width double <zh>εĿ.</zh>
   * <en>The width of I-shaped shape.</en>
   * @param height double <zh>εĸ߶.</zh>
   * <en>The height of I-shaped shape.</en>
   */
  public IShapedChartShape(Point2D bottomLeft, double width, double height) {
    if (bottomLeft == null) {
      throw new IllegalArgumentException("Null 'bottomLeft' argument.") ;
    }

    this.leftX = bottomLeft.getX() ;
    this.bottomY = bottomLeft.getY() ;
    this.width = width ;
    this.height = height ;
  }

  /**
   * <zh>½ǵĵX.</zh>
   * <en>Returns the x coordinate of left-hand corner.</en>
   *
   * @return double
   */
  public double getLeftX() {
    return this.leftX ;
  }

  /**
   * <zh>½ǵĵX,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the x coordinate of left-hand corner,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param leftX double
   */
  public void setLeftX(double leftX) {
    if (this.leftX != leftX) {
      this.leftX = leftX ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>½ǵĵY.</zh>
   * <en>Returns the y coordinate of left-hand corner.</en>
   *
   * @return double
   */
  public double getBottomY() {
    return this.bottomY ;
  }

  /**
   * <zh>½ǵĵY,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the y coordinate of left-hand corner,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param bottomY double
   */
  public void setBottomY(double bottomY) {
    if (this.bottomY != bottomY) {
      this.bottomY = bottomY ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ùεĿ.</zh>
   * <en>Sets the width of I-shaped shape.</en>
   *
   * @return double
   */
  public double getWidth() {
    return this.width ;
  }

  /**
   * <zh>ùεĿ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the width of I-shaped shape,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param width double
   */
  public void setWidth(double width) {
    if (this.width != width) {
      this.width = width ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>عεĸ߶.</zh>
   * <en>Returns the height of I-shaped shape.</en>
   *
   * @return double
   */
  public double getHeight() {
    return this.height ;
  }

  /**
   * <zh>ùεĸ߶,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the height of I-shaped shape,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param height double
   */
  public void setHeight(double height) {
    if (this.height != height) {
      this.height = height ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻƴͼ</zh>
   * <en>Draw the ChartShape on a Java 2d graphics device(such as the screen or the printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>The graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  public void draw(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (!isVisible()) {
      return ;
    }

    Graphics2DState state = new Graphics2DState(g2) ;
    state.saveOldState() ;

    g2.translate(this.leftX, this.bottomY) ;
    g2.rotate(getRotate()) ;
    beforeDraw(g2) ;

    g2.setStroke(getBorderStroke()) ;
    g2.setPaint(getBorderPaint()) ;

    GeneralPath path = new GeneralPath() ;
    path.moveTo(0, 0) ;
    path.lineTo((float)this.width, 0) ;
    path.lineTo((float) (0.5 * this.width), 0) ;
    path.lineTo((float) (0.5 * this.width), (float) - this.height) ;
    path.lineTo(0, (float) - this.height) ;
    path.lineTo((float)this.width, (float) - this.height) ;
    path.lineTo((float) (0.5 * this.width), (float) - this.height) ;
    path.lineTo((float) (0.5 * this.width), 0) ;
    path.closePath() ;

    g2.draw(path) ;

    afterDraw(g2) ;
    g2.rotate( -getRotate()) ;
    state.restoreOldState() ;
    drawSelect(g2) ;
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻͼεѡв,Ϊͼοܻص,ԻͼʱҪȻеͼκ,
   * ٵô˷ѡв.</zh>
   * <en>Draw the selected ChartShape on a Java 2d graphics device(such
   * as the screen or the printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>the graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  protected void drawSelect(Graphics2D g2) {
    if (isSelected()) {
      double halfWidth = 0.5 * this.width ;
      boolean drawCenter = this.width > 2 * getSelectRectWidth() ;

      Point2D.Double[] points = new Point2D.Double[] {
                                new Point2D.Double(0, 0),
                                drawCenter ? new Point2D.Double(halfWidth, 0) : null,
                                new Point2D.Double(this.width, 0),

                                new Point2D.Double(0, -this.height),
                                drawCenter ? new Point2D.Double(halfWidth, -this.height) : null,
                                new Point2D.Double(this.width, -this.height)} ;

      drawAnchorByPoints(g2, this.leftX, this.bottomY, points) ;
    }
  }

  /**
   * <zh>ƶͼ,X᷽ƫtx,Y᷽ƫty,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Concatenates the current Graphics2D Transform with a translation transform.and sends
   * a {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.
   * Subsequent rendering is translated by the specified
   * distance relative to the previous position.
   * This is equivalent to calling transform(T), where T is an
   * AffineTransform represented by the following matrix:
   * <pre>
   *		[   1    0    tx  ]
   *		[   0    1    ty  ]
   *		[   0    0    1   ]
   * </pre>.</en>
   *
   * @param tx double <zh>X᷽ľ.</zh>
   * <en>The distance to translate along the x-axis.</en>
   * @param ty double <zh>Y᷽ľ.</zh>
   * <en>The distance to translate along the y-axis.</en>
   */
  public void translate(double tx, double ty) {
    if (!(tx == 0 && ty == 0)) {
      this.leftX += tx ;
      this.bottomY += ty ;

      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼϽǵ(ͼεת).</zh>
   * <en>Returns the top left corner coordinates of this shape (not consider rotation).</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getTopLeftPoint() {
    return new Point2D.Double(this.leftX, this.bottomY - this.height) ;
  }

  /**
   * <zh>ʱͼεĵ.</zh>
   * <en>Returns the lock coordinates of this shape.</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getLockPoint() {
    GeneralPath path = new GeneralPath() ;
    path.moveTo(0, 0) ;
    path.lineTo((float)this.width, 0) ;
    path.lineTo((float) (0.5 * this.width), 0) ;
    path.lineTo((float) (0.5 * this.width), (float) - this.height) ;
    path.lineTo(0, (float) - this.height) ;
    path.lineTo((float)this.width, (float) - this.height) ;
    path.lineTo((float) (0.5 * this.width), (float) - this.height) ;
    path.lineTo((float) (0.5 * this.width), 0) ;
    path.closePath() ;

    AffineTransform transform = AffineTransform.getTranslateInstance(this.leftX, this.bottomY) ;
    transform.rotate(getRotate()) ;
    Shape rotateShape = transform.createTransformedShape(path) ;
    Rectangle2D rotateBound = rotateShape.getBounds2D() ;

    return new Point2D.Double(rotateBound.getCenterX(), rotateBound.getCenterY()) ;
  }

  /**
   * <zh>任ͼ.</zh>
   * <en>Calculate the transformed shape.</en>
   */
  protected void calculateTransformedShape() {
    if (transformedShape == null) {
      float halfWidth = (float) (0.5 * this.width) ;
      float fWidth = (float)this.width ;
      float fHeight = (float)this.height ;
      final float expandWidth = 1.5F ;

      GeneralPath path = new GeneralPath() ;
      path.moveTo(0, -expandWidth) ;
      path.lineTo(halfWidth - expandWidth, -expandWidth) ;
      path.lineTo(halfWidth - expandWidth, -fHeight + expandWidth) ;
      path.lineTo(0, -fHeight + expandWidth) ;
      path.lineTo(0, -fHeight - expandWidth) ;
      path.lineTo(fWidth, -fHeight - expandWidth) ;
      path.lineTo(fWidth, -fHeight + expandWidth) ;
      path.lineTo(halfWidth + expandWidth, -fHeight + expandWidth) ;
      path.lineTo(halfWidth + expandWidth, -expandWidth) ;
      path.lineTo(fWidth, -expandWidth) ;
      path.lineTo(fWidth, expandWidth) ;
      path.lineTo(0, expandWidth) ;
      path.closePath() ;

      AffineTransform transform = new AffineTransform() ;
      transform.translate(this.leftX, this.bottomY) ;
      transform.rotate(getRotate()) ;
      transformedShape = transform.createTransformedShape(path) ;
    }
  }

  /**
   * <zh>ȵͼ.</zh>
   * <en>Returns the hotspot shape.</en>
   *
   * @return Shape <zh>ȵͼ (Ϊnull).</zh>
   * <en>The hotspot shape (maybe null).</en>
   */
  protected Shape getHotspotShape() {
    calculateTransformedShape() ;
    return transformedShape ;
  }

  /**
   * <zh>ͼε.</zh>
   * <en>The type for describe this shape.</en>
   *
   * @return ShapeType <zh>{@link ShapeType}ඨĳ (Ϊnull).</zh>
   * <en>The constant value in class {@link ShapeType}.</en>
   */
  public ShapeType getShapeType() {
    return ShapeType.SHAPE_I_SHAPED ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates whether some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false
   *  otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }
    if (!(obj instanceof IShapedChartShape)) {
      return false ;
    }

    final double theta = 0.00001 ;
    IShapedChartShape that = (IShapedChartShape) obj ;
    if (!EqualsUtil.equals(this.leftX, that.leftX, theta)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.bottomY, that.bottomY, theta)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.width, that.width, theta)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.height, that.height, theta)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double hashCode = super.hashCode() ;
    hashCode = 37 * hashCode + this.leftX ;
    hashCode = 37 * hashCode + this.bottomY ;
    hashCode = 37 * hashCode + this.width ;
    hashCode = 37 * hashCode + this.height ;

    return (int) hashCode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object. </en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[leftX=" + this.leftX
      + ",bottomY=" + this.bottomY
      + ",width=" + this.width
      + ",height=" + this.height
      + "]" ;
  }

}
