package com.ijchart.xychart.shape ;

import java.awt.AlphaComposite ;
import java.awt.Graphics2D ;
import java.awt.Shape ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.Line2D ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;

import com.ijchart.xychart.constant.ShapeType ;
import com.ijchart.xychart.constant.VerticalAlignment ;
import com.ijchart.xychart.ui.Graphics2DState ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;
import com.ijchart.xychart.util.StrokeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>һͼκһֱɵĸͼ.</zh>
 * <en>A chart shape composited by a line and a shape.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class LineAndShapeChartShape extends AbstractRectangleTextChartShape implements Cloneable,
  Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 4428700269943560757L ;

  /**
   * <zh>СΪ0ľ.</zh>
   * <en>A constant representing zero rectangle.</en>
   */
  public static final Rectangle2D ZERO_RECT = new Rectangle2D.Double(0, 0, 0, 0) ;

  /**
   * <zh>ͼƫԭX.</zh>
   * <en>The x coordinate offset of the shape and the line from origin.</en>
   */
  private double offsetX ;

  /**
   * <zh>ͼƫԭY.</zh>
   * <en>The y coordinate offset of the shape and the line from origin.</en>
   */
  private double offsetY ;

  /**
   * <zh>ͼζ.</zh>
   * <en>A shape.</en>
   */
  private transient Shape shape ;

  /**
   * <zh>һֱ.</zh>
   * <en>A line.</en>
   */
  private transient Line2D line ;

  /**
   * <zh>һͼκһֱɵĸͼ.</zh>
   * <en>A shape composited by a line and a shape.</en>
   *
   * @param offsetX <zh>ͼƫԭX.</zh>
   * <en>The x coordinate offset of the shape and the line from origin.</en>
   * @param offsetY <zh>ͼƫԭY.</zh>
   * <en>The y coordinate offset of the shape and the line from origin.</en>
   * @param shape Shape <zh>ͼζ (Ϊnull).</zh>
   * <en>A shape (null permitted).</en>
   * @param line Line2D <zh> (Ϊnull).</zh>
   * <en>A line (null permitted).</en>
   */
  public LineAndShapeChartShape(double offsetX, double offsetY, Shape shape, Line2D line) {
    this(offsetX, offsetY, shape, line, null) ;
  }

  /**
   * <zh>һͼκһֱɵĸͼ.</zh>
   * <en>A shape composited by a line and a shape.</en>
   *
   * @param offsetX <zh>ͼƫԭX.</zh>
   * <en>The x coordinate offset of the shape and the line from origin.</en>
   * @param offsetY <zh>ͼƫԭY.</zh>
   * <en>The y coordinate offset of the shape and the line from origin.</en>
   * @param shape Shape <zh>ͼζ (Ϊnull).</zh>
   * <en>A shape (null permitted).</en>
   * @param line Line2D <zh> (Ϊnull).</zh>
   * <en>A line (null permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   */
  public LineAndShapeChartShape(double offsetX, double offsetY,
                                Shape shape, Line2D line, String text) {
    super((shape == null) ? ZERO_RECT : shape.getBounds2D(), text) ;

    this.offsetX = offsetX ;
    this.offsetY = offsetY ;
    this.shape = shape ;
    this.line = line ;
    setTextVerticalAlign(VerticalAlignment.CENTER) ;
  }

  /**
   * <zh>Xƫֵ.</zh>
   * <en>Returns the x-offset.</en>
   *
   * @return double <zh>Xƫֵ.</zh>
   * <en>The x-offset.</en>
   */
  public double getOffsetX() {
    return this.offsetX ;
  }

  /**
   * <zh>Xƫֵ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the x-offset,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param offsetX double <zh>Xƫֵ.</zh>
   * <en>The x-offset.</en>
   */
  public void setOffsetX(double offsetX) {
    if (this.offsetX != offsetX) {
      this.offsetX = offsetX ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>Yƫֵ.</zh>
   * <en>Returns the y-offset.</en>
   *
   * @return double <zh>Yƫֵ.</zh>
   * <en>The y-offset.</en>
   */
  public double getOffsetY() {
    return this.offsetY ;
  }

  /**
   * <zh>Yƫֵ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the y-offset,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param offsetY double <zh>Yƫֵ.</zh>
   * <en>The y-offset.</en>
   */
  public void setOffsetY(double offsetY) {
    if (this.offsetY != offsetY) {
      this.offsetY = offsetY ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼζ.</zh>
   * <en>Returns the shape.</en>
   *
   * @return Shape <zh>ͼζ (Ϊnull).</zh>
   * <en>A shape (maybe null).</en>
   */
  public Shape getShape() {
    return this.shape ;
  }

  /**
   * <zh>ͼζ,м{@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the shape,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param shape Shape <zh>ͼζ (Ϊnull).</zh>
   * <en>A shape (null permitted).</en>
   */
  public void setShape(Shape shape) {
    if (!(this.shape == null && shape == null)) {
      this.shape = shape ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh> (Ϊnull).</zh>
   * <en>A line (null permitted).</en>
   *
   * @return Line2D <zh> (Ϊnull).</zh>
   * <en>A line (maybe null).</en>
   */
  public Line2D getLine() {
    return this.line ;
  }

  /**
   * <zh>,м{@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the line,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param line Line2D <zh> (Ϊnull).</zh>
   * <en>A line (null permitted).</en>
   */
  public void setLine(Line2D line) {
    if (!(this.line == null && line == null)) {
      this.line = line ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻƴͼ</zh>
   * <en>Draw the ChartShape on a Java 2d graphics device(such as the screen or the printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>The graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  public void draw(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (!isVisible()) {
      return ;
    }

    Graphics2DState state = new Graphics2DState(g2) ;
    state.saveOldState() ;

    if (this.line != null) {
      g2.translate(this.offsetX, this.offsetY) ;
      g2.rotate(getRotate()) ;
      beforeDraw(g2) ;
      //    g2.setClip(cloneArea) ;

      drawBorder(g2, this.line) ;
      afterDraw(g2) ;
      g2.rotate( -getRotate()) ;
    }

    if (this.shape != null) {
      state.restoreTransform() ;
      g2.translate(this.offsetX, this.offsetY) ;
      g2.rotate(getRotate()) ;
      beforeDraw(g2) ;
      //    g2.setClip(cloneArea) ;

      drawBackground(g2, this.shape) ;
      drawBorder(g2, this.shape) ;
      afterDraw(g2) ;
      g2.rotate( -getRotate()) ;
    }

    drawText(g2, new Rectangle2D.Double(this.offsetX, this.offsetY, 0, 0)) ;
    state.restoreOldState() ;
    drawSelect(g2) ;
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻͼεѡв,Ϊͼοܻص,ԻͼʱҪȻеͼκ,
   * ٵô˷ѡв.</zh>
   * <en>Draw the selected ChartShape on a Java 2d graphics device(such as the screen or the
   * printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>the graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  protected void drawSelect(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }

    if (isSelected() && this.line != null) {
      double x1 = this.line.getX1() ;
      double y1 = this.line.getY1() ;
      double x2 = this.line.getX2() ;
      double y2 = this.line.getY2() ;

      Graphics2DState state = new Graphics2DState(g2) ;
      state.saveOldState() ;

      g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getSelectRectAlpha())) ;

      double rectWidth = getSelectRectWidth() ;
      final double halfRect = rectWidth / 2 ;

      g2.setPaint(getSelectAnchorPaint()) ;
      g2.translate(this.offsetX + x1, this.offsetY + y1) ;
      g2.rotate(getRotate()) ;

      //ê
      if (y1 == y2) { //ˮƽ
        g2.fill(new Rectangle2D.Double(0, -halfRect, rectWidth, rectWidth)) ;
        g2.fill(new Rectangle2D.Double(x2 - x1 - rectWidth, -halfRect,
                                       rectWidth, rectWidth)) ;
      } else if (x1 == x2) { //ֱ
        g2.fill(new Rectangle2D.Double( -halfRect, 0, rectWidth, rectWidth)) ;
        g2.fill(new Rectangle2D.Double( -halfRect, y2 - y1 - rectWidth,
                                       rectWidth, rectWidth)) ;
      } else {
        //ˮƽֱê㾭ƫתбê
        double theta = Math.atan((y2 - y1) / (x2 - x1)) ; //бߵƫת
        if (y1 > y2 && x1 > x2) {
          theta -= Math.PI ;
        }

        double lineLen = Math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)) ; //֮ľ
        g2.rotate(theta) ;

        g2.fill(new Rectangle2D.Double(0, -halfRect, rectWidth, rectWidth)) ;
        g2.fill(new Rectangle2D.Double(lineLen - rectWidth, -halfRect, rectWidth, rectWidth)) ;

        g2.rotate( -theta) ;
      }

      g2.rotate( -getRotate()) ;
      state.restoreOldState() ;
    }

  }

  /**
   * <zh>任ͼ.</zh>
   * <en>Calculate the transformed shape.</en>
   */
  protected void calculateTransformedShape() {
    if (transformedShape == null) {
      AffineTransform transform = new AffineTransform() ;
      transform.translate(this.offsetX, this.offsetY) ;
      transform.rotate(getRotate()) ;

      GeneralPath path = new GeneralPath() ;
      if (this.shape != null) {
        path.append(this.shape.getPathIterator(null), false) ;
      }
      if (this.line != null) {
        Shape shape = StrokeUtil.lineToRectangle(line, getBorderStroke(), 0) ;
        path.append(shape, false) ;
      }

      transformedShape = transform.createTransformedShape(path) ;
    }
  }

  /**
   * <zh>ƶͼ,X᷽ƫtx,Y᷽ƫty,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Concatenates the current Graphics2D Transform with a translation transform.and sends
   * a {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.
   * Subsequent rendering is translated by the specified
   * distance relative to the previous position.
   * This is equivalent to calling transform(T), where T is an
   * AffineTransform represented by the following matrix:
   * <pre>
   *		[   1    0    tx  ]
   *		[   0    1    ty  ]
   *		[   0    0    1   ]
   * </pre>.</en>
   *
   * @param tx double <zh>X᷽ľ.</zh>
   * <en>The distance to translate along the x-axis.</en>
   * @param ty double <zh>Y᷽ľ.</zh>
   * <en>The distance to translate along the y-axis.</en>
   */
  public void translate(double tx, double ty) {
    if (!(tx == 0 && ty == 0)) {
      if (this.shape != null) {
        AffineTransform transform = AffineTransform.getTranslateInstance(tx, ty) ;
        this.shape = transform.createTransformedShape(this.shape) ;
      }

      if (this.line != null) {
        this.line.setLine(this.line.getX1() + tx, this.line.getY1() + ty,
                          this.line.getX2() + tx, this.line.getY2() + ty) ;
      }

      fireChartShapeChanged() ;
    }

  }

  /**
   * <zh>ͼϽǵ(ͼεת).</zh>
   * <en>Returns the top left corner coordinates of this shape (not consider rotation).</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getTopLeftPoint() {
    if (this.shape != null && this.line == null) {
      Rectangle2D rect = this.shape.getBounds2D() ;
      return new Point2D.Double(rect.getMinX() + this.offsetX, rect.getMinY() + this.offsetY) ;
    } else if (this.shape == null && this.line != null) {
      return new Point2D.Double(Math.min(this.line.getX1(), this.line.getX2()) + this.offsetX,
                                Math.min(this.line.getY1(), this.line.getY2()) + this.offsetY) ;
    } else if (this.shape != null && this.line != null) {
      Rectangle2D rect1 = this.shape.getBounds2D() ;
      Rectangle2D rect2 = this.line.getBounds2D() ;
      return new Point2D.Double(Math.min(rect1.getMinX(), rect2.getMinX()) + this.offsetX,
                                Math.min(rect1.getMinY(), rect2.getMinY()) + this.offsetY) ;
    }

    return null ;
  }

  /**
   * <zh>ʱͼεĵ.</zh>
   * <en>Returns the lock coordinates of this shape.</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getLockPoint() {
    Shape lockShape = null ;
    if (this.shape != null) {
      lockShape = this.shape ;
    }

    if (lockShape == null && this.line != null) {
      lockShape = this.line ;
    }

    if (lockShape == null) {
      return null ;
    }

    AffineTransform transform = AffineTransform.getTranslateInstance(this.offsetX, this.offsetY) ;
    transform.rotate(getRotate()) ;
    Shape rotateShape = transform.createTransformedShape(lockShape) ;
    Rectangle2D rotateBound = rotateShape.getBounds2D() ;

    return new Point2D.Double(rotateBound.getCenterX(), rotateBound.getCenterY()) ;
  }

  /**
   * <zh>ȵͼ.</zh>
   * <en>Returns the hotspot shape.</en>
   *
   * @return Shape <zh>ȵͼ (Ϊnull).</zh>
   * <en>The hotspot shape (maybe null).</en>
   */
  protected Shape getHotspotShape() {
    calculateTransformedShape() ;
    return transformedShape ;
  }

  /**
   * <zh>ͼε.</zh>
   * <en>The type for describe this shape.</en>
   * @return ShapeType <zh>{@link com.ijchart.xychart.constant.ShapeType}ඨĳ
   * (Ϊnull).</zh>
   * <en>The constant value in class {@link com.ijchart.xychart.constant.ShapeType}.</en>
   */
  public ShapeType getShapeType() {
    return ShapeType.SHAPE_LINE_AND_SHAPE ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates whether some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument;
   * false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return false ;
    }
    if (!(obj instanceof LineAndShapeChartShape)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    LineAndShapeChartShape that = (LineAndShapeChartShape) obj ;
    if (this.offsetX != that.offsetX) {
      return false ;
    }
    if (this.offsetY != that.offsetY) {
      return false ;
    }
    if (!EqualsUtil.equals(this.shape, that.shape)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.line, that.line)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double hashCode = super.hashCode() ;
    hashCode = 37 * hashCode + this.offsetX ;
    hashCode = 37 * hashCode + this.offsetY ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.shape) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.line) ;

    return (int) hashCode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object. </en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[offsetX=" + this.offsetX + ",offsetY=" + this.offsetY
      + ",shape=" + this.shape + ",line=" + this.line + "]" ;
  }

  /**
   * <zh>л</zh>
   * <en>Provides serialization support</en>
   * @param stream <zh></zh>
   * <en>the output stream</en>
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;

    SerialUtil.writeShape(stream, this.shape) ;
    SerialUtil.writeShape(stream, this.line) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;

    this.shape = SerialUtil.readShape(stream) ;
    this.line = (Line2D) SerialUtil.readShape(stream) ;
  }
}
