package com.ijchart.xychart.shape ;

import java.awt.AlphaComposite ;
import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Graphics2D ;
import java.awt.Paint ;
import java.awt.Shape ;
import java.awt.Stroke ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.Ellipse2D ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.Line2D ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;

import com.ijchart.xychart.constant.ShapeType ;
import com.ijchart.xychart.ui.Graphics2DState ;
import com.ijchart.xychart.util.AnchorUtil ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;
import com.ijchart.xychart.util.StrokeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>Сֵֵͼ.</zh>
 * <en>Draw a min and max value {@link com.ijchart.xychart.shape.IChartShape}.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class MinMaxChartShape extends AbstractChartShape implements Serializable, Cloneable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 3642632434285996119L ;

  /**
   * <zh>ȱʡͼδС.</zh>
   * <en>The default shape size.</en>
   */
  private static final double DEFAULT_SHAPE_SIZE = 4 ;

  /**
   * <zh>ȱʡСֵӦͼ.</zh>
   * <en>The default minimum value shape.</en>
   */
  public static final Shape DEFAULT_MIN_SHAPE =
    new Ellipse2D.Double( -DEFAULT_SHAPE_SIZE, -DEFAULT_SHAPE_SIZE, 2 * DEFAULT_SHAPE_SIZE,
                         2 * DEFAULT_SHAPE_SIZE) ;

  /**
   * <zh>ȱʡĴֵӦͼ.</zh>
   * <en>The default maximum value shape.</en>
   */
  public static final Shape DEFAULT_MAX_SHAPE =
    new Ellipse2D.Double( -DEFAULT_SHAPE_SIZE, -DEFAULT_SHAPE_SIZE, 2 * DEFAULT_SHAPE_SIZE,
                         2 * DEFAULT_SHAPE_SIZE) ;

  /**
   * <zh>ȱʡĻСֵͼεpaint.</zh>
   * <en>The default paint used to draw minimum value shape.</en>
   */
  public static final Paint DEFAULT_MIN_SHAPE_PAINT = Color.BLACK ;

  /**
   * <zh>ȱʡĻƴֵͼεpaint.</zh>
   * <en>The default paint used to draw maximum value shape.</en>
   */
  public static final Paint DEFAULT_MAX_SHAPE_PAINT = Color.BLACK ;

  /**
   * <zh>ȱʡĻߵpaint.</zh>
   * <en>The default paint used to draw group line.</en>
   */
  public static final Paint DEFAULT_GROUP_LINE_PAINT = Color.BLACK ;

  /**
   * <zh>ȱʡĻСֵͼεstroke.</zh>
   * <en>The default stroke used to draw minimum value shape.</en>
   */
  public static final Stroke DEFAULT_MIN_SHAPE_STROKE = new BasicStroke(1.0F) ;

  /**
   * <zh>ȱʡĻƴֵͼεstroke.</zh>
   * <en>The default stroke used to draw maximum value shape.</en>
   */
  public static final Stroke DEFAULT_MAX_SHAPE_STROKE = new BasicStroke(1.0F) ;

  /**
   * <zh>ȱʡĻߵstroke.</zh>
   * <en>The default stroke used to draw group line.</en>
   */
  public static final Stroke DEFAULT_GROUP_LINE_STROKE = new BasicStroke(1.0F) ;

  /**
   * <zh>ȱʡǷ.</zh>
   * <en>A default flag controls whether or not draw group line.</en>
   */
  public static final boolean DEFAULT_DRAW_GROUP_LINE = true ;

  /**
   * <zh>СֵӦͼ.</zh>
   * <en>The minimum value shape.</en>
   */
  private transient Shape minShape ;

  /**
   * <zh>ֵӦͼ.</zh>
   * <en>The maximum value shape.</en>
   */
  private transient Shape maxShape ;

  /**
   * <zh>Сֵͼεpaint.</zh>
   * <en>The paint used to draw minimum value shape.</en>
   */
  private transient Paint minShapePaint ;

  /**
   * <zh>ƴֵͼεpaint.</zh>
   * <en>The paint used to draw maximum value shape.</en>
   */
  private transient Paint maxShapePaint ;

  /**
   * <zh>ߵpaint.</zh>
   * <en>The paint used to draw group line.</en>
   */
  private transient Paint groupLinePaint ;

  /**
   * <zh>Сֵͼεstroke.</zh>
   * <en>The stroke used to draw minimum value shape.</en>
   */
  private transient Stroke minShapeStroke ;

  /**
   * <zh>ƴֵͼεstroke.</zh>
   * <en>The stroke used to draw maximum value shape.</en>
   */
  private transient Stroke maxShapeStroke ;

  /**
   * <zh>ߵstroke.</zh>
   * <en>The stroke used to draw group line.</en>
   */
  private transient Stroke groupLineStroke ;

  /**
   * <zh>Сֵͼελ.</zh>
   * <en>The position of minimum value shape.</en>
   */
  private Point2D minShapePos ;

  /**
   * <zh>ֵͼελ.</zh>
   * <en>The position of maximum value shape.</en>
   */
  private Point2D maxShapePos ;

  /**
   * <zh>Ƿ.</zh>
   * <en>A flag controls whether or not draw group line.</en>
   */
  private boolean drawGroupLine ;

  /**
   * <zh>ͼ.</zh>
   * <en>Constructor.</en>
   *
   * @param minShapePos Point2D <zh>Сֵͼελ (Ϊnull).</zh>
   * <en>The position of minimum value shape (null permitted).</en>
   * @param maxShapePos Point2D <zh>ֵͼελ (Ϊnull).</zh>
   * <en>The position of maximum value shape (null permitted).</en>
   */
  public MinMaxChartShape(Point2D minShapePos, Point2D maxShapePos) {
    this(minShapePos, maxShapePos, DEFAULT_MIN_SHAPE, DEFAULT_MAX_SHAPE) ;
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Constructor.</en>
   *
   * @param minShapePos Point2D <zh>Сֵͼελ (Ϊnull).</zh>
   * <en>The position of minimum value shape (null permitted).</en>
   * @param maxShapePos Point2D <zh>ֵͼελ (Ϊnull).</zh>
   * <en>The position of maximum value shape (null permitted).</en>
   * @param minShape Shape <zh>СֵӦͼ (Ϊnull).</zh>
   * <en>The minimum value shape (null permitted).</en>
   * @param maxShape Shape <zh>ֵӦͼ (Ϊnull).</zh>
   * <en>The maximum value shape (null permitted).</en>
   */
  public MinMaxChartShape(Point2D minShapePos, Point2D maxShapePos, Shape minShape, Shape maxShape) {
    this.minShapePos = minShapePos ;
    this.maxShapePos = maxShapePos ;
    this.minShape = minShape ;
    this.maxShape = maxShape ;

    this.minShapePaint = DEFAULT_MIN_SHAPE_PAINT ;
    this.maxShapePaint = DEFAULT_MAX_SHAPE_PAINT ;
    this.groupLinePaint = DEFAULT_GROUP_LINE_PAINT ;
    this.minShapeStroke = DEFAULT_GROUP_LINE_STROKE ;
    this.maxShapeStroke = DEFAULT_MAX_SHAPE_STROKE ;
    this.groupLineStroke = DEFAULT_GROUP_LINE_STROKE ;

    this.drawGroupLine = DEFAULT_DRAW_GROUP_LINE ;
  }

  /**
   * <zh>СֵӦͼ.</zh>
   * <en>Returns the minimum value shape.</en>
   *
   * @return Shape <zh>СֵӦͼ (Ϊnull).</zh>
   * <en>The minimum value shape (maybe null).</en>
   */
  public Shape getMinShape() {
    return this.minShape ;
  }

  /**
   * <zh>СֵӦͼ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the minimum value shape and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param shape Shape <zh>СֵӦͼ (Ϊnull).</zh>
   * <en>The minimum value shape (null permitted).</en>
   */
  public void setMinShape(Shape shape) {
    if (!(this.minShape == null && shape == null)) {
      this.minShape = shape ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>شֵӦͼ.</zh>
   * <en>Returns the maximum value shape.</en>
   *
   * @return Shape <zh>ֵӦͼ (Ϊnull).</zh>
   * <en>The maximum value shape (maybe null).</en>
   */
  public Shape getMaxShape() {
    return this.maxShape ;
  }

  /**
   * <zh>ôֵӦͼ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the maximum value shape and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param shape Shape <zh>ֵӦͼ (Ϊnull).</zh>
   * <en>The maximum value shape (null permitted).</en>
   */
  public void setMaxShape(Shape shape) {
    if (!(this.maxShape == null && shape == null)) {
      this.maxShape = shape ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػСֵͼεpaint.</zh>
   * <en>Returns the paint used to draw minimum value shape.</en>
   *
   * @return Paint <zh>Сֵͼεpaint (Ϊnull).</zh>
   * <en>The paint used to draw minimum value shape (maybe null).</en>
   */
  public Paint getMinShapePaint() {
    return this.minShapePaint ;
  }

  /**
   * <zh>ûСֵͼεpaint,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the paint used to draw minimum value shape and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Сֵͼεpaint (Ϊnull).</zh>
   * <en>The paint used to draw minimum value shape (null permitted).</en>
   */
  public void setMinShapePaint(Paint paint) {
    if (!(this.minShapePaint == null && paint == null)) {
      this.minShapePaint = paint ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػƴֵͼεpaint.</zh>
   * <en>Returns the paint used to draw maximum value shape.</en>
   *
   * @return Paint <zh>ƴֵͼεpaint (Ϊnull).</zh>
   * <en>The paint used to draw maximum value shape (maybe null).</en>
   */
  public Paint getMaxShapePaint() {
    return this.maxShapePaint ;
  }

  /**
   * <zh>ûƴֵͼεpaint,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the paint used to draw maximum value shape and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>ƴֵͼεpaint (Ϊnull).</zh>
   * <en>The paint used to draw maximum value shape (null permitted).</en>
   */
  public void setMaxShapePaint(Paint paint) {
    if (!(this.maxShapePaint == null && paint == null)) {
      this.maxShapePaint = paint ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػߵpaint.</zh>
   * <en>Returns the paint used to draw group line.</en>
   *
   * @return Paint <zh>ߵpaint (Ϊnull).</zh>
   * <en>The paint used to draw group line (maybe null).</en>
   */
  public Paint getGroupLinePaint() {
    return this.groupLinePaint ;
  }

  /**
   * <zh>ûߵpaint,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the paint used to draw group line and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>ߵpaint (Ϊnull).</zh>
   * <en>The paint used to draw group line (null permitted).</en>
   */
  public void setGroupLinePaint(Paint paint) {
    if (!(this.groupLinePaint == null && paint == null)) {
      this.groupLinePaint = paint ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػСֵͼεstroke.</zh>
   * <en>Returns the stroke used to draw minimum value shape.</en>
   *
   * @return Stroke <zh>Сֵͼεstroke (Ϊnull).</zh>
   * <en>The stroke used to draw minimum value shape (maybe null).</en>
   */
  public Stroke getMinShapeStroke() {
    return this.minShapeStroke ;
  }

  /**
   * <zh>ûСֵͼεstroke,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the stroke used to draw minimum value shape and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>Сֵͼεstroke (Ϊnull).</zh>
   * <en>The stroke used to draw minimum value shape (null permitted).</en>
   */
  public void setMinShapeStroke(Stroke stroke) {
    if (!(this.minShapeStroke == null && stroke == null)) {
      this.minShapeStroke = stroke ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػƴֵͼεstroke.</zh>
   * <en>Returns the stroke used to draw maximum value shape.</en>
   *
   * @return Stroke <zh>ƴֵͼεstroke (Ϊnull).</zh>
   * <en>The stroke used to draw maximum value shape (maybe null).</en>
   */
  public Stroke getMaxShapeStroke() {
    return this.maxShapeStroke ;
  }

  /**
   * <zh>ûƴֵͼεstroke,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the stroke used to draw maximum value shape and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>ƴֵͼεstroke (Ϊnull).</zh>
   * <en>The stroke used to draw maximum value shape (null permitted).</en>
   */
  public void setMaxShapeStroke(Stroke stroke) {
    if (!(this.maxShapeStroke == null && stroke == null)) {
      this.maxShapeStroke = stroke ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػߵstroke.</zh>
   * <en>Returns the stroke used to draw group line.</en>
   *
   * @return Stroke <zh>ߵstroke (Ϊnull).</zh>
   * <en>The stroke used to draw group line (maybe null).</en>
   */
  public Stroke getGroupLineStroke() {
    return this.groupLineStroke ;
  }

  /**
   * <zh>ûߵstroke,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the stroke used to draw group line and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>ߵstroke (Ϊnull).</zh>
   * <en>The stroke used to draw group line (null permitted).</en>
   */
  public void setGroupLineStroke(Stroke stroke) {
    if (!(this.groupLineStroke == null && stroke == null)) {
      this.groupLineStroke = stroke ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>Сֵͼελ.</zh>
   * <en>Returns the position of minimum value shape.</en>
   *
   * @return Point2D <zh>Сֵͼελ (Ϊnull).</zh>
   * <en>The position of minimum value shape (maybe null).</en>
   */
  public Point2D getMinShapePos() {
    return this.minShapePos ;
  }

  /**
   * <zh>Сֵͼελ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the position of minimum value shape and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param point Point2D <zh>Сֵͼελ (Ϊnull).</zh>
   * <en>The position of minimum value shape (null permitted).</en>
   */
  public void setMinShapePos(Point2D point) {
    if (!(this.minShapePos == null && point == null)) {
      this.minShapePos = point ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>شֵͼελ.</zh>
   * <en>Returns the position of maximum value shape.</en>
   *
   * @return Point2D <zh>ֵͼελ (Ϊnull).</zh>
   * <en>The position of maximum value shape (maybe null).</en>
   */
  public Point2D getMaxShapePos() {
    return this.maxShapePos ;
  }

  /**
   * <zh>ôֵͼελ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the position of maximum value shape and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param point Point2D <zh>ֵͼελ (Ϊnull).</zh>
   * <en>The position of maximum value shape (null permitted).</en>
   */
  public void setMaxShapePos(Point2D point) {
    if (!(this.maxShapePos == null && point == null)) {
      this.maxShapePos = point ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Returns a flag controls whether or not draw group line.</en>
   *
   * @return boolean <zh>Ƿ.</zh>
   * <en>A flag controls whether or not draw group line.</en>
   */
  public boolean isDrawGroupLine() {
    return this.drawGroupLine ;
  }

  /**
   * <zh>Ƿ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets a flag controls whether or not draw group line and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>Ƿ.</zh>
   * <en>A flag controls whether or not draw group line.</en>
   */
  public void setDrawGroupLine(boolean flag) {
    if (this.drawGroupLine != flag) {
      this.drawGroupLine = flag ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻƴͼ</zh>
   * <en>Draw the ChartShape on a Java 2d graphics device(such as the screen or the printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull).</zh>
   * <en>The graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  public void draw(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (!isVisible()) {
      return ;
    }

    if (this.minShapePos == null && this.maxShapePos == null) {
      return ;
    }

    Graphics2DState state = new Graphics2DState(g2) ;
    state.saveOldState() ;

    Point2D translatePos = this.minShapePos ;
    if (translatePos == null) {
      translatePos = this.maxShapePos ;
    }

    g2.translate(translatePos.getX(), translatePos.getY()) ;
    g2.rotate(getRotate()) ;
    beforeDraw(g2) ;

    //draw group line
    if (this.drawGroupLine && this.minShapePos != null && this.maxShapePos != null) {
      if (this.groupLinePaint != null) {
        g2.setPaint(this.groupLinePaint) ;
      }
      if (this.groupLineStroke != null) {
        g2.setStroke(this.groupLineStroke) ;
      }

      Line2D groupLine = new Line2D.Double(this.minShapePos.getX() - translatePos.getX(),
                                           this.minShapePos.getY() - translatePos.getY(),
                                           this.maxShapePos.getX() - translatePos.getX(),
                                           this.maxShapePos.getY() - translatePos.getY()) ;
      g2.draw(groupLine) ;
    }

    //draw min value shape
    if (this.minShape != null && this.minShapePos != null) {
      if (this.minShapePaint != null) {
        g2.setPaint(this.minShapePaint) ;
      }
      if (this.minShapeStroke != null) {
        g2.setStroke(this.minShapeStroke) ;
      }

      g2.translate(this.minShapePos.getX() - translatePos.getX(),
                   this.minShapePos.getY() - translatePos.getY()) ;
      g2.draw(this.minShape) ;
      g2.translate( -(this.minShapePos.getX() - translatePos.getX()),
                   -(this.minShapePos.getY() - translatePos.getY())) ;
    }

    //draw max value shape
    if (this.maxShape != null && this.maxShapePos != null) {
      if (this.maxShapePaint != null) {
        g2.setPaint(this.maxShapePaint) ;
      }
      if (this.maxShapeStroke != null) {
        g2.setStroke(this.maxShapeStroke) ;
      }

      g2.translate(this.maxShapePos.getX() - translatePos.getX(),
                   this.maxShapePos.getY() - translatePos.getY()) ;
      g2.draw(this.maxShape) ;
      g2.translate( -(this.maxShapePos.getX() - translatePos.getX()),
                   -(this.maxShapePos.getY() - translatePos.getY())) ;
    }

    afterDraw(g2) ;
    g2.rotate( -getRotate()) ;
    state.restoreOldState() ;
    drawSelect(g2) ;
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻͼεѡв,Ϊͼοܻص,ԻͼʱҪȻеͼκ,
   * ٵô˷ѡв.</zh>
   * <en>Draw the selected ChartShape on a Java 2d graphics device(such
   * as the screen or the printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>the graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  protected void drawSelect(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }

    if (isSelected()) {
      if (this.minShapePos == null && this.maxShapePos == null) {
        return ;
      }

      Graphics2DState state = new Graphics2DState(g2) ;
      state.saveOldState() ;
      g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getSelectRectAlpha())) ;
      double rectWidth = getSelectRectWidth() ;

      g2.setPaint(getSelectAnchorPaint()) ;
      Point2D translatePos = this.minShapePos ;
      if (translatePos == null) {
        translatePos = this.maxShapePos ;
      }

      g2.translate(translatePos.getX(), translatePos.getY()) ;
      g2.rotate(getRotate()) ;

      if (this.drawGroupLine && this.minShapePos != null && this.maxShapePos != null) {
        Point2D point1 = new Point2D.Double(this.minShapePos.getX() - translatePos.getX(),
                                            this.minShapePos.getY() - translatePos.getY()) ;
        Rectangle2D anchor1 = AnchorUtil.getAnchorRect(point1, rectWidth) ;
        g2.fill(anchor1) ;

        double point2X = ((this.minShapePos.getX() - translatePos.getX())
                          + (this.maxShapePos.getX() - translatePos.getX())) * 0.5 ;
        double point2Y = ((this.minShapePos.getY() - translatePos.getY())
                          + (this.maxShapePos.getY() - translatePos.getY())) * 0.5 ;
        Point2D point2 = new Point2D.Double(point2X, point2Y) ;
        Rectangle2D anchor2 = AnchorUtil.getAnchorRect(point2, rectWidth) ;
        g2.fill(anchor2) ;

        Point2D point3 = new Point2D.Double(this.maxShapePos.getX() - translatePos.getX(),
                                            this.maxShapePos.getY() - translatePos.getY()) ;
        Rectangle2D anchor3 = AnchorUtil.getAnchorRect(point3, rectWidth) ;
        g2.fill(anchor3) ;

        state.restoreTransform() ;
      }

      g2.rotate( -getRotate()) ;
      state.restoreOldState() ;
    }

  }

  /**
   * <zh>ƶͼ,X᷽ƫtx,Y᷽ƫty,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Concatenates the current Graphics2D Transform with a translation transform.and sends
   * a {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.
   * Subsequent rendering is translated by the specified
   * distance relative to the previous position.
   * This is equivalent to calling transform(T), where T is an
   * AffineTransform represented by the following matrix:
   * <pre>
   *		[   1    0    tx  ]
   *		[   0    1    ty  ]
   *		[   0    0    1   ]
   * </pre>.</en>
   *
   * @param tx double <zh>X᷽ľ.</zh>
   * <en>The distance to translate along the x-axis.</en>
   * @param ty double <zh>Y᷽ľ.</zh>
   * <en>The distance to translate along the y-axis.</en>
   */
  public void translate(double tx, double ty) {
    if (!(tx == 0 && ty == 0)) {
      if (this.minShapePos != null) {
        this.minShapePos.setLocation(this.minShapePos.getX() + tx, this.minShapePos.getY() + ty) ;
      }
      if (this.maxShapePos != null) {
        this.maxShapePos.setLocation(this.maxShapePos.getX() + tx, this.maxShapePos.getY() + ty) ;
      }

      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼϽǵ(ͼεת).</zh>
   * <en>Returns the top left corner coordinates of this shape (not consider rotation).</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getTopLeftPoint() {
    if (this.minShapePos == null && this.maxShapePos == null) {
      return new Point2D.Double(0, 0) ;
    }

    double x = Double.NaN ;
    double y = Double.NaN ;
    if (this.minShapePos != null) {
      x = this.minShapePos.getX() ;
      y = this.minShapePos.getY() ;
    }

    if (this.maxShapePos != null) {
      if (Double.isNaN(x)) {
        x = this.maxShapePos.getX() ;
      } else {
        x = Math.min(x, this.maxShapePos.getX()) ;
      }

      if (Double.isNaN(y)) {
        y = this.maxShapePos.getY() ;
      } else {
        y = Math.min(y, this.maxShapePos.getY()) ;
      }
    }

    if (Double.isNaN(x) || Double.isNaN(y)) {
      return new Point2D.Double(0, 0) ;
    } else {
      return new Point2D.Double(x, y) ;
    }
  }

  /**
   * <zh>ʱͼεĵ.</zh>
   * <en>Returns the lock coordinates of this shape.</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getLockPoint() {
    if (this.minShapePos == null && this.maxShapePos == null) {
      return null ;
    } else if (this.minShapePos == null || this.maxShapePos == null) {
      return (this.minShapePos != null) ? this.minShapePos : this.maxShapePos ;
    } else {
      double centerX = (this.minShapePos.getX() + this.maxShapePos.getX()) * 0.5 ;
      double centerY = (this.minShapePos.getY() + this.maxShapePos.getY()) * 0.5 ;
      return new Point2D.Double(centerX, centerY) ;
    }
  }

  /**
   * <zh>任ͼ.</zh>
   * <en>Calculate the transformed shape.</en>
   */
  protected void calculateTransformedShape() {
    if (transformedShape == null) {
      if (this.minShapePos == null && this.maxShapePos == null) {
        transformedShape = null ;
        return ;
      }

      GeneralPath transformShapePath = new GeneralPath() ;
      Point2D translatePos = this.minShapePos ;
      if (translatePos == null) {
        translatePos = this.maxShapePos ;
      }

      if (this.drawGroupLine && this.minShapePos != null && this.maxShapePos != null) {
        Line2D groupLine = new Line2D.Double(this.minShapePos.getX() - translatePos.getX(),
                                             this.minShapePos.getY() - translatePos.getY(),
                                             this.maxShapePos.getX() - translatePos.getX(),
                                             this.maxShapePos.getY() - translatePos.getY()) ;
        Shape lineRect = StrokeUtil.lineToRectangle(groupLine, this.groupLineStroke, 0) ;
        transformShapePath.append(lineRect, false) ;
      }

      if (this.minShape != null && this.minShapePos != null) {
        float transX = (float) (this.minShapePos.getX() - translatePos.getX()) ;
        float transY = (float) (this.minShapePos.getY() - translatePos.getY()) ;
        AffineTransform transform = AffineTransform.getTranslateInstance(transX, transY) ;
        Shape transformShape = transform.createTransformedShape(this.minShape) ;
        transformShapePath.append(transformShape, false) ;
      }

      if (this.maxShape != null && this.maxShapePos != null) {
        float transX = (float) (this.maxShapePos.getX() - translatePos.getX()) ;
        float transY = (float) (this.maxShapePos.getY() - translatePos.getY()) ;
        AffineTransform transform = AffineTransform.getTranslateInstance(transX, transY) ;
        Shape transformShape = transform.createTransformedShape(this.maxShape) ;
        transformShapePath.append(transformShape, false) ;
      }

      AffineTransform transform = AffineTransform.getTranslateInstance(translatePos.getX(),
                                                                       translatePos.getY()) ;
      transform.rotate(getRotate()) ;
      transformedShape = transform.createTransformedShape(transformShapePath) ;
    }
  }

  /**
   * <zh>ȵͼ.</zh>
   * <en>Returns the hotspot shape.</en>
   *
   * @return Shape <zh>ȵͼ (Ϊnull).</zh>
   * <en>The hotspot shape (maybe null).</en>
   */
  protected Shape getHotspotShape() {
    calculateTransformedShape() ;
    return transformedShape ;
  }

  /**
   * <zh>ͼε.</zh>
   * <en>The type for describe this shape.</en>
   *
   * @return ShapeType <zh>{@link ShapeType}ඨĳ (Ϊnull).</zh>
   * <en>The constant value in class {@link ShapeType}.</en>
   */
  public ShapeType getShapeType() {
    return ShapeType.SHAPE_MIN_AND_MAX ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates whether some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false
   *  otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    if (!(obj instanceof MinMaxChartShape)) {
      return false ;
    }

    MinMaxChartShape that = (MinMaxChartShape) obj ;
    if (!EqualsUtil.equals(this.minShape, that.minShape)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.maxShape, that.maxShape)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.minShapePaint, that.minShapePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.maxShapePaint, that.maxShapePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.groupLinePaint, that.groupLinePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.minShapeStroke, that.minShapeStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.maxShapeStroke, that.maxShapeStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.groupLineStroke, that.groupLineStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.minShapePos, that.minShapePos)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.maxShapePos, that.maxShapePos)) {
      return false ;
    }
    if (this.drawGroupLine != that.drawGroupLine) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double hashCode = super.hashCode() ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.minShape) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.maxShape) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.minShapePaint) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.maxShapePaint) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.groupLinePaint) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.minShapeStroke) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.maxShapeStroke) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.groupLineStroke) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.minShapePos) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.maxShapePos) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.drawGroupLine) ;

    return (int) hashCode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object. </en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[minShapePos=" + this.minShapePos + ",maxShapePos=" + this.maxShapePos + "]" ;
  }

  /**
   * <zh>л.</zh>
   * <en>Provides serialization support.</en>
   * @param stream <zh>.</zh>
   * <en>the output stream.</en>
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writePoint2D(stream, this.minShapePos) ;
    SerialUtil.writePoint2D(stream, this.maxShapePos) ;
    SerialUtil.writeShape(stream, this.minShape) ;
    SerialUtil.writeShape(stream, this.maxShape) ;
    SerialUtil.writePaint(stream, this.minShapePaint) ;
    SerialUtil.writePaint(stream, this.maxShapePaint) ;
    SerialUtil.writePaint(stream, this.groupLinePaint) ;
    SerialUtil.writeStroke(stream, this.minShapeStroke) ;
    SerialUtil.writeStroke(stream, this.maxShapeStroke) ;
    SerialUtil.writeStroke(stream, this.groupLineStroke) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.minShapePos = SerialUtil.readPoint2D(stream) ;
    this.maxShapePos = SerialUtil.readPoint2D(stream) ;
    this.minShape = SerialUtil.readShape(stream) ;
    this.maxShape = SerialUtil.readShape(stream) ;
    this.minShapePaint = SerialUtil.readPaint(stream) ;
    this.maxShapePaint = SerialUtil.readPaint(stream) ;
    this.groupLinePaint = SerialUtil.readPaint(stream) ;
    this.minShapeStroke = SerialUtil.readStroke(stream) ;
    this.maxShapeStroke = SerialUtil.readStroke(stream) ;
    this.groupLineStroke = SerialUtil.readStroke(stream) ;
  }
}
