package com.ijchart.xychart.shape ;

import java.awt.AlphaComposite ;
import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Graphics2D ;
import java.awt.Paint ;
import java.awt.Shape ;
import java.awt.Stroke ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.Arc2D ;
import java.awt.geom.Area ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;

import com.ijchart.xychart.constant.HorizontalAlignment ;
import com.ijchart.xychart.constant.ShapeType ;
import com.ijchart.xychart.ui.Graphics2DState ;
import com.ijchart.xychart.ui.RectangleInsets ;
import com.ijchart.xychart.util.AngleUtil ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.GeometryUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.MathUtil ;
import com.ijchart.xychart.util.SerialUtil ;
import com.ijchart.xychart.util.StringUtil ;
import com.ijchart.xychart.util.StrokeUtil ;
import com.ijchart.xychart.util.TextUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ͼ.</zh>
 * <en>Draw a sector shape</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class SectorChartShape extends AbstractRectangleTextChartShape implements Cloneable,
  Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -7117940820413771251L ;

  /**
   * <zh>ȱʡǷư뾶.</zh>
   * <en>A default boolean flag controls whether or not draw arc radius.</en>
   */
  public static final boolean DEFAULT_DRAW_RADIUS = false ;

  /**
   * <zh>Ʊ߿paint.</zh>
   * <en>The paint used to draw border.</en>
   */
  public static final Paint DEFAULT_ARC_BORDER_PAINT = Color.lightGray ;

  /**
   * <zh>ȱʡıǩıpaint.</zh>
   * <en>The default section label background paint.</en>
   */
  public static final Paint DEFAULT_LABEL_BACKGROUND_PAINT = new Color(255, 255, 192) ;

  /**
   * <zh>Ʊǩı߿paint.</zh>
   * <en>The default section label outline paint.</en>
   */
  public static final Paint DEFAULT_LABEL_OUTLINE_PAINT = Color.black ;

  /**
   * <zh>ȱʡĻƱ߿stroke.</zh>
   * <en>The default section label outline stroke.</en>
   */
  public static final Stroke DEFAULT_LABEL_OUTLINE_STROKE = new BasicStroke(0.5f) ;

  /**
   * <zh>ȱʡĻƱǩӰpaint.</zh>
   * <en>The default section label shadow paint.</en>
   */
  public static final Paint DEFAULT_LABEL_SHADOW_PAINT = new Color(151, 151, 151, 128) ;

  /**
   * <zh>ȱʡǩĲ.</zh>
   * <en>The default padding between the labels and the label outlines.</en>
   */
  public static final RectangleInsets DEFAULT_LABEL_PADDING = new RectangleInsets(2, 2, 2, 2) ;

  /**
   * <zh>ȱʡıǩxӰ.</zh>
   * <en>The default x-offset for the label shadow effect.</en>
   */
  public static final double DEFAULT_LABEL_SHADOW_X_OFFSET = 4F ;

  /**
   * <zh>ȱʡıǩyӰ.</zh>
   * <en>The default y-offset for the label shadow effect.</en>
   */
  public static final double DEFAULT_LABEL_SHADOW_Y_OFFSET = 4F ;

  /**
   * <zh>ȱʡĻͼεpaint.</zh>
   * <en>The default paint used to draw sector shadow.</en>
   */
  public static final Paint DEFAULT_SECTOR_SHADOW_PAINT = Color.gray ;

  /**
   * <zh>ȱʡxӰ.</zh>
   * <en>The default x-offset for the sector shadow effect.</en>
   */
  public static final double DEFAULT_SECTOR_SHADOW_X_OFFSET = 4F ;

  /**
   * <zh>ȱʡyӰ.</zh>
   * <en>The default y-offset for the sector shadow effect.</en>
   */
  public static final double DEFAULT_SECTOR_SHADOW_Y_OFFSET = 4F ;

  /**
   * <zh>ȱʡǷ3D͵ͼ.</zh>
   * <en>A default flag that controls whether or not the sector is 3D.</en>
   */
  public static final boolean DEFAULT_SECTOR_IS_3D = false ;

  /**
   * <zh>ȱʡĻ3DЧ.</zh>
   * <en>The default depth of 3D pie.</en>
   */
  public static final double DEFAULT_DEPTH_3D = 8.0 ;

  /**
   * <zh>ͼ.</zh>
   * <en>The arc shape.</en>
   */
  private Arc2D arc ;

  /**
   * <zh>Ƿư뾶.</zh>
   * <en>A boolean flag controls whether or not draw arc radius.</en>
   */
  private boolean drawRadius ;

  /**
   * <zh>ǩıɫ.</zh>
   * <en>The color used to draw the background of the section labels.</en>
   */
  private transient Paint labelBackgroundPaint ;

  /**
   * <zh>Ʊǩı߿paint.</zh>
   * <en>The paint used to draw the outline of the section labels.</en>
   */
  private transient Paint labelOutlinePaint ;

  /**
   * <zh>Ʊ߿stroke.</zh>
   * <en>The stroke used to draw the outline of the section labels.</en>
   */
  private transient Stroke labelOutlineStroke ;

  /**
   * <zh>ƱǩӰpaint.</zh>
   * <en> The paint used to draw the shadow for the section labels.</en>
   */
  private transient Paint labelShadowPaint ;

  /**
   * <zh>ǩĲ.</zh>
   * <en>The padding between the labels and the label outlines.</en>
   */
  private RectangleInsets labelPadding ;

  /**
   * <zh>ǩxӰ.</zh>
   * <en>The x-offset for the label shadow effect.</en>
   */
  private double labelShadowXOffset ;

  /**
   * <zh>ǩyӰ.</zh>
   * <en>The y-offset for the label shadow effect.</en>
   */
  private double labelShadowYOffset ;

  /**
   * <zh>ͼεpaint.</zh>
   * <en>The paint used to draw sector shadow.</en>
   */
  private transient Paint sectorShadowPaint ;

  /**
   * <zh>xӰ.</zh>
   * <en>The x-offset for the sector shadow effect.</en>
   */
  private double sectorShadowXOffset ;

  /**
   * <zh>yӰ.</zh>
   * <en>The y-offset for the sector shadow effect.</en>
   */
  private double sectorShadowYOffset ;

  /**
   * <zh>Ƿ3D͵ͼ.</zh>
   * <en>A flag that controls whether or not the sector is 3D.</en>
   */
  private boolean sector3D ;

  /**
   * <zh>3DЧ.</zh>
   * <en>The depth of 3D pie.</en>
   */
  private double depth3D ;

  /**
   * <zh>.</zh>
   * <en>A constructor.</en>
   *
   * @param arc Arc2D <zh>ͼ (Ϊnull).</zh>
   * <en>A arc shape (null not permitted).</en>
   * @param drawArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The draw area (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   */
  public SectorChartShape(Arc2D arc, Rectangle2D drawArea, String text) {
    this(arc, drawArea, text, DEFAULT_SECTOR_IS_3D) ;
  }

  /**
   * <zh>.</zh>
   * <en>A constructor.</en>
   *
   * @param arc Arc2D <zh>ͼ (Ϊnull).</zh>
   * <en>A arc shape (null not permitted).</en>
   * @param drawArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The draw area (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param sector3D boolean <zh>Ƿ3D͵ͼ.</zh>
   * <en>A flag that controls whether or not the sector is 3D.</en>
   */
  public SectorChartShape(Arc2D arc, Rectangle2D drawArea, String text, boolean sector3D) {
    super(drawArea, text) ;
    if (arc == null) {
      throw new IllegalArgumentException("Null 'arc' argument.") ;
    }

    this.arc = arc ;
    this.drawRadius = DEFAULT_DRAW_RADIUS ;

    this.labelBackgroundPaint = DEFAULT_LABEL_BACKGROUND_PAINT ;
    this.labelOutlinePaint = DEFAULT_LABEL_OUTLINE_PAINT ;
    this.labelOutlineStroke = DEFAULT_LABEL_OUTLINE_STROKE ;
    this.labelPadding = DEFAULT_LABEL_PADDING ;

    this.labelShadowPaint = DEFAULT_LABEL_SHADOW_PAINT ;
    this.labelShadowXOffset = DEFAULT_LABEL_SHADOW_X_OFFSET ;
    this.labelShadowYOffset = DEFAULT_LABEL_SHADOW_Y_OFFSET ;

    this.sectorShadowPaint = DEFAULT_SECTOR_SHADOW_PAINT ;
    this.sectorShadowXOffset = DEFAULT_SECTOR_SHADOW_X_OFFSET ;
    this.sectorShadowYOffset = DEFAULT_SECTOR_SHADOW_Y_OFFSET ;

    this.sector3D = sector3D ;
    this.depth3D = DEFAULT_DEPTH_3D ;

    setFillShape(true) ;
    setDrawBorder(false) ;
    setBorderPaint(DEFAULT_ARC_BORDER_PAINT) ;
    setTextHorizontalAlign(HorizontalAlignment.CENTER) ;
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Returns the arc shape.</en>
   *
   * @return Arc2D <zh>ͼ (Ϊnull).</zh>
   * <en>The arc shape (never null).</en>
   */
  public Arc2D getArc() {
    return this.arc ;
  }

  /**
   * <zh>ͼ,м{@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the arc shape,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param arc Arc2D <zh>ͼ (Ϊnull).</zh>
   * <en>The arc shape (null not permitted).</en>
   */
  public void setArc(Arc2D arc) {
    if (arc == null) {
      throw new IllegalArgumentException("Null 'arc' argument.") ;
    }

    this.arc = arc ;
    fireChartShapeChanged() ;
  }

  /**
   * <zh>Ƿư뾶.</zh>
   * <en>A boolean flag controls whether or not draw arc radius.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDrawRadius() {
    return this.drawRadius ;
  }

  /**
   * <zh>Ƿư뾶,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>A boolean flag controls whether or not draw arc radius,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setDrawRadius(boolean flag) {
    if (this.drawRadius != flag) {
      this.drawRadius = flag ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>رǩıɫ.</zh>
   * <en>Returns the color used to draw the background of the section labels.</en>
   *
   * @return Paint <zh>paint.</zh>
   * <en>The paint.</en>
   */
  public Paint getLabelBackgroundPaint() {
    return this.labelBackgroundPaint ;
  }

  /**
   * <zh>ñǩıɫ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the color used to draw the background of the section labels and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>paint (Ϊnull).</zh>
   * <en>The paint (null permitted).</en>
   */
  public void setLabelBackgroundPaint(Paint paint) {
    if (!(this.labelBackgroundPaint == null && paint == null)) {
      this.labelBackgroundPaint = paint ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػƱǩı߿paint.</zh>
   * <en>Returns the paint used to draw the outline of the section labels.</en>
   *
   * @return Paint <zh>paint.</zh>
   * <en>The paint.</en>
   */
  public Paint getLabelOutlinePaint() {
    return this.labelOutlinePaint ;
  }

  /**
   * <zh>ûƱǩı߿paint,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the paint used to draw the outline of the section labels and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>paint (Ϊnull).</zh>
   * <en>The paint (null permitted).</en>
   */
  public void setLabelOutlinePaint(Paint paint) {
    if (!(this.labelBackgroundPaint == null && paint == null)) {
      this.labelOutlinePaint = paint ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػƱ߿stroke.</zh>
   * <en>Returns the stroke used to draw the outline of the section labels.</en>
   *
   * @return Stroke <zh>stroke.</zh>
   * <en>The stroke.</en>
   */
  public Stroke getLabelOutlineStroke() {
    return this.labelOutlineStroke ;
  }

  /**
   * <zh>ûƱ߿stroke,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the stroke used to draw the outline of the section labels and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>stroke (Ϊnull).</zh>
   * <en>The stroke (null permitted).</en>
   */
  public void setLabelOutlineStroke(Stroke stroke) {
    if (!(this.labelOutlineStroke == null && stroke == null)) {
      this.labelOutlineStroke = stroke ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػͼεpaint.</zh>
   * <en>Returns the paint used to draw sector shadow.</en>
   *
   * @return Paint <zh>paint (Ϊnull).</zh>
   * <en>The paint (maybe null).</en>
   */
  public Paint getLabelShadowPaint() {
    return this.labelShadowPaint ;
  }

  /**
   * <zh>ûͼεpaint,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the paint used to draw sector shadow and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>paint (Ϊnull).</zh>
   * <en>The paint (null permitted).</en>
   */
  public void setLabelShadowPaint(Paint paint) {
    if (!(this.labelShadowPaint == null && paint == null)) {
      this.labelShadowPaint = paint ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>رǩĲ.</zh>
   * <en>Returns the padding between the labels and the label outlines.</en>
   *
   * @return RectangleInsets <zh> (Ϊnull).</zh>
   * <en>The padding (never null).</en>
   */
  public RectangleInsets getLabelPadding() {
    return this.labelPadding ;
  }

  /**
   * <zh>ñǩĲ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the padding between the labels and the label outlines and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param padding RectangleInsets <zh> (Ϊnull).</zh>
   * <en>The padding (null not permitted).</en>
   */
  public void setLabelPadding(RectangleInsets padding) {
    if (padding == null) {
      throw new IllegalArgumentException("Null 'padding' argument") ;
    }

    this.labelPadding = padding ;
    fireChartShapeChanged() ;
  }

  /**
   * <zh>رǩxӰ.</zh>
   * <en>Returns the x-offset for the label shadow effect.</en>
   *
   * @return double <zh>ƫֵ.</zh>
   * <en>The x-offset.</en>
   */
  public double getLabelShadowXOffset() {
    return this.labelShadowXOffset ;
  }

  /**
   * <zh>ñǩxӰ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the x-offset for the label shadow effect and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param offset double <zh>ƫֵ.</zh>
   * <en>The x-offset.</en>
   */
  public void setLabelShadowXOffset(double offset) {
    if (offset < 0) {
      throw new IllegalArgumentException("Argument 'offset' must be >=0.") ;
    }

    if (this.labelShadowXOffset != offset) {
      this.labelShadowXOffset = offset ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>رǩxӰ.</zh>
   * <en>Returns the y-offset for the label shadow effect.</en>
   *
   * @return double <zh>ƫֵ.</zh>
   * <en>The y-offset.</en>
   */
  public double getLabelShadowYOffset() {
    return this.labelShadowYOffset ;
  }

  /**
   * <zh>ñǩyӰ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the y-offset for the label shadow effect and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param offset double <zh>ƫֵ.</zh>
   * <en>The y-offset.</en>
   */
  public void setLabelShadowYOffset(double offset) {
    if (this.labelShadowYOffset != offset) {
      this.labelShadowYOffset = offset ;
      fireChartShapeChanged() ;
    }
  }

  /**
   <zh>ػƱǩӰpaint.</zh>
   * <en>Returns the paint used to draw the shadow for the section labels.</en>
   *
   * @return Paint <zh>paint.</zh>
   * <en>The paint.</en>
   */
  public Paint getSectorShadowPaint() {
    return this.sectorShadowPaint ;
  }

  /**
   * <zh>ûƱǩӰpaint,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the paint used to draw the shadow for the section labels and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint
   */
  public void setSectorShadowPaint(Paint paint) {
    this.sectorShadowPaint = paint ;
  }

  /**
   * <zh>xӰ.</zh>
   * <en>Returns the x-offset for the sector shadow effect.</en>
   *
   * @return double <zh>ƫֵ.</zh>
   * <en>The x-offset.</en>
   */
  public double getSectorShadowXOffset() {
    return this.sectorShadowXOffset ;
  }

  /**
   * <zh>xӰ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the x-offset for the sector shadow effect and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param offset double <zh>ƫֵ.</zh>
   * <en>The x-offset.</en>
   */
  public void setSectorShadowXOffset(double offset) {
    if (offset < 0) {
      throw new IllegalArgumentException("Argument 'offset' must be >=0.") ;
    }

    if (this.sectorShadowXOffset != offset) {
      this.sectorShadowXOffset = offset ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>xӰ.</zh>
   * <en>Returns the y-offset for the sector shadow effect.</en>
   *
   * @return double <zh>ƫֵ.</zh>
   * <en>The y-offset.</en>
   */
  public double getSectorShadowYOffset() {
    return this.sectorShadowYOffset ;
  }

  /**
   * <zh>yӰ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the y-offset for the sector shadow effect and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param offset double <zh>ƫֵ.</zh>
   * <en>The y-offset.</en>
   */
  public void setSectorShadowYOffset(double offset) {
    if (this.sectorShadowYOffset != offset) {
      this.sectorShadowYOffset = offset ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>Ƿ3D͵ͼ.</zh>
   * <en>Returns the flag that controls whether or not the sector is 3D.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isSector3D() {
    return this.sector3D ;
  }

  /**
   * <zh>Ƿ3D͵ͼ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the flag that controls whether or not the sector is 3D and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setSector3D(boolean flag) {
    if (this.sector3D != flag) {
      this.sector3D = flag ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ػ3DЧ.</zh>
   * <en>Returns the depth of 3D pie.</en>
   *
   * @return double <zh>.</zh>
   * <en>The depth.</en>
   */
  public double getDepth3D() {
    return this.depth3D ;
  }

  /**
   * <zh>û3DЧ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the depth of 3D pie and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param depth3D double <zh>.</zh>
   * <en>The depth.</en>
   */
  public void setDepth3D(double depth3D) {
    if (depth3D >= 0 && this.depth3D != depth3D) {
      this.depth3D = depth3D ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻƴͼ.ʱshapeλ.</zh>
   * <en>Draw the ChartShape on a Java 2d graphics device(such as the screen or the printer).
   * The shape's coordinate is ignored.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>The graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  public void draw(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (!isVisible()) {
      return ;
    }

    Graphics2DState state = new Graphics2DState(g2) ;
    state.saveOldState() ;

    double angleStart = this.arc.getAngleStart() ; //ʼǶ
    double angleExtent = this.arc.getAngleExtent() ; //
    double arcX = this.arc.getX() ;
    double arcY = this.arc.getY() ;
    double arcWidth = this.arc.getWidth() ;
    double arcHeight = this.arc.getHeight() ;
    Arc2D newArc = new Arc2D.Double(0.0D, 0.0D, arcWidth, arcHeight,
                                    angleStart, angleExtent, this.arc.getArcType()) ;

    g2.translate(arcX, arcY) ;
    g2.rotate(getRotate()) ;
    beforeDraw(g2) ;
//    g2.setClip(newArc) ;

    //Ӱ
    if (this.sectorShadowPaint != null) {
      AffineTransform transform = AffineTransform.getTranslateInstance(this.sectorShadowXOffset,
                                                                       this.sectorShadowYOffset) ;
      Shape shadowSector = transform.createTransformedShape(newArc) ;
      g2.setPaint(this.sectorShadowPaint) ;
      g2.fill(shadowSector) ;
    }

    if (this.sector3D) { //Ҫ3DЧ
      Arc2D arcBottom = new Arc2D.Double(newArc.getX(), newArc.getY() + this.depth3D,
                                         newArc.getWidth(), newArc.getHeight(),
                                         newArc.getAngleStart(), newArc.getAngleExtent(),
                                         newArc.getArcType()) ;
      Point2D[] topPs = getPointsOfArc(newArc) ;
      Point2D[] bottomPs = getPointsOfArc(arcBottom) ;

      //Front site
      GeneralPath front = new GeneralPath() ;
      front.moveTo((float) topPs[1].getX(), (float) topPs[1].getY()) ;
      front.lineTo((float) topPs[0].getX(), (float) topPs[0].getY()) ;
      front.lineTo((float) topPs[2].getX(), (float) topPs[2].getY()) ;
      front.lineTo((float) bottomPs[2].getX(), (float) bottomPs[2].getY()) ;
      front.lineTo((float) bottomPs[0].getX(), (float) bottomPs[0].getY()) ;
      front.lineTo((float) bottomPs[1].getX(), (float) bottomPs[1].getY()) ;
      front.closePath() ;
      Area frontSite = new Area(arcBottom) ;
      frontSite.add(new Area(front)) ;

      //Left site
      GeneralPath left = new GeneralPath() ;
      left.moveTo((float) topPs[0].getX(), (float) topPs[0].getY()) ;
      left.lineTo((float) topPs[1].getX(), (float) topPs[1].getY()) ;
      left.lineTo((float) bottomPs[1].getX(), (float) bottomPs[1].getY()) ;
      left.lineTo((float) topPs[0].getX(), (float) topPs[0].getY() + 3) ;
      left.closePath() ;
      Area leftSite = new Area(left) ;

      //Right site
      GeneralPath right = new GeneralPath() ;
      right.moveTo((float) topPs[0].getX(), (float) topPs[0].getY()) ;
      right.lineTo((float) topPs[2].getX(), (float) topPs[2].getY()) ;
      right.lineTo((float) bottomPs[2].getX(), (float) bottomPs[2].getY()) ;
      right.lineTo((float) topPs[0].getX(), (float) topPs[0].getY() + 3) ;
      right.closePath() ;
      Area rightSite = new Area(right) ;

      Paint fill = getFillPaint() ;
      if (fill instanceof Color) {
        g2.setPaint(((Color) fill).darker()) ;
      } else {
        g2.setPaint(fill) ;
      }

      g2.fill(frontSite) ;
      g2.fill(leftSite) ;
      g2.fill(rightSite) ;
    }

    drawBackground(g2, newArc) ;
    drawBorder(g2, newArc) ;

    //Ʊ߿,ҪǷư뾶
    if (this.drawRadius) {
      newArc.setArcType(Arc2D.PIE) ;
      drawBorder(g2, newArc) ;
    } else {
      newArc.setArcType(Arc2D.OPEN) ;
      drawBorder(g2, newArc) ;
    }

    //Ʊǩ
    String label = getText() ;
    if (isDrawText() && !StringUtil.isEmpty(label)) {
      g2.setFont(getTextFont()) ;
      Rectangle2D bounds = TextUtil.getStringBoundWithFm(g2, label) ;
      Rectangle2D out = getLabelPadding().createOutsetRectangle(bounds) ;

      Arc2D middleArc = new Arc2D.Double(arc.getX(), arc.getY(), arc.getWidth(), arc.getHeight(),
                                         arc.getAngleStart(), arc.getAngleExtent() / 2, Arc2D.PIE) ;
      Point2D arcMiddle = middleArc.getEndPoint() ;
      Point2D arcCenter = new Point2D.Double(arc.getCenterX(), arc.getCenterY()) ;
      double dis = GeometryUtil.distanceOfPoints(arcCenter, arcMiddle) * 0.75 ;
      Point2D labelPosition = GeometryUtil.extentPoint(arcCenter, arcMiddle, dis) ;

      float strokeWidth = StrokeUtil.getStrokeWidth(this.labelOutlineStroke) ;
      Rectangle2D labelBound = new Rectangle2D.Double(out.getX() + labelPosition.getX() - arcX
                                                      - 0.5 * out.getWidth(),
                                                      out.getY() + labelPosition.getY() - arcY
                                                      - 0.5 * out.getHeight(),
                                                      out.getWidth() + 2 * strokeWidth,
                                                      out.getHeight() + 2 * strokeWidth) ;
      //draw label shadow
      if (this.labelShadowPaint != null) {
        Rectangle2D shadowBound = new Rectangle2D.Double(labelBound.getX() + labelShadowXOffset,
                                                         labelBound.getY() + labelShadowYOffset,
                                                         labelBound.getWidth(),
                                                         labelBound.getHeight()) ;
        g2.setPaint(this.labelShadowPaint) ;
        g2.fill(shadowBound) ;
      }

      //draw label background
      if (this.labelBackgroundPaint != null) {
        g2.setPaint(this.labelBackgroundPaint) ;
        g2.fill(labelBound) ;
      }

      //draw label border
      if (this.labelOutlinePaint != null && this.labelOutlineStroke != null) {
        g2.setPaint(this.labelOutlinePaint) ;
        g2.setStroke(this.labelOutlineStroke) ;
        g2.draw(labelBound) ;
      }

      drawTextWithTrimMarginRect(g2, labelBound) ;
    }

    afterDraw(g2) ;
    g2.rotate( -getRotate()) ;
    state.restoreOldState() ;
    drawSelect(g2) ;
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻͼεѡв,Ϊͼοܻص,ԻͼʱҪȻеͼκ,
   * ٵô˷ѡв.</zh>
   * <en>Draw the selected ChartShape on a Java 2d graphics device(such as the screen or the
   * printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>The graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  protected void drawSelect(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }

    if (isSelected()) {
      Graphics2DState state = new Graphics2DState(g2) ;
      state.saveOldState() ;

      double angleStart = this.arc.getAngleStart() ; //ʼǶ
      double angleExtent = this.arc.getAngleExtent() ;
      double arcX = this.arc.getX() ;
      double arcY = this.arc.getY() ;
      double arcWidth = this.arc.getWidth() ;
      double arcHeight = this.arc.getHeight() ;
      Arc2D newArc = new Arc2D.Double(0.0D, 0.0D, arcWidth, arcHeight,
                                      angleStart, angleExtent, this.arc.getArcType()) ;

      g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getSelectRectAlpha())) ;
      g2.setPaint(getSelectAnchorPaint()) ;

      g2.translate(arcX, arcY) ;
      g2.rotate(getRotate()) ;
//      g2.setClip(newArc) ;

      double rectWidth = getSelectRectWidth() ;
      final double halfRect = rectWidth / 2 ;

      //ê
      double halfArcWidth = 0.5 * this.arc.getWidth() ;
      double halfArcHeight = 0.5 * this.arc.getHeight() ;

      double drawAngle = angleStart ;
      int sign = MathUtil.sign(angleExtent) ;
      final double tickSizeAngle = 30 * sign ;
      double actualMaxAngle = GeometryUtil.calculateArcEndPointAngle(newArc) ;

      int anchorCount = 0 ; //êĸ
      Point2D startPoint = newArc.getStartPoint() ;
      Point2D endPoint = newArc.getEndPoint() ;
      boolean drawEnd = false ; //ǷendPointڵê
      double endX = 0, endY = 0 ; //
      do { //ڿ̶֮
        double radians = drawAngle / 180 * Math.PI ;
        double tanAngle = Math.tan(radians) ;
        int quadrant = AngleUtil.getQuadrantByDegreeAngle(drawAngle) ;

        if (anchorCount == 0) {
          endX = startPoint.getX() ;
          endY = startPoint.getY() ;
        } else {
          double x = Math.sqrt(1 / (1 / Math.pow(halfArcWidth, 2)
                                    + Math.pow(tanAngle, 2) / Math.pow(halfArcHeight, 2))) ;
          double y = Math.abs(x * tanAngle) ;

          if (quadrant == 1 || quadrant == 4) {
            endX = halfArcWidth + x ;
          } else {
            endX = halfArcWidth - x ;
          }
          if (quadrant == 1 || quadrant == 2) {
            endY = halfArcHeight - y ;
          } else {
            endY = halfArcHeight + y ;
          }
        }

        g2.fill(new Rectangle2D.Double(endX - halfRect, endY - halfRect, rectWidth, rectWidth)) ;

        drawAngle += tickSizeAngle ;
        anchorCount++ ;

        if ((sign == 1 && drawAngle > actualMaxAngle) || (sign == -1 && drawAngle < actualMaxAngle)) {
          if (anchorCount >= 2) {
            break ;
          } else {
            drawAngle = actualMaxAngle ;
            drawEnd = true ;
          }
        }
      } while (true) ;

      if (!drawEnd) { //һêյľ̫,ûбҪê
        double distince = Math.sqrt((endX - endPoint.getX()) * (endX - endPoint.getX())
                                    + (endY - endPoint.getY()) * (endY - endPoint.getY())) ;
        if (distince > rectWidth * 3) {
          g2.fill(new Rectangle2D.Double(endPoint.getX() - halfRect, endPoint.getY() - halfRect,
                                         rectWidth, rectWidth)) ;
        }
      }

      g2.rotate( -getRotate()) ;
      state.restoreOldState() ;
    }
  }

  /**
   * <zh>ʱͼεĵ.</zh>
   * <en>Returns the lock coordinates of this shape.</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getLockPoint() {
    if (this.arc == null) {
      return null ;
    }

    double arcX = this.arc.getX() ;
    double arcY = this.arc.getY() ;

    Arc2D middleArc = new Arc2D.Double(arc.getX(), arc.getY(), arc.getWidth(), arc.getHeight(),
                                       arc.getAngleStart(), arc.getAngleExtent() / 2, Arc2D.PIE) ;
    Point2D arcMiddle = middleArc.getEndPoint() ;
    Point2D arcCenter = new Point2D.Double(arc.getCenterX(), arc.getCenterY()) ;
    double dis = GeometryUtil.distanceOfPoints(arcCenter, arcMiddle) * (1 - 0.5) ;
    Point2D labelPos = GeometryUtil.extentPoint(arcCenter, arcMiddle, dis) ;

    Rectangle2D labelBound = new Rectangle2D.Double(labelPos.getX() - arcX, labelPos.getY() - arcY,
                                                    1, 1) ;

    AffineTransform transform = AffineTransform.getTranslateInstance(arcX, arcY) ;
    transform.rotate(getRotate()) ;
    Shape rotateShape = transform.createTransformedShape(labelBound) ;
    Rectangle2D rotateBound = rotateShape.getBounds2D() ;

    return new Point2D.Double(rotateBound.getCenterX(), rotateBound.getCenterY()) ;
  }

  /**
   * <zh>任ͼ.</zh>
   * <en>Calculate the transformed shape.</en>
   */
  protected void calculateTransformedShape() {
    if (transformedShape == null) {
      double angleStart = this.arc.getAngleStart() ; //ʼǶ
      double angleExtent = this.arc.getAngleExtent() ; //
      double arcX = this.arc.getX() ;
      double arcY = this.arc.getY() ;
      Arc2D newArc = new Arc2D.Double(0.0D, 0.0D, this.arc.getWidth(), this.arc.getHeight(),
                                      angleStart, angleExtent, this.arc.getArcType()) ;

      AffineTransform transform = new AffineTransform() ;
      transform.translate(arcX, arcY) ;
      transform.rotate(getRotate()) ;
      transformedShape = transform.createTransformedShape(newArc) ;
    }
  }

  /**
   * <zh>ƶͼ,X᷽ƫtx,Y᷽ƫty,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Concatenates the current Graphics2D Transform with a translation transform.and sends
   * a {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.
   * Subsequent rendering is translated by the specified
   * distance relative to the previous position.
   * This is equivalent to calling transform(T), where T is an
   * AffineTransform represented by the following matrix:
   * <pre>
   *		[   1    0    tx  ]
   *		[   0    1    ty  ]
   *		[   0    0    1   ]
   * </pre>.</en>
   *
   * @param tx double <zh>X᷽ľ.</zh>
   * <en>The distance to translate along the x-axis.</en>
   * @param ty double <zh>Y᷽ľ.</zh>
   * <en>The distance to translate along the y-axis.</en>
   */
  public void translate(double tx, double ty) {
    if (!(tx == 0 && ty == 0)) {
      this.arc.setArc(this.arc.getX() + tx, this.arc.getY() + ty, this.arc.getWidth(),
                      this.arc.getHeight(), this.arc.getAngleStart(), this.arc.getAngleExtent(),
                      this.arc.getArcType()) ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼϽǵ.</zh>
   * <en>Returns the top left corner coordinates of this shape.</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getTopLeftPoint() {
    return new Point2D.Double(this.arc.getX(), this.arc.getY()) ;
  }

  /**
   * <zh>ȵͼ.</zh>
   * <en>Returns the hotspot shape.</en>
   *
   * @return Shape <zh>ȵͼ (Ϊnull).</zh>
   * <en>The hotspot shape (maybe null).</en>
   */
  protected Shape getHotspotShape() {
    GeneralPath arcPath = GeometryUtil.getArcPath(this.arc) ;
    Arc2D newArc = new Arc2D.Double(0.0D, 0.0D, this.arc.getWidth(), this.arc.getHeight(),
                                    this.arc.getAngleStart(), this.arc.getAngleExtent(),
                                    this.arc.getArcType()) ;
    arcPath.lineTo((float) newArc.getCenterX(), (float) newArc.getCenterY()) ;
    arcPath.closePath() ;

    AffineTransform transform = AffineTransform.getTranslateInstance(this.arc.getX(), this.arc.getY()) ;
    transform.rotate(getRotate()) ;

    return transform.createTransformedShape(arcPath) ;
  }

  /**
   * <zh>ͼε.</zh>
   * <en>The type for describe this shape.</en>
   * @return ShapeType <zh>{@link com.ijchart.xychart.constant.ShapeType}ඨĳ
   * (Ϊnull).</zh>
   * <en>The constant value in class {@link com.ijchart.xychart.constant.ShapeType}.</en>
   */
  public ShapeType getShapeType() {
    return ShapeType.SHAPE_ARC ;
  }

  /**
   * <zh>ȡArcϵ,ڶArc: center, left, right.</zh>
   * <en>Returns the points of arc:center,left,right.</en>
   *
   * @param arc Arc2D
   * @return Point2D[]
   */
  protected Point2D[] getPointsOfArc(Arc2D arc) {
    Point2D center = new Point2D.Double(arc.getCenterX(), arc.getCenterY()) ;
    Point2D left = arc.getStartPoint() ;
    Point2D right = arc.getEndPoint() ;

    return new Point2D[] {center, left, right} ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() ;
    result = 37 * result + HashcodeUtil.hashCode(this.arc) ;
    result = 37 * result + HashcodeUtil.hashCode(this.drawRadius) ;

    result = 37 * result + HashcodeUtil.hashCode(this.labelBackgroundPaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.labelOutlinePaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.labelOutlineStroke) ;
    result = 37 * result + HashcodeUtil.hashCode(this.labelPadding) ;

    result = 37 * result + HashcodeUtil.hashCode(this.labelShadowPaint) ;
    result = 37 * result + this.labelShadowXOffset ;
    result = 37 * result + this.labelShadowYOffset ;

    result = 37 * result + HashcodeUtil.hashCode(this.sectorShadowPaint) ;
    result = 37 * result + this.sectorShadowXOffset ;
    result = 37 * result + this.sectorShadowYOffset ;

    result = 37 * result + HashcodeUtil.hashCode(this.sector3D) ;
    result = 37 * result + this.depth3D ;

    return (int) result ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates whether some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof SectorChartShape)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    SectorChartShape that = (SectorChartShape) obj ;
    if (!EqualsUtil.equals(this.arc, that.arc)) {
      return false ;
    }
    if (this.drawRadius != that.drawRadius) {
      return false ;
    }
    if (!EqualsUtil.equals(this.labelBackgroundPaint, that.labelBackgroundPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.labelOutlinePaint, that.labelOutlinePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.labelOutlineStroke, that.labelOutlineStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.labelPadding, that.labelPadding)) {
      return false ;
    }

    if (!EqualsUtil.equals(this.labelShadowPaint, that.labelShadowPaint)) {
      return false ;
    }
    if (this.labelShadowXOffset != that.labelShadowXOffset) {
      return false ;
    }
    if (this.labelShadowYOffset != that.labelShadowYOffset) {
      return false ;
    }

    if (!EqualsUtil.equals(this.sectorShadowPaint, that.sectorShadowPaint)) {
      return false ;
    }
    if (this.sectorShadowXOffset != that.sectorShadowXOffset) {
      return false ;
    }
    if (this.sectorShadowYOffset != that.sectorShadowYOffset) {
      return false ;
    }
    if (this.sector3D != that.sector3D) {
      return false ;
    }
    if (this.depth3D != that.depth3D) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;

    SerialUtil.writePaint(stream, this.labelBackgroundPaint) ;
    SerialUtil.writePaint(stream, this.labelOutlinePaint) ;
    SerialUtil.writePaint(stream, this.labelShadowPaint) ;
    SerialUtil.writeStroke(stream, this.labelOutlineStroke) ;
    SerialUtil.writePaint(stream, this.sectorShadowPaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;

    this.labelBackgroundPaint = SerialUtil.readPaint(stream) ;
    this.labelOutlinePaint = SerialUtil.readPaint(stream) ;
    this.labelShadowPaint = SerialUtil.readPaint(stream) ;
    this.labelOutlineStroke = SerialUtil.readStroke(stream) ;
    this.sectorShadowPaint = SerialUtil.readPaint(stream) ;
  }
}
