package com.ijchart.xychart.ui ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>εĹ.</zh>
 * <en>A utility class for aligning rectangles.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public final class Align implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 6465094977727886842L ;

  /**
   * <zh>ж.</zh>
   * <en>Center alignment.</en>
   */
  public static final Align CENTER = new Align(0x00) ;

  /**
   * <zh>˶.</zh>
   * <en>Top alignment.</en>
   */
  public static final Align TOP = new Align(0x01) ;

  /**
   * <zh>׶˶.</zh>
   * <en>Bottom alignment. </en>
   */
  public static final Align BOTTOM = new Align(0x02) ;

  /**
   * <zh>.</zh>
   * <en>Left alignment.</en>
   */
  public static final Align LEFT = new Align(0x04) ;

  /**
   * <zh>Ҷ.</zh>
   * <en>Right alignment. </en>
   */
  public static final Align RIGHT = new Align(0x08) ;

  /**
   * <zh>/˶.</zh>
   * <en>Top/Left alignment.</en>
   */
  public static final Align TOP_LEFT = new Align(TOP.getValue() | LEFT.getValue()) ;

  /**
   * <zh>/Ҷ˶.</zh>
   * <en>Top/Right alignment. </en>
   */
  public static final Align TOP_RIGHT = new Align(TOP.getValue() | RIGHT.getValue()) ;

  /**
   * <zh>׶/˶.</zh>
   * <en>Bottom/Left alignment.</en>
   */
  public static final Align BOTTOM_LEFT = new Align(BOTTOM.getValue() | LEFT.getValue()) ;

  /**
   * <zh>׶/Ҷ˶</zh>
   * <en>Bottom/Right alignment.</en>
   */
  public static final Align BOTTOM_RIGHT = new Align(BOTTOM.getValue() | RIGHT.getValue()) ;

  /**
   * <zh>ˮƽӦС.</zh>
   * <en>Horizontal fit. </en>
   */
  public static final Align FIT_HORIZONTAL = new Align(LEFT.getValue() | RIGHT.getValue()) ;

  /**
   * <zh>ֱӦС.</zh>
   * <en>Vertical fit. </en>
   */
  public static final Align FIT_VERTICAL = new Align(TOP.getValue() | BOTTOM.getValue()) ;

  /**
   * <zh>ˮƽʹֱӦС.</zh>
   * <en>Complete fit. </en>
   */
  public static final Align FIT = new Align(FIT_HORIZONTAL.getValue() | FIT_VERTICAL.getValue()) ;

  /**
   * <zh>(ͬ˶.).</zh>
   * <en>North alignment (same as TOP).</en>
   */
  public static final Align NORTH = TOP ;

  /**
   * <zh>ϱ(ͬ׶˶).</zh>
   * <en>South alignment (same as BOTTOM).</en>
   */
  public static final Align SOUTH = BOTTOM ;

  /**
   * <zh>(ͬ).</zh>
   * <en>West alignment (same as LEFT).</en>
   */
  public static final Align WEST = LEFT ;

  /**
   * <zh>(ͬҶ).</zh>
   * <en>East alignment (same as RIGHT).</en>
   */
  public static final Align EAST = RIGHT ;

  /**
   * <zh>(ͬTOP_LEFT).</zh>
   * <en>North/West alignment (same as TOP_LEFT).</en>
   */
  public static final Align NORTH_WEST = TOP_LEFT ;

  /**
   * <zh>(ͬTOP_RIGHT).</zh>
   * <en>North/East alignment (same as TOP_RIGHT).</en>
   */
  public static final Align NORTH_EAST = TOP_RIGHT ;

  /**
   * <zh>Ϸ(ͬBOTTOM_LEFT).</zh>
   * <en>South/West alignment (same as BOTTOM_LEFT).</en>
   */
  public static final Align SOUTH_WEST = BOTTOM_LEFT ;

  /**
   * <zh>Ϸ(ͬBOTTOM_RIGHT).</zh>
   * <en>South/East alignment (same as BOTTOM_RIGHT). </en>
   */
  public static final Align SOUTH_EAST = BOTTOM_RIGHT ;

  /**
   * <zh>ֵ.</zh>
   * <en>The value.</en>
   */
  private int value ;

  /**
   * <zh>˽йֹʵ.</zh>
   * <en>Private constructor to prevent instance.</en>
   *
   * @param value int <zh>ֵ֮һ.</zh>
   * <en>One of a constant of this class.</en>
   */
  private Align(int value) {
    super() ;
    this.value = value ;
  }

  /**
   * <zh>ֵ.</zh>
   * <en>Returns the value.</en>
   * @return int
   */
  public int getValue() {
    return this.value ;
  }

  /**
   * <zh>ݶ뷽ʽ(rect)һ(frame)Ķλ.</zh>
   * <en>Aligns one rectangle (rect) relative
   * to another rectangle (frame).</en>
   *
   * @param rect <zh>Ҫľ(Ϊnull).</zh>
   * <en>The rectangle to be aligned (null not permitted).</en>
   * @param frame <zh>Ծ(Ϊnull).</zh>
   * <en>The reference frame (null not permitted).</en>
   * @param align Align <zh>뷽ʽ(Ϊnull).</zh>
   * <en>The alignment.</en>
   * @return Rectangle2D <zh>rectʵλ.</zh>
   * <en>The aligned rect.</en>
   */
  public static Rectangle2D align(Rectangle2D rect, Rectangle2D frame, Align align) {
    if (rect == null) {
      throw new IllegalArgumentException("Null 'rect' argument.") ;
    }
    if (frame == null) {
      throw new IllegalArgumentException("Null 'frame' argument.") ;
    }
    if (align == null) {
      throw new IllegalArgumentException("Null 'align' argument.") ;
    }

    double x = frame.getCenterX() - rect.getWidth() / 2.0 ;
    double y = frame.getCenterY() - rect.getHeight() / 2.0 ;
    double w = rect.getWidth() ;
    double h = rect.getHeight() ;

    if ((align.getValue() & FIT_VERTICAL.getValue()) == FIT_VERTICAL.getValue()) {
      h = frame.getHeight() ;
    }

    if ((align.getValue() & FIT_HORIZONTAL.getValue()) == FIT_HORIZONTAL.getValue()) {
      w = frame.getWidth() ;
    }

    if ((align.getValue() & TOP.getValue()) == TOP.getValue()) {
      y = frame.getMinY() ;
    }

    if ((align.getValue() & BOTTOM.getValue()) == BOTTOM.getValue()) {
      y = frame.getMaxY() - h ;
    }

    if ((align.getValue() & LEFT.getValue()) == LEFT.getValue()) {
      x = frame.getX() ;
    }

    if ((align.getValue() & RIGHT.getValue()) == RIGHT.getValue()) {
      x = frame.getMaxX() - w ;
    }

    return new Rectangle2D.Double(x, y, w, h) ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Override toString method.</en>
   * @return String
   */
  public String toString() {
    return "[value=" + this.value + "]" ;
  }

  /**
   * <zh>hashCode.</zh>
   * <en>Override hashCode method.</en>
   * @return int
   */
  public int hashCode() {
    long bits = Double.doubleToLongBits(this.value) * 43 ;

    return (((int) bits) ^ ((int) (bits >> 32))) ;
  }

  /**
   * <zh>equals.</zh>
   * <en>Override equals method.</en>
   * @param obj Object
   * @return boolean
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }

    if (!(obj instanceof Align)) {
      return false ;
    }

    Align other = (Align) obj ;

    return this.value == other.value ;
  }

  /**
   * <zh>equals.</zh>
   * <en>Override equals method.</en>
   * @return Object
   * @throws CloneNotSupportedException
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }
}
