package com.ijchart.xychart.ui ;

import java.io.Serializable ;

import com.ijchart.xychart.shape.ParallelepipedTextChartShape ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ʵֽӿ{@link IEffect3D}</zh>
 * <en>A class that implements interface {@link IEffect3D}</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class DefaultEffect3D implements IEffect3D, Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 7988978616311694048L ;

  /**
   * <zh>X᷽ƫֵ.</zh>
   * <en>The offset.</en>
   */
  private double xOffset ;

  /**
   * <zh>Y᷽ƫֵ.</zh>
   * <en>The offset.</en>
   */
  private double yOffset ;

  /**
   * <zh>ȱʡĹ.</zh>
   * <en>The default constructor.</en>
   */
  public DefaultEffect3D() {
    this(ParallelepipedTextChartShape.DEFAULT_X_OFFSET,
         ParallelepipedTextChartShape.DEFAULT_Y_OFFSET) ;
  }

  /**
   * <zh>ȱʡĹ.</zh>
   * <en>The default constructor.</en>
   *
   * @param xOffset double <zh>X᷽ƫֵ.</zh>
   * <en>The offset.</en>
   * @param yOffset double <zh>Y᷽ƫֵ.</zh>
   * <en>The offset.</en>
   */
  public DefaultEffect3D(double xOffset, double yOffset) {
    this.xOffset = xOffset ;
    this.yOffset = yOffset ;
  }

  /**
   * <zh>ػ3DЧʱX᷽ƫֵ.</zh>
   * <en>Returns the x-offset (in Java2D units) for the 3D effect.</en>
   *
   * @return double <zh>X᷽ƫֵ.</zh>
   * <en>The offset.</en>
   */
  public double getXOffset() {
    return this.xOffset ;
  }

  /**
   * <zh>ػ3DЧʱY᷽ƫֵ.</zh>
   * <en>Returns the y-offset (in Java2D units) for the 3D effect.</en>
   *
   * @return double <zh>Y᷽ƫֵ.</zh>
   * <en>The offset.</en>
   */
  public double getYOffset() {
    return this.yOffset ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof DefaultEffect3D)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    DefaultEffect3D that = (DefaultEffect3D) obj ;
    if (this.xOffset != that.xOffset) {
      return false ;
    }
    if (this.yOffset != that.yOffset) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() ;
    result = 37 * result + this.xOffset ;
    result = 37 * result + this.yOffset ;

    return (int) result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[xOffset=" + this.xOffset + ",yOffset=" + this.yOffset + "]" ;
  }
}
