package com.ijchart.xychart.util ;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.GradientPaint ;
import java.awt.Graphics ;
import java.awt.Image ;
import java.awt.Paint ;
import java.awt.Shape ;
import java.awt.Stroke ;
import java.awt.geom.Arc2D ;
import java.awt.geom.Ellipse2D ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.Line2D ;
import java.awt.geom.PathIterator ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.awt.image.BufferedImage ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.text.AttributedCharacterIterator ;
import java.text.AttributedString ;
import java.text.CharacterIterator ;
import java.util.Enumeration ;
import java.util.HashMap ;
import java.util.Hashtable ;
import java.util.Iterator ;
import java.util.Map ;
import java.util.Set ;

import javax.imageio.ImageIO ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>л,ЩJava׼ûʵjava.io.Serializableӿ,вֱл,Ҫʹô
 * л</zh>
 * <en>A class containing useful utility methods relating to serialization</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public abstract class SerialUtil {

  /**
   * <zh>һʵSerializableӿڷtrue,򷵻false</zh>
   * <en>Returns true if a class implements Serializable and false otherwise</en>
   * @param c  the class.
   *
   * @return A boolean.
   */
  public static boolean isSerializable(final Class c) {
    return Serializable.class.isAssignableFrom(c) ;
  }

  /**
   * <zh>ȡ{@link #writePaint(ObjectOutputStream,Paint)}лPaint</zh>
   * <en>Reads a Paint object that has been serialised by the
   * {@link #writePaint(ObjectOutputStream,Paint)} method</en>
   *
   * @param stream ObjectOutputStream <zh>ObjectInputStream(Ϊnull)</zh>
   * <en>the input stream (null not permitted).</en>
   * @return The paint object (possibly null).
   * @throws IOException  if there is an I/O problem.
   * @throws ClassNotFoundException  if there is a problem loading a class.
   */
  public static Paint readPaint(final ObjectInputStream stream) throws IOException,
    ClassNotFoundException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument") ;
    }

    Paint result = null ;
    final boolean isNull = stream.readBoolean() ;
    if (!isNull) {
      final Class c = (Class) stream.readObject() ;
      if (isSerializable(c)) {
        result = (Paint) stream.readObject() ;
      } else if (c.equals(GradientPaint.class)) {
        final float x1 = stream.readFloat() ;
        final float y1 = stream.readFloat() ;
        final Color c1 = (Color) stream.readObject() ;
        final float x2 = stream.readFloat() ;
        final float y2 = stream.readFloat() ;
        final Color c2 = (Color) stream.readObject() ;
        final boolean isCyclic = stream.readBoolean() ;
        result = new GradientPaint(x1, y1, c1, x2, y2, c2, isCyclic) ;
      }
    }

    return result ;
  }

  /**
   * <zh>лPaint</zh>
   * <en>Serialises a Paint object.</en>
   *
   * @param stream ObjectOutputStream <zh>(Ϊnull)</zh>
   * <en>the output stream (null not permitted).</en>
   * @param paint <zh>Paint (Ϊnull)</zh>
   * <en>the paint object (null permitted).</en>
   * @throws IOException if there is an I/O error.
   */
  public static void writePaint(final ObjectOutputStream stream, final Paint paint) throws
    IOException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument") ;
    }

    if (paint != null) {
      stream.writeBoolean(false) ;
      stream.writeObject(paint.getClass()) ;
      if (paint instanceof Serializable) {
        stream.writeObject(paint) ;
      } else if (paint instanceof GradientPaint) {
        final GradientPaint gp = (GradientPaint) paint ;
        stream.writeFloat((float) gp.getPoint1().getX()) ;
        stream.writeFloat((float) gp.getPoint1().getY()) ;
        stream.writeObject(gp.getColor1()) ;
        stream.writeFloat((float) gp.getPoint2().getX()) ;
        stream.writeFloat((float) gp.getPoint2().getY()) ;
        stream.writeObject(gp.getColor2()) ;
        stream.writeBoolean(gp.isCyclic()) ;
      }
    } else {
      stream.writeBoolean(true) ;
    }

  }

  /**
   * <zh>ȡ{@link SerialUtil#writeStroke(ObjectOutputStream,Stroke)}лStroke</zh>
   * <en>Reads a Stroke object that has been serialised by the
   *  {@link SerialUtil#writeStroke(ObjectOutputStream,Stroke)} method</en>
   *
   * @param stream ObjectOutputStream <zh>(Ϊnull)</zh>
   * <en>the input stream (null not permitted).</en>
   * @return Stroke <zh>Stroke (Ϊnull).</zh>
   * <en>The stroke object (possibly null).</en>
   * @throws IOException  if there is an I/O problem.
   * @throws ClassNotFoundException  if there is a problem loading a class.
   */
  public static Stroke readStroke(final ObjectInputStream stream) throws IOException,
    ClassNotFoundException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    Stroke result = null ;
    final boolean isNull = stream.readBoolean() ;
    if (!isNull) {
      final Class c = (Class) stream.readObject() ;
      if (c.equals(BasicStroke.class)) {
        final float width = stream.readFloat() ;
        final int cap = stream.readInt() ;
        final int join = stream.readInt() ;
        final float miterLimit = stream.readFloat() ;
        final float[] dash = (float[]) stream.readObject() ;
        final float dashPhase = stream.readFloat() ;
        result = new BasicStroke(
          width, cap, join, miterLimit, dash, dashPhase
                 ) ;
      } else {
        result = (Stroke) stream.readObject() ;
      }
    }

    return result ;
  }

  /**
   * <zh>лStroke,˴ֻStrokeJDKȱʡʵBasicStroke</zh>
   * <en>Serialises a Stroke object.This code handles the BasicStroke class which is
   * the only Stroke implementation provided by the JDK (and isn't directly Serializable).</en>
   *
   * @param stream ObjectOutputStream <zh>(Ϊnull)</zh>
   * <en>the output stream (null not permitted).</en>
   * @param stroke <zh>Stroke(Ϊnull)</zh>
   * <en>the stroke object (null permitted).</en>
   * @throws IOException if there is an I/O error.
   */
  public static void writeStroke(final ObjectOutputStream stream, final Stroke stroke) throws
    IOException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    if (stroke != null) {
      stream.writeBoolean(false) ;
      if (stroke instanceof BasicStroke) {
        final BasicStroke s = (BasicStroke) stroke ;
        stream.writeObject(BasicStroke.class) ;
        stream.writeFloat(s.getLineWidth()) ;
        stream.writeInt(s.getEndCap()) ;
        stream.writeInt(s.getLineJoin()) ;
        stream.writeFloat(s.getMiterLimit()) ;
        stream.writeObject(s.getDashArray()) ;
        stream.writeFloat(s.getDashPhase()) ;
      } else {
        stream.writeObject(stroke.getClass()) ;
        stream.writeObject(stroke) ;
      }
    } else {
      stream.writeBoolean(true) ;
    }

  }

  /**
   * <zh>ȡ{@link #writeShape(ObjectOutputStream,Shape)}лShape</zh>
   * <en>Reads a Shape object that has been serialised by the
   * {@link #writeShape(ObjectOutputStream,Shape)} method</en>
   *
   * @param stream ObjectOutputStream <zh>(Ϊnull)</zh>
   * <en>the input stream (null not permitted).</en>
   * @return <zh>Shape (Ϊnull)</zh><en>The shape object (possibly null).</en>
   * @throws IOException <zh>IO</zh><en>if there is an I/O problem.</en>
   * @throws ClassNotFoundException <zh></zh>
   * <en>if there is a problem loading a class</en>
   */
  public static Shape readShape(final ObjectInputStream stream) throws IOException,
    ClassNotFoundException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    Shape result = null ;
    final boolean isNull = stream.readBoolean() ;
    if (!isNull) {
      final Class c = (Class) stream.readObject() ;
      if (c.equals(Line2D.class)) {
        final double x1 = stream.readDouble() ;
        final double y1 = stream.readDouble() ;
        final double x2 = stream.readDouble() ;
        final double y2 = stream.readDouble() ;
        result = new Line2D.Double(x1, y1, x2, y2) ;
      } else if (c.equals(Rectangle2D.class)) {
        final double x = stream.readDouble() ;
        final double y = stream.readDouble() ;
        final double w = stream.readDouble() ;
        final double h = stream.readDouble() ;
        result = new Rectangle2D.Double(x, y, w, h) ;
      } else if (c.equals(Ellipse2D.class)) {
        final double x = stream.readDouble() ;
        final double y = stream.readDouble() ;
        final double w = stream.readDouble() ;
        final double h = stream.readDouble() ;
        result = new Ellipse2D.Double(x, y, w, h) ;
      } else if (c.equals(Arc2D.class)) {
        final double x = stream.readDouble() ;
        final double y = stream.readDouble() ;
        final double w = stream.readDouble() ;
        final double h = stream.readDouble() ;
        final double as = stream.readDouble() ; // Angle Start
        final double ae = stream.readDouble() ; // Angle Extent
        final int at = stream.readInt() ; // Arc type
        result = new Arc2D.Double(x, y, w, h, as, ae, at) ;
      } else if (c.equals(GeneralPath.class)) {
        final GeneralPath gp = new GeneralPath() ;
        final float[] args = new float[6] ;
        boolean hasNext = stream.readBoolean() ;
        while (!hasNext) {
          final int type = stream.readInt() ;
          for (int i = 0 ; i < 6 ; i++) {
            args[i] = stream.readFloat() ;
          }
          switch (type) {
          case PathIterator.SEG_MOVETO:
            gp.moveTo(args[0], args[1]) ;
            break ;
          case PathIterator.SEG_LINETO:
            gp.lineTo(args[0], args[1]) ;
            break ;
          case PathIterator.SEG_CUBICTO:
            gp.curveTo(args[0], args[1], args[2], args[3], args[4], args[5]) ;
            break ;
          case PathIterator.SEG_QUADTO:
            gp.quadTo(args[0], args[1], args[2], args[3]) ;
            break ;
          case PathIterator.SEG_CLOSE:

            //result = gp;
            break ;
          default:
            throw new RuntimeException("IJChart - No path exists") ;
          }
          gp.setWindingRule(stream.readInt()) ;
          hasNext = stream.readBoolean() ;
        }
        result = gp ;
      } else {
        result = (Shape) stream.readObject() ;
      }
    }

    return result ;
  }

  /**
   * <zh>лShape</zh>
   * <en>Serialises a Shape object</en>
   *
   * @param stream ObjectOutputStream <zh> (Ϊnull)</zh>
   * <en>the output stream (null not permitted).</en>
   * @param shape <zh>ҪлĶ</zh>
   * <en>the shape object (null permitted).</en>
   * @throws IOException if there is an I/O error.
   */
  public static void writeShape(final ObjectOutputStream stream, final Shape shape) throws
    IOException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    if (shape != null) {
      stream.writeBoolean(false) ;
      if (shape instanceof Line2D) {
        final Line2D line = (Line2D) shape ;
        stream.writeObject(Line2D.class) ;
        stream.writeDouble(line.getX1()) ;
        stream.writeDouble(line.getY1()) ;
        stream.writeDouble(line.getX2()) ;
        stream.writeDouble(line.getY2()) ;
      } else if (shape instanceof Rectangle2D) {
        final Rectangle2D rectangle = (Rectangle2D) shape ;
        stream.writeObject(Rectangle2D.class) ;
        stream.writeDouble(rectangle.getX()) ;
        stream.writeDouble(rectangle.getY()) ;
        stream.writeDouble(rectangle.getWidth()) ;
        stream.writeDouble(rectangle.getHeight()) ;
      } else if (shape instanceof Ellipse2D) {
        final Ellipse2D ellipse = (Ellipse2D) shape ;
        stream.writeObject(Ellipse2D.class) ;
        stream.writeDouble(ellipse.getX()) ;
        stream.writeDouble(ellipse.getY()) ;
        stream.writeDouble(ellipse.getWidth()) ;
        stream.writeDouble(ellipse.getHeight()) ;
      } else if (shape instanceof Arc2D) {
        final Arc2D arc = (Arc2D) shape ;
        stream.writeObject(Arc2D.class) ;
        stream.writeDouble(arc.getX()) ;
        stream.writeDouble(arc.getY()) ;
        stream.writeDouble(arc.getWidth()) ;
        stream.writeDouble(arc.getHeight()) ;
        stream.writeDouble(arc.getAngleStart()) ;
        stream.writeDouble(arc.getAngleExtent()) ;
        stream.writeInt(arc.getArcType()) ;
      } else if (shape instanceof GeneralPath) {
        stream.writeObject(GeneralPath.class) ;
        final PathIterator pi = shape.getPathIterator(null) ;
        final float[] args = new float[6] ;
        stream.writeBoolean(pi.isDone()) ;
        while (!pi.isDone()) {
          final int type = pi.currentSegment(args) ;
          stream.writeInt(type) ;
          // TODO: could write this to only stream the values
          // required for the segment type
          for (int i = 0 ; i < 6 ; i++) {
            stream.writeFloat(args[i]) ;
          }
          stream.writeInt(pi.getWindingRule()) ;
          pi.next() ;
          stream.writeBoolean(pi.isDone()) ;
        }
      } else {
        stream.writeObject(shape.getClass()) ;
        stream.writeObject(shape) ;
      }
    } else {
      stream.writeBoolean(true) ;
    }

  }

  /**
   * <zh>ȡ{@link SerialUtil#writePoint2D(ObjectOutputStream,Point2D)}лPoint2D</zh>
   * <en>Reads a Point2D object that has been serialised by the
   * {@link SerialUtil#writePoint2D(ObjectOutputStream,Point2D)} method</en>
   *
   * @param stream ObjectOutputStream <zh> (Ϊnull)</zh>
   * <en>the input stream (null not permitted).</en>
   * @return <zh>Point2D</zh>
   * <en>The point object (possibly null).</en>
   * @throws IOException  if there is an I/O problem.
   */
  public static Point2D readPoint2D(final ObjectInputStream stream) throws IOException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    Point2D result = null ;
    final boolean isNull = stream.readBoolean() ;
    if (!isNull) {
      final double x = stream.readDouble() ;
      final double y = stream.readDouble() ;
      result = new Point2D.Double(x, y) ;
    }

    return result ;
  }

  /**
   * <zh>лPoint2D</zh>
   * <en>Serialises a Point2D object</en>
   *
   * @param stream ObjectOutputStream <zh> (Ϊnull)</zh>
   * <en>output stream (null not permitted).</en>
   * @param point <zh>Point2D (Ϊnull)</zh>
   * <en>the point object (null permitted).</en>
   * @throws IOException if there is an I/O error.
   */
  public static void writePoint2D(final ObjectOutputStream stream, final Point2D point) throws
    IOException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    if (point != null) {
      stream.writeBoolean(false) ;
      stream.writeDouble(point.getX()) ;
      stream.writeDouble(point.getY()) ;
    } else {
      stream.writeBoolean(true) ;
    }

  }

  /**
   * <zh>ȡ{@link SerialUtil#writeAttributedString(AttributedString,ObjectOutputStream)}л
   * AttributedString</zh>
   * <en>Reads a AttributedString object that has been serialised by the
   * {@link SerialUtil#writeAttributedString(AttributedString,ObjectOutputStream)} method</en>
   *
   * @param stream ObjectOutputStream <zh> (Ϊnull)</zh>
   * <en>the input stream (null not permitted).</en>
   * @return <zh>AttributedString(Ϊnull)</zh>
   * <en>The attributed string object (possibly null).</en>
   * @throws IOException <zh>IO쳣</zh>
   * <en>if there is an I/O problem</en>
   * @throws ClassNotFoundException  if there is a problem loading a class
   */
  public static AttributedString readAttributedString(ObjectInputStream stream) throws IOException,
    ClassNotFoundException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    AttributedString result = null ;
    final boolean isNull = stream.readBoolean() ;
    if (!isNull) {
      // read string and attributes then create result
      String plainStr = (String) stream.readObject() ;
      result = new AttributedString(plainStr) ;
      char c = stream.readChar() ;
      int start = 0 ;
      while (c != CharacterIterator.DONE) {
        int limit = stream.readInt() ;
        Map atts = (Map) stream.readObject() ;
        result.addAttributes(atts, start, limit) ;
        start = limit ;
        c = stream.readChar() ;
      }
    }

    return result ;
  }

  /**
   * <zh>лAttributedString</zh>
   * <en>Serialises an AttributedString object</en>
   *
   * @param as <zh>AttributedString (Ϊnull)</zh>
   * <en>the attributed string object (null permitted).</en>
   * @param stream ObjectOutputStream <zh> (Ϊnull)</zh>
   * <en>the output stream (null not permitted).</en>
   * @throws IOException if there is an I/O error.
   */
  public static void writeAttributedString(AttributedString as, ObjectOutputStream stream) throws
    IOException {

    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    if (as != null) {
      stream.writeBoolean(false) ;
      AttributedCharacterIterator aci = as.getIterator() ;
      // build a plain string from aci
      // then write the string
      StringBuffer plainStr = new StringBuffer() ;
      char current = aci.first() ;
      while (current != CharacterIterator.DONE) {
        plainStr = plainStr.append(current) ;
        current = aci.next() ;
      }
      stream.writeObject(plainStr.toString()) ;

      // then write the attributes and limits for each run
      current = aci.first() ;
      int begin = aci.getBeginIndex() ;
      while (current != CharacterIterator.DONE) {
        // write the current character - when the reader sees that this
        // is not CharacterIterator.DONE, it will know to read the
        // run limits and attributes
        stream.writeChar(current) ;

        // now write the limit, adjusted as if beginIndex is zero
        int limit = aci.getRunLimit() ;
        stream.writeInt(limit - begin) ;

        // now write the attribute set
        Map atts = new HashMap(aci.getAttributes()) ;
        stream.writeObject(atts) ;
        current = aci.setIndex(limit) ;
      }
      // write a character that signals to the reader that all runs
      // are done...
      stream.writeChar(CharacterIterator.DONE) ;
    } else {
      // write a flag that indicates a null
      stream.writeBoolean(true) ;
    }

  }

  /**
   * <zh>жȡMap,MapÿԪΪ(Object,Paint).</zh>
   * <en>Reads a Map of (Object, Paint) elements from a stream.</en>
   *
   * @param in <zh>.</zh>
   * <en>The input stream.</en>
   *
   * @return The map.
   *
   * @throws IOException
   * @throws ClassNotFoundException
   */
  public static Map readPaintMap(ObjectInputStream in) throws IOException, ClassNotFoundException {
    boolean isNull = in.readBoolean() ;
    if (isNull) {
      return null ;
    }

    Map result = new HashMap() ;
    int count = in.readInt() ;

    for (int i = 0 ; i < count ; i++) {
      Object category = in.readObject() ;
      Paint paint = readPaint(in) ;
      result.put(category, paint) ;
    }

    return result ;
  }

  /**
   * <zh>дMap,MapÿԪΪ(Object,Paint).</zh>
   * <en>Writes a map of (Comparable, Paint) elements to a stream.</en>
   *
   * @param out <zh>.</zh>
   * <en>The output stream.</en>
   * @param map <zh>Map,MapÿԪΪ(Object,Paint) (Ϊnull).</zh>
   * <en>The map (null permitted).</en>
   *
   * @throws IOException
   */
  public static void writePaintMap(ObjectOutputStream out, Map map) throws IOException {
    if (map == null) {
      out.writeBoolean(true) ;
    } else {
      out.writeBoolean(false) ;

      Set keys = map.keySet() ;
      int count = keys.size() ;
      out.writeInt(count) ;

      Iterator iterator = keys.iterator() ;
      while (iterator.hasNext()) {
        Object key = iterator.next() ;
        out.writeObject(key) ;
        writePaint(out, (Paint) map.get(key)) ;
      }
    }
  }

  /**
   * <zh>дhashtablepaint.</zh>
   * <en>Write paint objects of a hashtable.</en>
   * @param stream ObjectOutputStream
   * @param paints Hashtable <zh>painthashtable (Ϊnull).</zh>
   * <en>The hashtable (null permitted).</en>
   * @throws IOException
   */
  public static void writeHashtablePaint(ObjectOutputStream stream, Hashtable paints) throws
    IOException {
    if (paints != null) {
      Enumeration keys = paints.keys() ;
      stream.writeInt(paints.size()) ;
      while (keys.hasMoreElements()) {
        Object key = keys.nextElement() ;
        SerialUtil.writePaint(stream, (Paint) paints.get(key)) ;
      }
    } else {
      stream.writeObject(null) ;
    }
  }

  /**
   * <zh>ȡһhashtablepaint.</zh>
   * <en>Reads a hashtable of (Object, Paint) elements from a stream</en>
   * @param stream ObjectInputStream
   * @return Hashtable
   * @throws IOException
   * @throws ClassNotFoundException
   */
  public static Hashtable readHashtablePaint(ObjectInputStream stream) throws IOException,
    ClassNotFoundException {
    int keySize = stream.readInt() ;
    Hashtable paints = new Hashtable(keySize) ;

    for (int i = 0 ; i < keySize ; i++) {
      Object key = stream.readObject() ;
      Paint paint = SerialUtil.readPaint(stream) ;
      paints.put(key, paint) ;
    }

    return paints ;
  }

  /**
   * <zh>дhashtablestroke.</zh>
   * <en>Write stroke objects of a hashtable.</en>
   * @param stream ObjectOutputStream
   * @param strokes Hashtable
   * @throws IOException
   */
  public static void writeHashtableStroke(ObjectOutputStream stream, Hashtable strokes) throws
    IOException {
    if (strokes != null) {
      Enumeration keys = strokes.keys() ;
      stream.writeInt(strokes.size()) ;
      while (keys.hasMoreElements()) {
        Object key = keys.nextElement() ;
        SerialUtil.writeStroke(stream, (Stroke) strokes.get(key)) ;
      }
    } else {
      stream.writeObject(null) ;
    }
  }

  /**
   * <zh>ȡһhashtablestroke.</zh>
   * <en>Reads a hashtable of (Object, Stroke) elements from a stream</en>
   * @param stream ObjectInputStream
   * @return Hashtable
   * @throws IOException
   * @throws java.lang.ClassNotFoundException
   */
  public static Hashtable readHashtableStroke(ObjectInputStream stream) throws IOException,
    ClassNotFoundException {
    int keySize = stream.readInt() ;
    Hashtable strokes = new Hashtable(keySize) ;

    for (int i = 0 ; i < keySize ; i++) {
      Object key = stream.readObject() ;
      Stroke stroke = SerialUtil.readStroke(stream) ;
      strokes.put(key, stroke) ;
    }

    return strokes ;
  }

  /**
   * <zh>лͼƬ</zh>
   * <en>Serialises a Image object.</en>
   *
   * @param stream ObjectOutputStream <zh>(Ϊnull)</zh>
   * <en>the output stream (null not permitted).</en>
   * @param image Image <zh>Image(Ϊnull)</zh>
   * <en>the image object (null permitted).</en>
   * @throws IOException if there is an I/O error.
   */
  public static void writeImage(final ObjectOutputStream stream, final Image image) throws
    IOException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument") ;
    }

    if (image != null) {
      stream.writeBoolean(false) ;
      BufferedImage bufferImage = new BufferedImage(image.getWidth(null), image.getHeight(null),
                                                    BufferedImage.TYPE_INT_RGB) ;
      Graphics g = image.getGraphics() ;
      g.drawImage(bufferImage, 0, 0, null) ;
      ImageIO.write(bufferImage, null, stream) ;
    } else {
      stream.writeBoolean(true) ;
    }
  }

  /**
   * <zh>ȡ{@link #writeImage(ObjectOutputStream,Image)}лImage</zh>
   * <en>Reads a Image object that has been serialised by the
   * {@link #writeImage(ObjectOutputStream,Image)} method</en>
   *
   * @param stream ObjectOutputStream <zh>ObjectInputStream(Ϊnull)</zh>
   * <en>the input stream (null not permitted).</en>
   * @return The image object (possibly null).
   * @throws IOException  if there is an I/O problem.
   * @throws ClassNotFoundException  if there is a problem loading a class.
   */
  public static Image readImage(final ObjectInputStream stream) throws IOException,
    ClassNotFoundException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument") ;
    }

    Image result = null ;
    final boolean isNull = stream.readBoolean() ;
    if (!isNull) {
      result = ImageIO.read(stream) ;
    }

    return result ;
  }

}
