package com.ijchart.xychart.util ;

import java.awt.BasicStroke ;
import java.awt.Shape ;
import java.awt.Stroke ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.Line2D ;
import java.awt.geom.Rectangle2D ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>StrokeĹ.</zh>
 * <en>Nice static helpers for working with Stroke.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public abstract class StrokeUtil {

  /**
   * <zh>strokeĿ.strokeBasicStroke,BasicStroke.getLineWidth(),򷵻1.</zh>
   * <en>Returns the width of stroke shape.If stroke instanceof BasicStroke return
   *  BasicStroke.getLineWidth(),otherwise return 1.</en>
   * @param stroke Stroke <zh>Stroke.</zh>
   * <en>The Stroke object.</en>
   *
   * @return float <zh>strokeĿ.</zh>
   * <en>The stroke width.</en>
   */
  public static float getStrokeWidth(Stroke stroke) {
    if (stroke == null) {
      return 0 ;
    }

    if (stroke instanceof BasicStroke) {
      BasicStroke basicStroke = (BasicStroke) stroke ;
      return basicStroke.getLineWidth() ;
    }

    return 1 ;
  }

  /**
   * <zh>ֱߺͻֱߵĻ(stroke),ֱߵľ.</zh>
   * <en>Returns the high-precision bounding box of this Line2D.</en>
   *
   * @param line Line2D <zh>ֱ (Ϊnull).</zh>
   * <en>The line (null permitted).</en>
   * @param stroke Stroke <zh> (Ϊnull).</zh>
   * <en>The stroke object (null permitted).</en>
   * @param rotate double <zh>תĻ.</zh>
   * <en>The rotate angle.</en>
   *
   * @return Shape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  public static Shape lineToRectangle(Line2D line, Stroke stroke, double rotate) {
    if (line == null) {
      return null ;
    }

    float strokeWidth = (stroke == null) ? 1 : getStrokeWidth(stroke) ;
    if (strokeWidth < 3) {
      strokeWidth = 3 ;
    }
    double halfStrokeWidth = 0.5 * strokeWidth ;

    Shape result = null ;
    double x1 = line.getX1() ;
    double y1 = line.getY1() ;
    double x2 = line.getX2() ;
    double y2 = line.getY2() ;

    if (x1 <= x2) {
      if (y1 < y2) { //2-->4
        double boundWidth = Math.abs(x2 - x1) ;
        double boundHeight = Math.abs(y2 - y1) ;
        double angle = Math.atan(boundWidth / boundHeight) ; //н

        double x = x1 - halfStrokeWidth * Math.cos(angle) ;
        double y = y1 + halfStrokeWidth * Math.sin(angle) ;
        double height = Math.sqrt(Math.pow((x2 - x1), 2) + Math.pow((y2 - y1), 2)) ;
        result = new Rectangle2D.Double(0, 0, strokeWidth, height) ;

        AffineTransform transform = AffineTransform.getTranslateInstance(x, y) ;
        transform.rotate(rotate - angle) ;
        result = transform.createTransformedShape(result) ;
      } else if (y1 == y2) { //ˮƽ
        Rectangle2D rect = new Rectangle2D.Double(0, 0, x2 - x1, strokeWidth) ;

        AffineTransform transform = AffineTransform.getTranslateInstance(x1, y1 - halfStrokeWidth) ;
        transform.rotate(rotate) ;
        result = transform.createTransformedShape(rect) ;
      } else { //3-->1
        double boundWidth = Math.abs(x2 - x1) ;
        double boundHeight = Math.abs(y2 - y1) ;
        double angle = Math.atan(boundWidth / boundHeight) ; //н

        double x = x1 - halfStrokeWidth * Math.cos(angle) ;
        double y = y1 - halfStrokeWidth * Math.sin(angle) ;
        double height = Math.sqrt(Math.pow((x2 - x1), 2) + Math.pow((y2 - y1), 2)) ;
        result = new Rectangle2D.Double(0, 0, height, strokeWidth) ;

        AffineTransform transform = AffineTransform.getTranslateInstance(x, y) ;
        transform.rotate(rotate - 0.5 * Math.PI + angle) ;
        result = transform.createTransformedShape(result) ;
      }
    } else {
      if (y1 < y2) { //3-->1
        double boundWidth = Math.abs(x2 - x1) ;
        double boundHeight = Math.abs(y2 - y1) ;
        double angle = Math.atan(boundWidth / boundHeight) ; //н

        double x = x1 - halfStrokeWidth * Math.cos(angle) ;
        double y = y1 - halfStrokeWidth * Math.sin(angle) ;
        double height = Math.sqrt(Math.pow((x2 - x1), 2) + Math.pow((y2 - y1), 2)) ;
        result = new Rectangle2D.Double(0, 0, strokeWidth, height) ;

        AffineTransform transform = AffineTransform.getTranslateInstance(x, y) ;
        transform.rotate(rotate + angle) ;
        result = transform.createTransformedShape(result) ;
      } else if (y1 == y2) { //ˮƽ
        Rectangle2D rect = new Rectangle2D.Double(0, 0, x1 - x2, strokeWidth) ;

        AffineTransform transform = AffineTransform.getTranslateInstance(x1, y1 + halfStrokeWidth) ;
        transform.rotate(rotate + Math.PI) ;
        result = transform.createTransformedShape(rect) ;
      } else { //4-->2
        double boundWidth = Math.abs(x2 - x1) ;
        double boundHeight = Math.abs(y2 - y1) ;
        double angle = Math.atan(boundWidth / boundHeight) ; //н

        double x = x1 - halfStrokeWidth * Math.cos(angle) ;
        double y = y1 + halfStrokeWidth * Math.sin(angle) ;
        float height = (float) Math.sqrt(Math.pow((x2 - x1), 2) + Math.pow((y2 - y1), 2)) ;
        GeneralPath path = new GeneralPath() ;
        path.moveTo(0, 0) ;
        path.lineTo(0F, -height) ;
        path.lineTo(strokeWidth, -height) ;
        path.lineTo(strokeWidth, 0) ;
        path.closePath() ;

        AffineTransform transform = AffineTransform.getTranslateInstance(x, y) ;
        transform.rotate(rotate - angle) ;
        result = transform.createTransformedShape(path) ;
      }
    }

    return result ;
  }
}
